/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_dump_self.c,v 1.117 2007/02/21 03:19:16 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/smp_lock.h>
#include <linux/pipe_fs_i.h>
#include <linux/binfmts.h>
#include <linux/major.h>

#include "vmadump.h"
#include "cr_context.h"

// Create a core file
static int cr_do_coredump(cr_chkpt_req_t *req, cr_chkpt_proc_req_t *proc_req, struct file *filp, int shared)
{
	int result = -EBADF;

	if (shared) {
		down(&req->dest.mutex);
	} 

	if (filp) {
		lock_kernel();
		result = -ENOSYS;
		if (current->binfmt && current->binfmt->core_dump) {
			unsigned long core_mask ;

			// core_mask has PF_DUMPCORE bit iff current did before
			core_mask = current->flags | ~PF_DUMPCORE;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,17))
			// In kernel versions prior to 2.4.17 you should hold
			// the mmap_sem for reading when invoking core dumps.
			cr_mmap_sem_down_read(&current->mm->mmap_sem);
			result = !current->binfmt->core_dump(0,
						    get_pt_regs(current),
						    filp);
			cr_mmap_sem_up_read(&current->mm->mmap_sem);
#else
			// In kernel versions 2.4.17 and up the core dump
			// method handles the necessary locking.
			result = !current->binfmt->core_dump(0,
							get_pt_regs(current),
							filp);
#endif

			// restore PF_DUMPCORE bit to its previous state
			current->flags &= core_mask;
		}
		unlock_kernel();
	}
 
	if (shared) {
		up(&req->dest.mutex);
	} 

	return result;
}

/*
 * XXX:  Fix to return number of bytes written
 */
static int cr_save_linkage(cr_chkpt_req_t *req, cr_chkpt_proc_req_t *proc_req, struct file *filp)
{
    struct list_head *l;
    struct cr_context_tasklinkage *linkage;
    int linkage_size;
    cr_task_t *cr_task;
    struct task_struct *task;
    int task_count;
    int task_num;
    int retval;

    /* first count to find out how many tasks there are */
    read_lock(&req->lock);
    task_count = 0;
    list_for_each(l, &req->tasks) {
        cr_task = list_entry(l, cr_task_t, req_list);
	if (cr_task->chkpt_proc_req != proc_req) continue;
        ++task_count;
    }
    read_unlock(&req->lock);

    /* check task count, just ensure non-zero and that we're not bigger 
     * than 128kB (for kmalloc) */
    linkage = NULL;
    linkage_size = sizeof(*linkage) * task_count;
    if ((task_count > 0) && linkage_size < 131072) {
        /* allocate a cr_context_linkage large enough to hold all of the
         * available tasks */
        linkage = kmalloc(linkage_size, GFP_KERNEL);
    }
    if (linkage == NULL) {
        retval = -ENOMEM;
	goto out;
    }

    /* XXX:  There is a race here.  Theoretically, someone could add
     * or remove a task while we're not holding req->lock...  Probably
     * should punt, force maximum memory allocation early on, and then
     * just write out the number of tasks scanned. */

    read_lock(&tasklist_lock);
    read_lock(&req->lock);
    /* loop over each task, and copy important fields into the linkage */
    task_num = 0;
    list_for_each(l, &req->tasks) {
        cr_task = list_entry(l, cr_task_t, req_list);
	if (cr_task->chkpt_proc_req != proc_req) continue;
	if (cr_task == NULL) {
	    CR_KTRACE_BADPARM("NULL cr_task found on task list!");
	    retval = -EINVAL;
	    goto out_nulltask;
	}
	task = cr_task->task;

	/* XXX: can a task that has exited still be on the task list?  If so,
	 * how would we ever now? 
	 *   -- Must we even care?  Even if the task_struct is garbage, it
	 *   only matters we'll build a new process around it... 
	 */

	linkage[task_num].myptr       = task;
#ifdef CR_REAL_PARENT
	linkage[task_num].real_parent = CR_REAL_PARENT(task);
#endif
	linkage[task_num].parent      = CR_PARENT(task);

	/* Handle init specially */
#ifdef CR_REAL_PARENT
	if (linkage[task_num].real_parent == child_reaper) {
	    linkage[task_num].real_parent = CR_PARENT_INIT;
	}
#endif
	if (linkage[task_num].parent == child_reaper) {
	    linkage[task_num].parent = CR_PARENT_INIT;
	}

	linkage[task_num].pid     = task->pid;
	linkage[task_num].tgid    = task->tgid;
	linkage[task_num].pgrp    = cr_task_pgrp(task);
	linkage[task_num].tty_old_pgrp = cr_task_tty_old_pgrp(task);
	linkage[task_num].session = cr_task_session(task);
	linkage[task_num].leader  = cr_task_leader(task);
	linkage[task_num].exit_signal  = task->exit_signal;

	linkage[task_num].cr_task.stopped   = cr_task->stopped;
	linkage[task_num].cr_task.phase     = cr_task->phase;
	linkage[task_num].cr_task.handler   = cr_task->handler;
	linkage[task_num].cr_task.orig_filp = cr_task->filp;
	++task_num;
    }

out_nulltask:
    read_unlock(&req->lock);
    read_unlock(&tasklist_lock);

    /* This will catch the goto, and a possible race condition */
    if (task_num < task_count - 1) {
        CR_KTRACE_BADPARM("Warning: wrong number of tasks counted: task_count = %d, task_num = %d", 
	    task_count, task_num);
	retval = -EINVAL;
    } else {
        /* write the length of the linkage structure to the context file */
        retval = cr_kwrite(filp, &linkage_size, sizeof(linkage_size));
	if (retval < sizeof(linkage_size)) {
            CR_KTRACE_LOW_LVL("cr_kwrite wrote %d of %d bytes", retval, 
			      (int)sizeof(linkage_size));
            /* XXX: This is an error...  we should abort! */
	} else {
	    /* now write the linkage structure */
            retval = cr_kwrite(filp, linkage, linkage_size);
	    if (retval < linkage_size) {
                CR_KTRACE_LOW_LVL("cr_kwrite wrote %d of %d bytes", 
				  retval, linkage_size);
	    }
	}
    }

    kfree(linkage);
out:
    return retval;
}

static struct file *
cr_dest_gett(cr_chkpt_req_t *req, struct task_struct *task, cr_obj_t what)
{
    int shared = -1;

    return cr_dest_get(&req->dest, &shared);
    CR_KTRACE_LOW_LVL("shared = %d", shared);
}

static void
cr_dest_putt(cr_chkpt_req_t *req, struct file *filp)
{
    cr_dest_put(&req->dest, filp);
}

/*
 * cr_fputs
 * 
 * Slightly different than fputs, since it writes the NULL also, and refuses to
 * write anything larger than a PAGE_SIZE
 * 
 * Returns number of bytes written on success.
 */
static int
cr_fputs(char *buf, struct file *filp)
{
    int ret;
    int wrote=0;
    int len;

    ret = -EIO;

    /* we actually want to know when someone wrote NULL to an area, so we
     * can restore the NULL on restart. */
    if (!buf) {
        /* magic length of 0 for NULL buf */
        len = 0;
    } else {
        /* we write out the length WITH the null termination. */
        len = strlen(buf)+1;
    }

    /* compare l+1 to avoid type issue wrt promotion of len+1 to unsigned
     * long in comparison against PAGE_SIZE */
    if (len < 0 || len+1 > PAGE_SIZE+1) {
        CR_ERR("cr_fputs:  String length (%d) out of bounds.", len);
        ret = -EINVAL;
        goto out;
    }

    ret = cr_kwrite(filp, &len, sizeof(len));
    wrote += ret;
    if (ret != sizeof(len)) {
        CR_ERR("cr_fputs: Bad file write");
        ret = -1;
        goto out;
    }
  
    if (len > 0) {
        ret = cr_kwrite(filp, buf, len);
        wrote += ret;
        if (ret != len) {
            CR_ERR("cr_fputs: Bad file write");
            ret = -1;
            goto out;
        }
    }

    ret = wrote;

out:
    return ret;
}

/*
 * XXX: change from PAGE_SIZE to PATH_MAX?
 */
static int 
cr_save_fs_struct(cr_chkpt_req_t *req)
{
    struct dentry *dentry;
    struct vfsmount *mnt;
    int retval;
    char *path;
    int umask;
    void *buf;
    struct file *filp;

    buf = (void *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_nomem;
    }

    filp = cr_dest_gett(req, current, cr_fs_obj);
    if (!filp) {
        retval = -EBADF;
        goto out_nofile;
    }

    read_lock(&current->fs->lock);
    umask = current->fs->umask;
    read_unlock(&current->fs->lock);

    /* save umask */
    retval = cr_kwrite(filp, &umask, sizeof(umask));
    if (retval < sizeof(umask)) {
        CR_ERR("umask: Bad file write!");
        goto out_writerr;
    }

    /* save root */
    read_lock(&current->fs->lock);
    dentry = current->fs->root;
    mnt = current->fs->rootmnt;
    read_unlock(&current->fs->lock);
    path = cr_getpath(dentry, mnt, buf, PAGE_SIZE);
    retval = cr_fputs(path, filp);
    if (retval < 0) {
        CR_ERR("root: Bad file write!");
        goto out_writerr;
    }

    /* save altroot -- can be NULL */
    read_lock(&current->fs->lock);
    dentry = current->fs->altroot;
    mnt = current->fs->altrootmnt;
    read_unlock(&current->fs->lock);
    if (dentry) {
        path = cr_getpath(dentry, mnt, buf, PAGE_SIZE);
    } else {
        path = NULL;
    }
    retval = cr_fputs(path, filp);
    if (retval < 0) {
        CR_ERR("altroot: Bad file write!");
        goto out_writerr;
    }

    /* save pwd */
    read_lock(&current->fs->lock);
    dentry = current->fs->pwd;
    mnt = current->fs->pwdmnt;
    read_unlock(&current->fs->lock);
    path = cr_getpath(dentry, mnt, buf, PAGE_SIZE);
    retval = cr_fputs(path, filp);
    if (retval < 0) {
        CR_ERR("pwd: Bad file write!");
        goto out_writerr;
    }

out_writerr:
    cr_dest_putt(req, filp);
out_nofile:
    free_page((unsigned long) buf);
out_nomem:
    return retval;
}

/*
 * Saves the files_struct
 *
 * This looks silly, but we actually restore the values of max_fds array
 * later.  Since we know how many files are open, we don't have to guess
 * the way get_unused_fd does.
 *
 * returns max_fds or error
 */
static int cr_save_files_struct(cr_chkpt_req_t *req, struct files_struct *files)
{
    cr_fdtable_t *fdt;
    struct cr_files_struct cr_fs;
    struct file *c_filp;
    int retval=0;

    c_filp = cr_dest_gett(req, current, cr_files_obj);
    if (!c_filp) {
        CR_ERR("Unable to find destination file for cr_files_struct.");
        retval = -EINVAL;
        goto out_no_cfilp;
    }

    cr_fs.cr_obj = cr_files_obj; 

#if !HAVE_STRUCT_FDTABLE
    cr_file_lock_read(&files->file_lock);
#endif

    rcu_read_lock();
    fdt = cr_fdtable(files);
    cr_fs.cr_max_fds   = fdt->max_fds;
#if HAVE_FILES_STRUCT_NEXT_FD
    cr_fs.cr_next_fd   = files->next_fd;
#else
    cr_fs.cr_next_fd   = fdt->next_fd;
#endif
    rcu_read_unlock();

#if !HAVE_STRUCT_FDTABLE
    cr_file_lock_unread(&files->file_lock);
#endif

    retval = cr_kwrite(c_filp, &cr_fs, sizeof(cr_fs));
    if (retval != sizeof(cr_fs)) {
        CR_ERR("cr_fs: Bad file write! (err=%d)", retval);
        retval = -EIO;
        goto out_writerr;
    }

    retval = cr_fs.cr_max_fds;
out_writerr:
    cr_dest_putt(req, c_filp);
out_no_cfilp:
    return retval;
}

/*
 * cr_get_fd_info
 *
 * Saves flags associated with the file descriptor.  (stored in files_struct)
 *
 * Must be called with the file_lock held, unless HAVE_STRUCT_FDTABLE for RCU.
 * 
 * Assumes that the file is open.  (descriptor refers to an open file in fd)
 */
static int
cr_get_fd_info(struct files_struct *files, int fd, struct cr_file_info *file_info)
{
    cr_fdtable_t *fdt;

    rcu_read_lock();
    fdt = cr_fdtable(files);
    if (FD_ISSET(fd, fdt->open_fds)) {
        file_info->fd      = fd;
        file_info->cloexec = FD_ISSET(fd, fdt->close_on_exec);
        file_info->orig_filp = fcheck(fd);
    } else {
        CR_WARN("cr_get_fd_info: Called on closed file!");
    }
    rcu_read_unlock();

    return 0;
}

static int
cr_get_file_info(cr_chkpt_req_t *req, struct file *filp, struct cr_file_info *file_info)
{
    struct dentry *dentry = filp->f_dentry;
    int retval=0;

    /* The nlink check might be redundant, but it costs us very little to be cautious here. */
    file_info->unlinked = (!IS_ROOT(dentry) && d_unhashed(dentry)) ||
			  (dentry->d_inode->i_nlink == 0) ||
			  (dentry->d_flags & DCACHE_NFSFS_RENAMED);

    if (cr_insert_object(req->map, file_info->orig_filp, file_info->orig_filp)) {
        /* Was in the object table (and thus is dup) */
        file_info->cr_file_type = cr_open_dup;
    } else {
        struct inode *inode = filp->f_dentry->d_inode;
        unsigned int mode = inode->i_mode;
        if (S_ISREG(mode)) {
            file_info->cr_file_type = cr_open_file;
        } else if (S_ISDIR(mode)) {
            file_info->cr_file_type = cr_open_directory;
        } else if (S_ISLNK(mode)) {
            file_info->cr_file_type = cr_open_link;
        } else if (S_ISFIFO(mode)) {
            file_info->cr_file_type = cr_open_fifo;
        } else if (S_ISSOCK(mode)) {
            file_info->cr_file_type = cr_open_socket;
        } else if (S_ISCHR(mode)) {
	    if (MAJOR(inode->i_rdev) == MEM_MAJOR) {
		/* Kludge for /dev/{null,zero,full,random,urandom} and friends */
                file_info->cr_file_type = cr_open_file;
	    } else {
                file_info->cr_file_type = cr_open_chr;
	    }
        } else if (S_ISBLK(mode)) {
            file_info->cr_file_type = cr_open_blk;
        } else {
            /* completely unknown */
            file_info->cr_file_type = cr_unknown_file;
        }
    }

    return retval;
}

static int
cr_save_file_info(cr_chkpt_req_t *req, struct cr_file_info *file_info)
{
    struct file *filp;
    int retval;

    filp = cr_dest_gett(req, current, cr_file_obj);
    if (!filp) {
        retval = -EINVAL;
        goto out;
    }

    retval = cr_kwrite(filp, file_info, sizeof(*file_info));
    if (retval != sizeof(*file_info)) {
        retval = -EINVAL;
        CR_ERR("cr_save_file_info: Bad write!");
        goto out_badwrite;
    }

    retval = 0;

out_badwrite:
    cr_dest_putt(req, filp);
out:
    return retval;
}

static int
cr_save_filename(struct file *c_filp, struct file *filp)
{
    int retval;
    char *buf;
    char *name;

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out;
    }

    /* find the file name and length */
    name = cr_getpath(filp->f_dentry, filp->f_vfsmnt, buf, PAGE_SIZE);
    if (name == NULL) {
        CR_ERR("Bad file name!");
        retval = -EBADF;
        goto out_bad;
    }

    /* now write out the name */
    retval = cr_fputs(name, c_filp);
    if (retval < strlen(name)) {
        CR_ERR("cr_save_filename - Bad file write! (fputs returned %d)", retval);
        goto out_bad;
    }

out_bad:
    free_page((unsigned long) buf);
out:
    return retval;
}

static int
cr_save_open_file(cr_chkpt_req_t *req, struct file *filp)
{
    struct file *c_filp;                /* the context file */
    struct cr_open_file open_file;
    int retval = -EBADF;

    c_filp = cr_dest_gett(req, current, cr_file_obj);
    if (!c_filp) {
        goto out;
    }

    open_file.cr_type = cr_open_file_obj;
    /* XXX:  Fix to use better representation. */
    open_file.i_mode  = filp->f_dentry->d_inode->i_mode;
    open_file.i_size  = filp->f_dentry->d_inode->i_size;
    open_file.f_flags = filp->f_flags;
    open_file.f_pos   = filp->f_pos;

    /* write out the open_file struct */
    retval = cr_kwrite(c_filp, &open_file, sizeof(open_file));
    if (retval != sizeof(open_file)) {
        retval = -EIO;
        CR_ERR("cr_save_open_file - Bad file write!");
        goto out_putt;
    }

    retval = cr_save_filename(c_filp, filp);
    if (retval < 0) {
        CR_ERR("cr_save_open_file - Bad file write (filename)!");
    }

out_putt:
    cr_dest_putt(req, c_filp);
out:
    return retval;
}

static int 
cr_save_open_dir(cr_chkpt_req_t *req, struct file *filp)
{
    struct file *c_filp;                /* the context file */
    struct cr_open_dir open_dir;
    int retval = -EBADF;

    c_filp = cr_dest_gett(req, current, cr_dir_obj);
    if (!c_filp) {
        goto out;
    }

    open_dir.cr_type = cr_open_dir_obj;
    /* XXX:  Fix to use better representation. */
    open_dir.i_mode  = filp->f_dentry->d_inode->i_mode;
    open_dir.f_flags = filp->f_flags;
    open_dir.f_pos   = filp->f_pos;

    /* write out the open_dir struct */
    retval = cr_kwrite(c_filp, &open_dir, sizeof(open_dir));
    if (retval != sizeof(open_dir)) {
        retval = -EIO;
        CR_ERR("cr_save_open_dir - Bad file write!");
        goto out_putt;
    }

    retval = cr_save_filename(c_filp, filp);
    if (retval < 0) {
        CR_ERR("cr_save_open_dir - Bad file write (filename)!");
    }

out_putt:
    cr_dest_putt(req, c_filp);
out:
    return retval;
}

static int 
cr_save_open_link(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

/* Deterimine if a given FIFO is "internal", defined has having
 * at least one reader and one writer in the checkpoint set.
 *
 * XXX: This is almost certainly the worst possible way to do this.
 * Alas, I don't have a better way in mind that can be implemented
 * w/o need to move file saves/restores to the end (after memory of
 * all processes).
 */
static int
cr_fifo_is_internal(cr_chkpt_req_t *req, struct inode *inode) {
    struct files_struct *prev = NULL;
    struct list_head *l;
    int have_reader = 0;
    int have_writer = 0;
    unsigned long retval = 0; /* For size match w/ void* in object map */
    char *map_key = 1 + (char *)inode;

    read_lock(&req->lock);

    /* Look for cached value first.  Note "creative" use of key. */
    if (cr_find_object(req->map, map_key, (void **) &retval)) {
	goto out_cached;
    }

    list_for_each(l, &req->tasks) {
	int max_fds, fd;
	cr_fdtable_t *fdt;
	cr_task_t *cr_task = list_entry(l, cr_task_t, req_list);
	struct task_struct *task = cr_task->task;

	if (task->files == prev) continue; /* Same proc as previous thread */
	prev = task->files;

	cr_file_lock_read(&task->files->file_lock);
	rcu_read_lock();
	fdt = cr_fdtable(task->files);
	max_fds = fdt->max_fds;	/* Never shrinks, right? */
        rcu_read_unlock();

	for (fd = 0; !retval && (fd < max_fds); ++fd) {
	    struct file *filp = fcheck_files(task->files, fd);
	    if (!filp) continue;
	    get_file(filp);
	    if (filp->f_dentry->d_inode == inode) {
		if (filp->f_mode & FMODE_WRITE) have_writer = 1;
		if (filp->f_mode & FMODE_READ) have_reader = 1;
	    }
	    fput(filp);
	    retval = (have_reader && have_writer);
	}

	cr_file_lock_unread(&task->files->file_lock);
	if (retval) break;
    }

    cr_insert_object(req->map, map_key, (void *)retval);
out_cached:
    read_unlock(&req->lock);
    CR_KTRACE_FUNC_EXIT("Return '%sternal'", retval ? "in" : "ex");
    return (int)retval;
}

static int 
cr_save_open_fifo(cr_chkpt_req_t *req, struct file *filp)
{
    struct cr_fifo cf_fifo;
    struct file *cf_filp;
    struct inode *inode;
    void *buf = NULL;
    int retval;

    retval = -EINVAL;
    if (!filp)
	goto out;

    cf_filp = cr_dest_gett(req, current, cr_fifo_obj);
    if (!cf_filp) {
        goto out;
    }

    inode = filp->f_dentry->d_inode;

    cf_fifo.cr_type = cr_fifo_obj;
    cf_fifo.fifo_id = inode;
    cf_fifo.fifo_dentry = filp->f_dentry;
    cf_fifo.fifo_vfsmnt = filp->f_vfsmnt;
    cf_fifo.fifo_len = -1;
    cf_fifo.fifo_internal = cr_fifo_is_internal(req, inode);

    if (!cr_insert_object(req->map, inode, inode)) {
	/* We are first to save: suck the data out of the pipe while holding the pipe semaphore */
	retval = -ERESTARTSYS;
	if (cr_pipe_lock_interruptible(inode)) {
	    goto out_putt;
	}
    #if HAVE_PIPE_INODE_INFO_BASE
	{
	    cf_fifo.fifo_len = PIPE_LEN(*inode);
	    buf = kmalloc(cf_fifo.fifo_len, GFP_KERNEL);
	    if (!buf) {
	        retval = -ENOMEM;
	        cr_pipe_unlock(inode);
	        goto out_putt;
	    }
	    memcpy(buf, (void*)PIPE_BASE(*inode)+PIPE_START(*inode), cf_fifo.fifo_len);
	}
    #else
	{
	    struct pipe_inode_info *pipe = inode->i_pipe;
	    int i, curbuf;
	    char *p;

	    buf = kmalloc(PIPE_BUFFERS * PAGE_SIZE, GFP_KERNEL); /* max size */
	    if (!buf) {
	        retval = -ENOMEM;
	        cr_pipe_unlock(inode);
	        goto out_putt;
	    }

	    p = buf;
	    cf_fifo.fifo_len = 0;
	    curbuf = pipe->curbuf;
	    for (i = 0; i < pipe->nrbufs; ++i) {
	        struct pipe_buffer *pbuf = pipe->bufs + curbuf;
	        struct pipe_buf_operations *ops = pbuf->ops;
	        void *addr;
#if HAVE_PIPE_BUF_OPERATIONS_PIN
		int error = ops->pin(pipe, pbuf);
		if (error) {
	          retval = error;
	          cr_pipe_unlock(inode);
	          goto out_free;
	        }
		addr = ops->map(pipe, pbuf, 0);
	        memcpy(p, addr, pbuf->len);
		ops->unmap(pipe, pbuf, addr);
#else
	        addr = ops->map(filp, pipe, pbuf);
	        memcpy(p, addr, pbuf->len);
	        ops->unmap(pipe, pbuf);
#endif
	        p += pbuf->len;
	        cf_fifo.fifo_len += pbuf->len;
	        curbuf = (curbuf + 1) & (PIPE_BUFFERS-1);
	    }
	}
    #endif
	cr_pipe_unlock(inode);
    } else {
	/* We've saved this fifo previously */
	/* cf_fifo.fifo_len == -1 will identify this case at restart time. */
    }

    /* pipe perms - for FIFOs, need to know what to pass to mknod() */
    cf_fifo.fifo_perms = inode->i_mode;

    /* posix flags (for open) */
    cf_fifo.fifo_pos = filp->f_pos;
    cf_fifo.fifo_flags = filp->f_flags;
    cf_fifo.fifo_version = filp->f_version;

    /* write out fifo structure */
    retval = cr_kwrite(cf_filp, &cf_fifo, sizeof(cf_fifo));
    if (retval != sizeof(cf_fifo)) {
	retval = -EIO;
	CR_WARN("cr_kwrite error! = %d", retval);
	goto out_free;
    }

    /* write the filename out */
    retval = cr_save_filename(cf_filp, filp);
    if (retval < 0) {
	CR_WARN("Error saving pipe filename. (err=%d)", retval);
	goto out_free;
    }
   
    /* write out pipe data last (unless saved previously) */
    if (cf_fifo.fifo_len != ((unsigned int)-1)) {
	retval = cr_kwrite(cf_filp, buf, cf_fifo.fifo_len);
	if (retval != cf_fifo.fifo_len) {
	    retval = -EIO;
	    CR_WARN("cr_kwrite error! = %d", retval);
	    goto out_free;
	}
    }

    retval = 0;

out_free:
    kfree(buf);  /* NULL ok */
out_putt:
    cr_dest_putt(req, cf_filp);
out:
    return retval;
}

static int 
cr_save_open_socket(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    /* XXX:  Hack to do absolutely nothing here... */
    CR_WARN("Skipping a socket.");
    retval = 0;

    return retval;
}

/*
 * Right now, saves major and minor numbers.
 */
static int 
cr_save_open_chr(cr_chkpt_req_t *req, struct file *filp)
{
    struct cr_chrdev cf_chrdev;
    struct file *cf_filp;
    struct inode *inode;
    int retval;

    retval = -EINVAL;
    if (!filp) 
      goto out;

    cf_filp = cr_dest_gett(req, current, cr_chrdev_obj);

    inode = filp->f_dentry->d_inode;

    cf_chrdev.cr_type  = cr_open_chr;
    cf_chrdev.cr_major = MAJOR(inode->i_rdev); 
    cf_chrdev.cr_minor = MINOR(inode->i_rdev);

    retval = cr_kwrite(cf_filp, &cf_chrdev, sizeof(cf_chrdev));
    if (retval != sizeof(cf_chrdev)) {
	retval = -EIO;
	CR_WARN("cr_kwrite error! = %d", retval);
	goto out_write;
    }
    retval = 0;

out_write:
    cr_dest_putt(req, cf_filp);
out:
    return retval;
}

static int 
cr_save_open_blk(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

/*
 * dup info is associated with a file descriptor, not a file,
 * so it's saved in cr_save_file_info, and nothing is saved here,
 * other than a placeholder data structure.
 */
static int 
cr_save_open_dup(cr_chkpt_req_t *req, struct file *filp)
{
    struct cr_dup cf_dup;
    struct file *cf_filp;
    int retval;

    retval = -EINVAL;
    if (!filp) 
      goto out;

    cf_filp = cr_dest_gett(req, current, cr_dup_obj);

    /* placeholder... just in case we need to do something later */
    cf_dup.cr_type  = cr_open_dup;

    retval = cr_kwrite(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
	retval = -EIO;
	CR_WARN("cr_kwrite error! = %d", retval);
	goto out_write;
    }
    retval = 0;

out_write:
    cr_dest_putt(req, cf_filp);
out:
    return retval;
}

static int
cr_save_file_locks(cr_chkpt_req_t *req, struct file *filp)
{
    int retval = -ENOSYS;

    return retval;
}

static int
cr_save_all_files(cr_chkpt_req_t *req)
{
    int retval;
    int fd;
    struct cr_file_info file_info;
    struct file *filp;
    int max_fds;

    CR_KTRACE_FUNC_ENTRY("");

    /* save the files info, and get max_fds as a side-effect */
    CR_KTRACE_HIGH_LVL("    ...files_struct");
    retval = cr_save_files_struct(req, current->files);
    if (retval < 0) {
        CR_ERR("cr_save_all_files: Error saving files_struct");
        goto out_nolocks;
    }
    max_fds = retval;

    CR_KTRACE_HIGH_LVL("    ...files");
    /* now save the per file info */
    cr_file_lock_read(&current->files->file_lock);
    for (fd = 0; fd < max_fds; ++fd) {
        /*
         * We have to do our own fget here to avoid a possible race on
         * file close.  (Probably impossible, but just to be on the safe
         * side.
         */

        /* loop around again if the file is not open */
        filp = fcheck(fd);
        if (! filp) {
            continue;
        }

        get_file(filp);

        // CR_KTRACE_LOW_LVL("fd %d is open", fd);

	memset(&file_info, 0, sizeof(file_info));
	file_info.cr_type = cr_file_info_obj;

        /* now save data for this file descriptor */
        cr_get_fd_info(current->files, fd, &file_info);

        cr_file_lock_unread(&current->files->file_lock);

        /* write out the file header */
        retval = cr_get_file_info(req, filp, &file_info);
        if (retval) {
            retval = -EBADF;
            CR_ERR("Unable to determine file info!"); 
        }
        retval = cr_save_file_info(req, &file_info);
        if (retval < 0) {
            retval = -EIO;
            CR_ERR("Write error!");
            fput(filp);
            goto out_nolocks;
        }

        switch(file_info.cr_file_type) {
        case cr_open_file:
            CR_KTRACE_LOW_LVL("    ...%d is regular file.", fd);
            retval = cr_save_open_file(req, filp);
            break;
        case cr_open_directory:
            CR_KTRACE_LOW_LVL("    ...%d is open directory.", fd);
            retval = cr_save_open_dir(req, filp);
            break; 
        case cr_open_link:
            CR_KTRACE_LOW_LVL("    ...%d is open symlink.", fd);
            retval = cr_save_open_link(req, filp);
            break;
        case cr_open_fifo:
            CR_KTRACE_LOW_LVL("    ...%d is open fifo.", fd);
            retval = cr_save_open_fifo(req, filp);
            break;
        case cr_open_socket:
            CR_KTRACE_LOW_LVL("    ...%d is open socket.", fd);
            retval = cr_save_open_socket(req, filp);
            break;
        case cr_open_chr:
            CR_KTRACE_LOW_LVL("    ...%d is open character device.", fd);
            retval = cr_save_open_chr(req, filp);
            break;
        case cr_open_blk:
            CR_KTRACE_LOW_LVL("    ...%d is an open block device.", fd);
            retval = cr_save_open_blk(req, filp);
            break;
        case cr_open_dup:
            CR_KTRACE_LOW_LVL("    ...%d is dup of %p", fd, file_info.orig_filp);
            retval = cr_save_open_dup(req, filp);
            break;
        case cr_bad_file_type:
            /* fall through */
        default:
            retval = -EBADF;
            break;
        }

        if (retval < 0) {
            fput(filp);
            retval = -EBADF;
            CR_ERR("Unable to save open file!");
            goto out_nolocks;
        }

        /* Now write out any locks we had on the file. */
        // CR_KTRACE_LOW_LVL("    ...locks on %d", fd);
        retval = cr_save_file_locks(req, filp);

        /* We did the fget() manually with the lock held. */
        fput(filp);

        cr_file_lock_read(&current->files->file_lock);
    }
    cr_file_lock_unread(&current->files->file_lock);

    /* now write one last record to indicate that there are no files left. */
    memset(&file_info, 0, sizeof(file_info));
    file_info.cr_type = cr_file_info_obj;
    file_info.cr_file_type = cr_end_of_files;
    file_info.fd = -1;
    retval = cr_save_file_info(req, &file_info);
    if (retval < 0) {
        retval = -EIO;
        CR_ERR("Write error!");
        goto out_nolocks;
    }

out_nolocks:
    return retval;
}

static int
cr_save_file_header(cr_chkpt_req_t *req, struct file *filp)
{
    int result;
    struct cr_context_file_header cf_fhead;

    result = 0;

    CR_KTRACE_LOW_LVL("Dumping file header");

    cf_fhead.magic[0] = 'C';
    cf_fhead.magic[1] = 'R';
    cf_fhead.version = CR_CONTEXT_VERSION;
    cf_fhead.scope = req->checkpoint_scope;
    result = cr_kwrite(filp, &cf_fhead, sizeof(cf_fhead));
    if (result != sizeof(cf_fhead)) {
	result = -EIO;
	CR_WARN("cr_kwrite error! = %d", result);
	goto out;
    }

    result = 0;

  out:
    return result;
}

static int
cr_save_header(cr_chkpt_req_t *req, struct file *filp, int thread_count)
{
    int result;
    struct cr_section_header cf_head;

    result = 0;

    CR_KTRACE_LOW_LVL("Dumping header for %d threads", thread_count);

    cf_head.num_threads = thread_count;
    cf_head.clone_flags = CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND;
#if HAVE_2_6_SIGNAL_STRUCT
    if (atomic_read(&current->signal->count) > 1) {
	cf_head.clone_flags |= CLONE_THREAD | CLONE_DETACHED;
    }
#endif

    result = cr_kwrite(filp, &cf_head, sizeof(cf_head));
    if (result != sizeof(cf_head)) {
	result = -EIO;
	CR_WARN("cr_kwrite error! = %d", result);
	goto out;
    }

    result = 0;

  out:
    return result;
}

// Create a vmadump file
static int cr_do_vmadump(cr_chkpt_req_t *req, cr_chkpt_proc_req_t *proc_req, struct file *filp, int shared)
{
    struct vmadump_threaddump_args vmad_args;
    int thread_count = proc_req->thread_count;
    int result=0;

    vmad_args.num_threads = thread_count;
    vmad_args.fd = -1;
    vmad_args.flags = 0;

    /* First arrival grabs mutex and starts to work...
       later arrivals find the work is all done */
    down(&proc_req->serial_mutex);

    /* Acquire dest mutex (if any) */
    if (!test_and_set_bit(0, &proc_req->done_init) && shared) {
	down(&req->dest.mutex);
    }

    /* Write out the header(s) */
    if (!test_and_set_bit(0, &proc_req->done_header)) {
        CR_KTRACE_HIGH_LVL("Preparing to dump %d threads of %s", thread_count, current->comm);
        result = cr_save_header(req, filp, thread_count);
        if (result < 0) {
            goto out_mutex;
        }
    }

    /* Now dump out the task linkage */
    if (!test_and_set_bit(0, &proc_req->done_linkage)) {
        CR_KTRACE_LOW_LVL("Writing the per-process linkage.");
        result = cr_save_linkage(req, proc_req, filp);
        if (result < 0) {
	    goto out_mutex;
        }
    }

    /* save the credentials (uid, gid, etc.) */
    if (!test_and_set_bit(0, &proc_req->done_creds)) {
	struct cr_context_creds cf_creds;

	cf_creds.uid   = current->uid;
	cf_creds.euid  = current->euid;
	cf_creds.suid  = current->suid;
	cf_creds.gid   = current->gid;
	cf_creds.egid  = current->egid;
	cf_creds.sgid  = current->sgid;

#if CR_RESTORE_IDS
	cf_creds.ngroups = current->ngroups;
	memcpy(cf_creds.groups, current->groups, sizeof(cf_creds.groups));
#endif

        CR_KTRACE_LOW_LVL("Writing credentials");

	result = cr_kwrite(filp, &cf_creds, sizeof(cf_creds));
        if (result < sizeof(cf_creds)) {
	    result = -EIO;
            goto out_mutex;
        }
    }
    up(&proc_req->serial_mutex);

    /* Now use vmadump to write out the state for the thread.
       ALL threads call this and serialization is done within */
    CR_KTRACE_LOW_LVL("Writing vmadump.");
    result = vmadump_freeze_threads(filp, get_pt_regs(current), &vmad_args);
    if (result > 0) {
	/* vmadump returns number of bytes written when successful */
	result = 0;
    } else {
	goto out;
    }
    CR_KTRACE_LOW_LVL("finished vmadump");

    /* Wait for all tasks to leave user space before saving files.
     * This is needed to ensure pipe bufs are not changing. */
    cr_barrier_wait(&req->prefiles_barrier);

    /* dump fs_struct (cwd, umask, etc.) */
    down(&proc_req->serial_mutex);
    if (!test_and_set_bit(0, &proc_req->done_fs)) {
        CR_KTRACE_HIGH_LVL("Writing the fs struct...");
        result = cr_save_fs_struct(req);
        if (result < 0) {
	    goto out_mutex;
        }
    }

    if (!test_and_set_bit(0, &proc_req->done_files)) {
        /* dump the open files */
        CR_KTRACE_HIGH_LVL("Writing the open file section...");
        result = cr_save_all_files(req);
        if (result < 0) {
	    goto out_mutex;
        }
    }

    result = 0; // XXX
out_mutex:
    up(&proc_req->serial_mutex);
out:
    if (!test_and_set_bit(0, &proc_req->done_fini) && shared) {
        up(&req->dest.mutex);
    }
    return result;
}

// Dump out to the given file
//
// XXX: When we define an options struct, it will need to be passed in too

/*
 * XXX:  Might want to change this a little (lot) when we support multiple
 * context files.  Could be a separate format for directory set, etc.  i.e.
 * cr_format_contextfile, cr_format_contextdir?
 */
static int cr_do_dump(cr_chkpt_req_t *req, cr_chkpt_proc_req_t *proc_req, struct file *filp, int shared)
{
	int result = -EINVAL;

	switch(req->dump_format) {
	case cr_format_vmadump:
		result = cr_do_vmadump(req, proc_req, filp, shared);
		break;

	case cr_format_coredump:
		result = cr_do_coredump(req, proc_req, filp, shared);
		break;

	default:
		CR_KTRACE_BADPARM("unrecognized dump_format '%d'", 
				  req->dump_format);
		break;
	}

	return result;
}


/*
 * Public functions
 */

/**
 * cr_dump_self - dump 'current'
 * @filp: The file on which the request was received.
 * @flag: Flags to modify behavior.
 *
 * DESCRIPTION:
 * Dumps the current process to a file descriptor supplied by the
 * requester.
 *
 * This is the function that the handlers will call into when
 * the CHECKPOINT portion of callbacks are finished.
 *
 * Currently the 'flags' argument is unused.
 */
int cr_dump_self(struct file *filp, unsigned long flags)
{
	cr_chkpt_proc_req_t *proc_req;
	struct file *dest_filp;
	cr_chkpt_req_t *req;
	cr_task_t *cr_task;
	int shared;
	int result = 0;

	CR_KTRACE_FUNC_ENTRY("flags=0x%lx", flags);

	// Lookup this task
	cr_task = cr_task_get(current);
	if (!cr_task) {
		// No matching task found.
		return -ESRCH;
	}

	// Find the matching request
	if (!(req = cr_task->chkpt_req) || !(proc_req = cr_task->chkpt_proc_req)) {
		// No matching request found.
		cr_task_put(cr_task);
		return -ESRCH;
	}

	// Ensure files save can start as soon as all tasks reach kernel space
	cr_barrier_notify(&req->prefiles_barrier);

	// XXX: Once we define the options structure, this is the place
	// to merge the per-process options (attached to 'filp'?) and the
	// options passed by the requester.

        // Find the destination
	dest_filp = cr_dest_get(&req->dest, &shared);
	if (IS_ERR(dest_filp)) {
		return PTR_ERR(dest_filp);
	}

	// Ensure that exactly one caller writes the file header
	down(&req->serial_mutex);
        if (!test_and_set_bit(0, &req->done_header)) {
            result = cr_save_file_header(req, dest_filp);
            if (result < 0) {
		req->result = result;
            }
        }
	if (req->result) {
	    result = req->result;
	    up(&req->serial_mutex);
	    goto cleanup;
	}
	up(&req->serial_mutex);

	// If we have a Phase1 then start Phase2 when complete.
	if (cr_task->phase == CR_PHASE1) {
		int rc = cr_barrier_enter(&proc_req->phase_barrier);
		// All Phase1 threads are done, so one thread triggers the Phase2 threads.
		if (rc) {
			cr_trigger_phase2(req);
		}
	}

	/* Check to see if error/abort has occurred */
	read_lock(&req->lock);
	if ( (result = req->result)) {
		read_unlock(&req->lock);
		cr_barrier_notify(&proc_req->predump_barrier);
		goto cleanup;
	}
	read_unlock(&req->lock);

	// Synchronize to ensure all tasks have stopped running.
	result = cr_barrier_enter(&proc_req->predump_barrier);

	// The handlers have all responded, so we should not "expire"
	// So "first" task out of the barrier zeros req->has_expiration.
	if (result) {
		write_lock(&req->lock);
		req->has_expiration = 0;
		write_unlock(&req->lock);
	}

	/* check again for abort/error */
	read_lock(&req->lock);
	if ( (result = req->result)) {
		read_unlock(&req->lock);
		goto cleanup;
	}
	read_unlock(&req->lock);

	// Send the dump to the proper destination filp.
	result = cr_do_dump(req, proc_req, dest_filp, shared);

	// If we experienced an error for the first time then
	// save it in the request so the requester will learn of it.
	if (result && !req->result) {
		req->result = result;
	}

	// Synchronize to ensure all tasks have saved before we continue
	cr_barrier_enter(&req->postdump_barrier);

	// Ensure that exactly one caller writes the file trailer
	down(&req->serial_mutex);
        if (!test_and_set_bit(0, &req->done_trailer)) {
            CR_KTRACE_LOW_LVL("Writing the trailer.");
            result = cr_save_header(req, dest_filp, 0);
            if (result < 0) {
		req->result = result;
            }
        }
	if (req->result) {
	    result = req->result;
	    up(&req->serial_mutex);
	    goto cleanup;
	}
	result = 0;
	up(&req->serial_mutex);

	// Here is where we raise(SIG_STOP) etc
	if (req->signal) {
		send_sig_info(req->signal, NULL, current);
	}

cleanup:
	cr_dest_put(&req->dest, dest_filp);
	if (req->die) {
		send_sig_info(SIGKILL, NULL, current);
		cr_chkpt_task_complete(cr_task);
	} else if (cr_task->phase == CR_NO_PHASE) {
		// task has no handlers, so finish up now.
		cr_chkpt_task_complete(cr_task);
	} else {
		// cr_chkpt_task_complete() will run from the HAND_CONT ioctl.
	}

	// Release the task, balancing the cr_task_get() above.
	cr_task_put(cr_task);

	return result;
}

/**
 * cr_cont_complete
 * @filp: The file on which the ioctl was received.
 *
 * DESCRIPTION:
 * Implements the HAND_CONT ioctl() which indicates to the kernel that
 * the calling task has completed running its CONTINUE handlers.  Thus
 * we can prevent checkpoint requests from overlapping these handlers.
 *
 * This is the function that the handlers will call into when
 * the CONTINUE portion of callbacks are finished.
 *
 */
int cr_cont_complete(struct file *filp)
{
	cr_task_t *cr_task;

	CR_KTRACE_FUNC_ENTRY();

	// Lookup this task
	cr_task = cr_task_get(current);
	if (!cr_task) {
		// No matching task found.
		return -ESRCH;
	}

	// Find the matching request
	if(!cr_task->chkpt_req) {
		// No matching request found.
		cr_task_put(cr_task);
		return -ESRCH;
	}

	// Release the request.
	cr_chkpt_task_complete(cr_task);

	// Release the task.
	cr_task_put(cr_task);

	return 0;
}
