/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_rstrt_req.c,v 1.164 2007/02/27 20:22:49 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/errno.h>
#include <linux/pipe_fs_i.h>
#include <linux/time.h>
#include <linux/mman.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>

#include "cr_context.h"
#include "vmadump.h"

#ifndef PID_MAX
   /* XXX: Is this always true? */
  #define PID_MAX 0x8000
#endif

/*
 * alloc_rstrt_request
 */
static cr_rstrt_req_t *
alloc_rstrt_req(void)
{
    cr_rstrt_req_t *req = NULL;
    cr_objectmap_t map;

    map = cr_alloc_objectmap();
    if (!map) {
	goto out;
    }

    req = kmalloc(sizeof(*req), GFP_ATOMIC);
    if (req) {
        CR_MODULE_GET();
	CR_KTRACE_REFCNT("Alloc cr_rstrt_req_t %p", req);
        memset(req, 0, sizeof(*req));
        atomic_set(&req->ref_count, 1);
	get_task_struct(current);
	req->cr_restart_task = current;
	req->state = CR_RSTRT_STATE_REQUESTER;
	set_bit(0, &req->need_children);
	init_waitqueue_head(&req->wait);
	cr_barrier_init(&req->barrier, 1);
	req->result = 0;
	req->die = 0;
	req->map = map;
	req->procs_lock = RW_LOCK_UNLOCKED;
	INIT_LIST_HEAD(&req->procs);
    } else {
	goto out_freemap;
    }

out:
    return req;

out_freemap:
    cr_release_objectmap(map);
    return req;
}

static void
release_rstrt_req(cr_rstrt_req_t *req)
{
    CR_KTRACE_REFCNT("ref count is approximately %d", 
		     atomic_read(&req->ref_count));

    if (atomic_dec_and_test(&req->ref_count)) {
	struct list_head *elem, *next;
	list_for_each_safe(elem, next, &req->procs) {
	    cr_rstrt_proc_req_t *tmp = list_entry(elem, cr_rstrt_proc_req_t, list);
            if (tmp->linkage) {
	        kfree(tmp->linkage);
	    }
	    kfree(tmp);
	}
        CR_KTRACE_REFCNT("fput()'ing file");
	fput(req->file0);
	put_task_struct(req->cr_restart_task);
	CR_KTRACE_ALLOC("kfree()'ing req");
	cr_release_objectmap(req->map);
        kfree(req);
        CR_MODULE_PUT();
	CR_KTRACE_REFCNT("Free cr_rstrt_req_t %p", req);
    } else {
	clear_bit(0, &req->need_children);
	wake_up(&req->wait);
    }
}

/*
 * source get typed.
 *
 * Doesn't quite get concurrency.
 */
static struct file *
__cr_src_gett(cr_rstrt_req_t *req, cr_obj_t object)
{
    return req->file0;
}

/*
 * cr_src_gett
 *
 * source get typed.  Knows what we're planning to read, and understands
 * how to serialize when necessary.
 */
static struct file *
cr_src_gett(cr_rstrt_req_t *req, cr_obj_t object)
{
    return __cr_src_gett(req, object);
}

/*
 * __cr_src_putt
 *
 * stands for src put typed.
 */
static int
__cr_src_putt(struct file *filp)
{
    return 0;
}

/*
 * cr_src_putt
 *
 * stands for src put typed.
 */
static int
cr_src_putt(struct file *filp)
{
    return __cr_src_putt(filp);
}

static int cr_restore_creds(cr_rstrt_proc_req_t *proc_req)
{
    struct cr_context_creds cf_creds;
    int retval;

    retval = cr_kread(proc_req->file, &cf_creds, sizeof(cf_creds));
    if (retval < sizeof(cf_creds)) {
	retval = -EIO;
        goto out;
    }
     
#if CR_RESTORE_IDS
    {
	int i, j;
	if (cf_creds.ngroups > NGROUPS) {
	    CR_KTRACE_LOW_LVL("%s: Invalid supplemental group count (%d)",
			      __FUNCTION__, (int)cf_creds.groups);
	    retval = -EINVAL;
	    goto out;
	}
	/* Search for any "expansion" of the group set */
	for (i = 0; i < cf_creds.groups; ++i) {
	    gid_t g = cf_creds.groups[i];
	    int gid_ok = 1;
	    for (j = 0; j < current->ngroups; ++j) {
		if (g == current->groups[j]) {
		    gid_ok = 0;
		    break;
		}
	    }
	    if (!gid_ok) {
	        /* If we reach here then we've seen a supplemental group in the
	         * saved context, which is not present in the current list.
		 * The set_groups() call checks permissions for us.  If we fail,
		 * then we must not have enough credentials.
		 */
		mm_segment_t oldfs = get_fs();
		set_fs(KERNEL_DS);
		retval = sys_setgroups(cf_creds.ngroups, cf_creds.groups);
		set_fs(oldfs);
		if (retval < 0) {
		    CR_KTRACE_LOW_LVL("%s: Failed to restore supplemental groups", __FUNCTION__);
		    goto out;
		}
		CR_KTRACE_LOW_LVL("%s: Restored supplemental groups", __FUNCTION__);
		break; /* no need to continue the i loop */
	    }
	}

	/* The set_setresgid() call checks permissions for us, always OK if no change . */
	retval = sys_setresgid(cf_creds.gid, cf_creds.egid, cf_creds.sgid);
	if (retval < 0) {
	    CR_KTRACE_LOW_LVL("%s: Failed to restore gids", __FUNCTION__);
	    goto out;
	}
	CR_KTRACE_LOW_LVL("%s: Restored gids", __FUNCTION__);

	/* The set_setresuid() call checks permissions for us, always OK if no change . */
	retval = sys_setresuid(cf_creds.uid, cf_creds.euid, cf_creds.suid);
	if (retval < 0) {
	    CR_KTRACE_LOW_LVL("%s: Failed to restore uids", __FUNCTION__);
	    goto out;
	}
	CR_KTRACE_LOW_LVL("%s: Restored uids", __FUNCTION__);
    }
#endif	 /* CR_RESTORE_IDS */

out:
    return retval;
}

/*
 * cr_load_linkage
 *
 * Loads the linkage structure stored in the context file into a new
 * buffer, attached to the proc_req, returns linkage count (or error)
 *
 * This linkage structure is used by cr_restore_{pids,parents}().
 *
 * This grabs and releases tasklist_lock.
 *
 * NOTE:  Can use this later to reconstruct the graph of processes,
 * to restore the session and thread counts
 */
static int cr_load_linkage(cr_rstrt_req_t *req, cr_rstrt_proc_req_t *proc_req)
{
    struct file *filp = proc_req->file;
    struct cr_context_tasklinkage *linkage;
    int linkage_bytes;                  /* size of linkage stored on disk */
    int linkage_size;                   /* number of entries in array */
    int retval;
    int i;

    /* find out how large the linkage structure is */
    retval = cr_kread(filp, &linkage_bytes, sizeof(linkage_bytes));
    if (retval < sizeof(linkage_bytes)) {
        if (retval >= 0) retval = -EIO;
        goto out;
    }

    /* allocate memory for the linkage structure */
    linkage = NULL;
    linkage_size = 0;
    if (linkage_bytes < 131072) {
        linkage = kmalloc(linkage_bytes, GFP_KERNEL);
	linkage_size = linkage_bytes/sizeof(*linkage);
    }
    if (linkage == NULL || linkage_size == 0) {
        CR_WARN("Too many processes (size = %d)", linkage_bytes);
        retval = -ENOMEM;
	goto out;
    }

    /* Now load the linkage from disk */
    retval = cr_kread(filp, linkage, linkage_bytes);
    if (retval < linkage_bytes) {
        if (retval >= 0) retval = -EIO;
        kfree(linkage);
        goto out;
    }

    /* 
     * Initialize all tl_task's to NULL
     *
     * Note:  Do not mess skip this.  We use it to check for process aborts in 
     * cr_restore_parents.
     */
    for (i=0; i<linkage_size; ++i) {
        linkage[i].tl_task = NULL;
    }

    /* return the linkage structure and size */
    proc_req->linkage = linkage;
    proc_req->linkage_size = linkage_size;

    retval = linkage_size;

out:
    return retval;
}

/*
 * cr_linkage_find_real_parent
 *
 * Figure out who your parent process ought to be.
 *
 * As a "Safety net" we default to becoming a child of the
 * cr_restart process if the original parent is not found.
 */
struct task_struct *
cr_linkage_find_real_parent(cr_rstrt_req_t *req, struct task_struct *old_parent)
{
    struct task_struct *parent = NULL;

    if (!cr_find_object(req->map, old_parent, (void **) &parent)) {
        /* No entry for this process, we'll reattatch it as a new child
         * of cr_restart. */
        parent = req->cr_restart_task;
    }

    return parent;
}

#if HAVE_2_6_TASK_LINKAGE
/* Code based heavily on kernel/pid.c */
#define BITS_PER_PAGE (8*PAGE_SIZE)
#define BITS_PER_PAGE_MASK (BITS_PER_PAGE-1)
#if defined(CR_KDATA_pidmap_array)
  #define cr_pidmap(nr) &(pidmap_array[(nr) / BITS_PER_PAGE])
#elif defined(CR_KDATA_init_pspace)
  #define cr_pidmap(nr) &(init_pspace.pidmap[(nr) / BITS_PER_PAGE])
#endif
static inline void
_cr_free_pid(int nr)
{
    struct pidmap *map = cr_pidmap(nr);
    int offset = nr & BITS_PER_PAGE_MASK;

    clear_bit(offset, map->page);
    atomic_inc(&map->nr_free);
    #if HAVE_ALLOC_PID
      put_pid(find_pid(nr));
    #endif
}
static int
_cr_alloc_pid(int nr) /* MUST call with no locks held */
{
    struct pidmap *map = cr_pidmap(nr);
    int offset = (nr & BITS_PER_PAGE_MASK);
    int result = 0;	/* assume failure */

    /* make sure we have a page */
    if (unlikely(map->page == NULL)) {
	unsigned long page = get_zeroed_page(GFP_KERNEL);

	spin_lock(&pidmap_lock);
	if (map->page) {
	    /* lost a race, free our page */
	    free_page(page);
	} else {
	    map->page = (void *)page;
	}
	spin_unlock(&pidmap_lock);
    }
    /* XXX assert (map->page != NULL) */

    if (!test_and_set_bit(offset, map->page)) {
    #if HAVE_ALLOC_PID
	if (!find_pid(nr)) {
            #define pid_hashfn(nr) hash_long((unsigned long)nr, pidhash_shift)
	    struct pid *pid = kmem_cache_alloc(pid_cachep, GFP_KERNEL);
	    enum pid_type type;

	    if (!pid) {
	        clear_bit(offset, map->page);
	        goto out;
	    }
	    atomic_set(&pid->count, 1);
	    pid->nr = nr;
	    for (type = 0; type < PIDTYPE_MAX; ++type) {
	        INIT_HLIST_HEAD(&pid->tasks[type]);
	    }

	    spin_lock_irq(&pidmap_lock);
	    hlist_add_head_rcu(&pid->pid_chain, &pid_hash[pid_hashfn(nr)]);
	    spin_unlock_irq(&pidmap_lock);
	}
    #endif
	atomic_dec(&map->nr_free);
	result = 1;
    }

#if HAVE_ALLOC_PID
out:
#endif
    return result;
}

/* Reserve all the pids we need
 * Returns non-zero on failure
 */
static int
cr_reserve_ids(struct cr_context_tasklinkage *linkage, int linkage_size)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    CR_KTRACE_HIGH_LVL("Now reserving required ids... ");

    /* Don't need to worry about session IDs, since exactly one member
     * of the session, the leader, will have a matching PID.  (When that
     * leader exits the entire session dies).
     */

#if 0 /* Not yet */
    /* _DO_ need to think about PGIDs.  Since it is possible for the process
     * which originally had pid==pgid to exit w/o terminating the process group,
     * we must call _cr_alloc_pid() for any such ids.  Easy way is to do this
     * for every pgid we restore and filter the pid==pgid cases in the pid loop.
     */
    foreach_unique_pgid_req {
        if (!_cr_alloc_pid(pgid)) {
            goto out_release_pgids;
        }
    }
#endif
    
    for (i = 0; i < linkage_size; ++i) {
	link = &linkage[i];

#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we already grabbed this id in the pgid loop above */
	    continue;
	}
#endif

	if (!_cr_alloc_pid(link->pid)) {
#if CR_KERNEL_TRACING
	    read_lock(&tasklist_lock);
	    printk("Found pid %d in use: pid=%d, tgid=%d, pgid=%d, sid=%d\n",
		   link->pid,
		   cr_have_pid(PIDTYPE_PID, link->pid),
#if HAVE_PIDTYPE_TGID
		   cr_have_pid(PIDTYPE_TGID, link->pid),
#else
		   0,
#endif
		   cr_have_pid(PIDTYPE_PGID, link->pid),
		   cr_have_pid(PIDTYPE_SID, link->pid));
	    read_unlock(&tasklist_lock);
#endif
	    goto out_release_pids;
	}
    }

    return 0;	/* success */

out_release_pids:
    /* If we failed w/ i=N, then 0<=j<N were allocated */
    while (i>0) {
	--i;
	_cr_free_pid(linkage[i].pid);
    }
#if 0	/* Not yet */
out_release_pgids:
    XXX release whichever pgids we had obtained before failure.
#endif

    return 1;	/* failure */
}

static void
cr_release_ids(struct cr_context_tasklinkage *linkage, int linkage_size)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    for (i = 0; i < linkage_size; ++i) {
	link = &linkage[i];
#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we will free this id in the pgid loop below */
	    continue;
	}
#endif
	_cr_free_pid(link->pid);
    }

#if 0	/* Not yet */
    foreach_unique_pgid_req {
	_cr_free_pid(pgid);
    }
#endif
}
#endif /* HAVE_2_6_TASK_LINKAGE */

/*
 * cr_is_linkage_restorable
 *
 * Scans through a linkage and checks to see if any of the PIDs, etc.
 * that are described here are in use.  (i.e. check for PID conflict)
 * Also check for missing (NULL) linkage fields.
 *
 * NOTE:  Called from request_restart _AND_ from cr_do_restore_pids!
 * As such, this thing _might_ be called with a spinlock held, so be
 * careful with what you go around doing in here, as you may get in trouble.
 *
 * NOTE: cr_is_linkage_restorable caller must hold tasklist_lock
 */
static int
cr_is_linkage_restorable(cr_rstrt_req_t *req,
			 struct cr_context_tasklinkage *linkage,
			 int linkage_size)
{
    int retval;
    int i;
    struct cr_context_tasklinkage *l;
#if HAVE_2_4_TASK_LINKAGE
    struct task_struct *p;
#endif

    retval = 0;

    /* Try to avoid a possible race against req->cr_restart_task exiting
     * before we get to restore the linkage */
    if (cri_task_dead(req->cr_restart_task)) {
        CR_ERR("%s [%d]:  Unrestorable linkage: cr_restart_task exited "
		"prematurely!", __FUNCTION__, current->pid);
        return 0;
    }

    for(i=0; i<linkage_size; ++i) {
	l = &linkage[i];
#if HAVE_2_4_TASK_LINKAGE
        if (l->pid >= PID_MAX) {
	    CR_WARN("PID out of bounds ... i=%d pid=%d", i, l->pid);
            return 0;
	}
#else
	// XXX: should we validate some equivalent of PID_MAX ?
#endif

#if HAVE_2_4_TASK_LINKAGE
        for_each_task(p) {
            if (l->pid == p->pid || l->tgid == p->tgid) {
		/* XXX: sessions not supported yet...
	        l->pgrp == p->pgrp || 
		l->sesion == p->session
		*/
	        CR_WARN("PID conflict ... i=%d pid=%d", i, l->pid);
                return 0;
	    }
	}
#else
	if (cr_have_pid(PIDTYPE_PID, l->pid)
#if HAVE_PIDTYPE_TGID
	    || cr_have_pid(PIDTYPE_TGID, l->tgid)
#endif
	    /* XXX: sessions not supported yet...
	    || cr_have_pid(PIDTYPE_PGID, l->pgrp)
	    || cr_have_pid(PIDTYPE_SID, l->session) */
	   ) {
	    CR_WARN("PID conflict ... i=%d pid=%d", i, l->pid);
            return 0;
	}
#endif
    }

    return 1;
}

/* 
 * cr_restore_parents
 *
 * NOTE:  Must only be called by a single thread.  This function acts on
 * every task contained in the restart request.  All others must be blocked
 * to prevent return to user space prior to entry into this function.
 *
 * Acquires and releases tasklist_lock held for writing.
 */
static int
cr_restore_parents(cr_rstrt_req_t *req)
{
    struct list_head *elem;
    int retval = 0;
    int i;

#if !HAVE_REMOVE_LINKS
    #define REMOVE_LINKS remove_parent
    #define SET_LINKS add_parent
#endif

    CR_KTRACE_HIGH_LVL("Now restoring the parent linkage... ");

    /* thread_group is already correct by virtue of how we create threads */

    /* Add linkage entry for init */
    cr_insert_object(req->map, CR_PARENT_INIT, child_reaper);

    read_lock(&req->procs_lock);

    /*
     * Make a pass through to translate old task pointers to new ones
     */
    list_for_each(elem, &req->procs) {
        cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
        struct cr_context_tasklinkage *linkage = proc_req->linkage;
        int linkage_size = proc_req->linkage_size;

        for (i=0; i<linkage_size; ++i) {
            struct cr_context_tasklinkage *link = &linkage[i];

	    if (!cr_find_object(req->map, link->myptr, (void **) &link->tl_task)) {
	        /* No entry for this process! */
	        CR_ERR("%s [%d]:  ERROR:  No task found for pid %d %p",
		         __FUNCTION__, current->pid, link->pid, link);
	        retval = -EINVAL;
	        goto out;
	    }

#ifdef CR_REAL_PARENT
	    link->real_parent = cr_linkage_find_real_parent(req, link->real_parent);
            link->parent = link->real_parent; /* XXX: until/unless we support ptrace */
#else
	    link->parent = cr_linkage_find_real_parent(req, link->parent);
#endif
	    if (!link->parent) {
	        /* No parent found for this process! */
	        CR_ERR("%s [%d]:  ERROR:  No parent found for pid %d",
		         __FUNCTION__, current->pid, link->pid);
	        retval = -EINVAL;
	        goto out;
	    }
        }
    }

    /* OK, now that the pointers have been calculated, we can go through
     * and actually perform the swap */
    write_lock_irq(&tasklist_lock);
    list_for_each(elem, &req->procs) {
        cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
        struct cr_context_tasklinkage *linkage = proc_req->linkage;
        int linkage_size = proc_req->linkage_size;

        for (i=0; i<linkage_size; ++i) {
            struct cr_context_tasklinkage *link = &linkage[i];
	    struct task_struct *task = link->tl_task;

	    REMOVE_LINKS(task);
            CR_PARENT(task)      = link->parent;
#ifdef CR_REAL_PARENT
	    CR_REAL_PARENT(task) = link->real_parent;
#endif
	    SET_LINKS(task);
        }
    }
    write_unlock_irq(&tasklist_lock);

    /* on success retval is zero when we get here */

out:
    read_unlock(&req->procs_lock);
    return retval;
}


// cr_rstrt_req(user_req)
//
// Processes a restart request received from user space.
//
// Returns 0 on sucess
// Returns negative on failure
// XXX: Return number of children to start when doing process groups?
int cr_rstrt_request_restart(struct file *filp, void *arg)
{
    struct cr_context_file_header cf_header;
    cr_pdata_t *priv;
    struct cr_rstrt_args args;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    /* Make sure that we don't have any requests already on this fd */
    priv = filp->private_data;
    if (priv->rstrt_req || priv->chkpt_req) {
        return -EAGAIN;
    }

    /* XXX:  We don't handle a cr_restart that has children
     * yet, so make sure that you don't have any before continuing */
    if (CR_CHILD(current) != NULL) {
        CR_KTRACE_UNEXPECTED("Aborting.  Process has children.");
        return -EAGAIN;
    }

    /* Allocate a request structure for bookkeeping */
    req = alloc_rstrt_req();
    if (req == NULL) {
        return -ENOMEM;
    }
    priv->rstrt_req = req;
    
    /* Copy in the user's arguments */
    if (copy_from_user(&args, arg, sizeof(args))) {
	retval = -EFAULT;
        goto out_free_req;
    }

    // XXX:  Copy in user request options

    req->signal = args.signal;
    
    /* Check the file descriptor */
    // XXX:  Needs to be replaced with cr_rstrt_src_get()...
    req->file0 = fget(args.cr_fd);
    if (!(req->file0)) {
	retval = -EBADF;
        goto out_free_req;
    }

    /* Read in the context file header */
    retval = cr_kread(req->file0, &cf_header, sizeof(cf_header));
    if (retval != sizeof(cf_header)) goto out_free_req;

    CR_KTRACE_HIGH_LVL("cr_magic = %d %d, " 
		       "cr_version = %d, "
		       "scope = %d,",
		       cf_header.magic[0], cf_header.magic[1],
		       cf_header.version, cf_header.scope);

    // Check magic and cr_version
    {
        static char the_magic[2] = CR_MAGIC;
        if ((the_magic[0] != cf_header.magic[0]) ||
            (the_magic[1] != cf_header.magic[1]) ||
            (cf_header.version != CR_CONTEXT_VERSION)) {
            retval = -EBADRQC;
	    goto out_free_req;
        }
    }

    req->scope = cf_header.scope;  // Currently unused
    
    /* Normal case.  We keep the req, since we will need it again */
    return 0;

out_free_req:
    release_rstrt_req(req);
    priv->rstrt_req = NULL;

    return retval;
}

// cr_rstrt_procs()
//
// Called by requester to get "next" process
//
// Returns > 0 if 1 or more processes remain, 0 if none.
// Returns negative on failure
int cr_rstrt_procs(struct file *filp, struct cr_procs_tbl *tbl_ptr)
{
    struct cr_section_header cf_header;
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    int retval, threads;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out_noreq;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out_noreq;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
	CR_ERR("%s: No rstrt_req attached to filp!", __FUNCTION__);
        goto out_noreq;
    }

    while (!req->die && (req->state == CR_RSTRT_STATE_CHILD) && !signal_pending(current)) {
	/* Wait for child to finish w/ the file */
	unsigned long timeout = MAX_SCHEDULE_TIMEOUT;
	timeout = wait_event_interruptible_timeout(req->wait, (req->die || (req->state != CR_RSTRT_STATE_CHILD)), timeout);
    }
    if (signal_pending(current)) {
	retval = -EINTR;
	goto out;
    } else if (req->die || (req->state == CR_RSTRT_STATE_DONE)) {
	retval = 0;
	goto out;
    }

    /* Read in the context header */
    retval = cr_kread(req->file0, &cf_header, sizeof(cf_header));
    if (retval != sizeof(cf_header)) {
        if (retval > 0) { retval = -EIO; }
	goto out;
    }

    threads = cf_header.num_threads;

    if (threads == 0  /* EOF */) {
	// XXX: additional serialization will be needed when restores are parallel
        retval = cr_restore_parents(req);
        if (retval) {
	    goto out;
        }

        /* XXX A hack to return a "lead" pid.
         * Currently we look for the first task that is a child of cr_restart.
         * The assumption was that there
         * is a single "lead" to waitpid() for, which might not be the case.
         * Looking for session or process group leaders would be good in the
         * case of multiple children of cr-restart.
         */
        if (!req->result) { /* don't clobber an error indication */
	    struct list_head *elem;
	    list_for_each(elem, &req->procs) {
        	cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
        	struct cr_context_tasklinkage *linkage = proc_req->linkage;
        	int i, linkage_size = proc_req->linkage_size;

                for (i=0; i<linkage_size; ++i) {
                    if (linkage[i].parent == req->cr_restart_task) {
		        req->result = linkage[i].pid;
	                break;
	            }
                }
	    }
        }

	retval = 0;
    } else {
        cr_rstrt_proc_req_t *proc_req;

	/* Copy the process-header info to user space */
	retval = -EFAULT;
	if (put_user(threads, &tbl_ptr->threads) ||
	    put_user(cf_header.clone_flags, &tbl_ptr->clone_flags)) {
	    goto out;
        }

        set_bit(0, &req->need_children); // XXX: Better way?

	proc_req = kmalloc(sizeof(*proc_req), GFP_KERNEL);
	retval = -ENOMEM;
	if (!proc_req) {
	    goto out;
	}

        memset(proc_req, 0, sizeof(*proc_req));

        init_MUTEX(&proc_req->serial_mutex);

        /* Set number of threads for this one. */
        proc_req->total_threads = threads;

        /* Now initialize the barrier count so we can block later */
        /* XXX:  This may not be correct.  We're assuming here that all the
         * threads that we create will actually call into the barrier code.
         * While this is true in all sane cases, it will not be true if one
         * of the threads dies prematurely...  pthreads has a habit of doing
         * this.  This should be fixed to use barrier_notify when needed. */
        atomic_set(&proc_req->final_counter, threads);
        atomic_set(&proc_req->linkage_counter, threads);
        cr_barrier_init(&proc_req->linkage_barrier, threads);

	proc_req->req = req;
    	proc_req->file = req->file0;	// XXX: KLUDGE
	proc_req->clone_flags = cf_header.clone_flags;

	write_lock(&req->procs_lock);
	list_add(&proc_req->list, &req->procs);
	write_unlock(&req->procs_lock);

        /* Normal case.  */
        req->state = CR_RSTRT_STATE_CHILD;
        retval = 1;
    }

out:
    if (retval != 1) {
	cr_barrier_notify(&req->barrier);
    }
out_noreq:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

// Check that a request is "done", from the point of view of the
// requestor.
// Returns >0 is completed, 0 if not completed, <0 on error
static
int check_done(cr_rstrt_req_t *req)
{
	/* Check that it is just me here,
	 * and that one or more children either started or died */
	return ((atomic_read(&req->ref_count) == 1) &&
		!test_bit(0, &req->need_children))	? 1 : 0;
}
	
// cr_rstrt_done(timeval)
//
// Test/wait for a restart request to complete (tasks are alive).
// Argument is:
//    Pointer to a (struct timeval) for a bounded wait
//    	if zero then we take care not to schedule().
//    NULL for an unbounded wait
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	>0 if complete
//	0 if not complete within the timeout (if any)
//
// XXX: *should* return PID of "root" process (session leader)
int cr_rstrt_done(struct file *filp, void *arg)
{
    cr_pdata_t		*priv;
    cr_rstrt_req_t	*req;
    struct timeval	timeval;
	
    unsigned long timeout;
    int retval = -EINVAL;

    CR_KTRACE_FUNC_ENTRY();

    priv = filp->private_data;
    if (!priv) {
	goto out;
    }
    req = priv->rstrt_req;
    if (!req) {
	goto out;
    }

    // Fetch/decode the optional timeout from userspace
    timeout = MAX_SCHEDULE_TIMEOUT;
    if (arg) {
	retval = copy_from_user(&timeval, arg, sizeof(timeval));
	if (retval) {
	    goto out;
	}

	timeout = timeval_to_jiffies(&timeval);
    } 

    timeout = wait_event_interruptible_timeout(req->wait, ((retval = check_done(req)) || req->die) , timeout);

    // Copy remaining time back to userspace if applicable
    if (arg) {
        struct timeval *tvp = (struct timeval *)arg;
	jiffies_to_timeval(timeout, &timeval);
        put_user(timeval.tv_sec, &tvp->tv_sec);
        put_user(timeval.tv_usec, &tvp->tv_usec);
	// like select() we fail silently if user timeval is R/O
    }

    // Check if we were interrupted
    if (!(retval || req->die) && timeout) {
	retval = -EINTR;
    }

out:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

// cr_rstrt_reap()
//
// Reap a completed restart.
// Returns:
// 	req->result on success
// 	-EINVAL if no un-reaped restart request is associated w/ this fd
// 	-EAGAIN if the request is not completed
int
cr_rstrt_reap(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    priv = filp->private_data;
    if (priv) {
	req = priv->rstrt_req;
	if (req) {
	    retval = -EAGAIN;
	    if (check_done(req)) {
		priv->rstrt_req = NULL;
		retval = req->result;
		(void) release_rstrt_req(req);
	    }
	}
    }

    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

/*
 * cr_linkage_find_by_old_pid
 *
 * Returns a pointer to a linkage entry with l.pid == pid
 * NULL if not found
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_old_pid(struct cr_context_tasklinkage *linkage,
			   int linkage_size,
			   int pid)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    for (i=0; i<linkage_size && entry == NULL; ++i) {
        if (linkage[i].pid == pid) {
	    entry = &(linkage[i]);
	}
    }

    return entry;
}

/*
 * Add task to all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_hash(struct task_struct *task)
{
#if HAVE_2_4_TASK_LINKAGE
    hash_pid(task);
#else
    int is_leader = thread_group_leader(task);
    if (is_leader) {
	attach_pid(task, PIDTYPE_SID, cr_task_session(task));
	attach_pid(task, PIDTYPE_PGID, cr_task_pgrp(task));
    }
  #if HAVE_PIDTYPE_TGID && HAVE_TASK_PIDS_PID
    attach_pid(task, PIDTYPE_TGID, task->tgid);
  #elif HAVE_PIDTYPE_TGID && HAVE_TASK_PIDS_PID_LINK
    if (is_leader) {
        attach_pid(task, PIDTYPE_TGID, task->tgid);
    } else {
        link_pid(task, task->pids + PIDTYPE_TGID,
		 &task->group_leader->pids[PIDTYPE_TGID].pid);
    }
  #endif
    attach_pid(task, PIDTYPE_PID, task->pid);
#endif
}

/*
 * Remove task from all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_unhash(struct task_struct *task)
{
#if HAVE_2_4_TASK_LINKAGE
    unhash_pid(task);
#else
    int is_leader = thread_group_leader(task);
    detach_pid(task, PIDTYPE_PID);
  #if HAVE_PIDTYPE_TGID
    detach_pid(task, PIDTYPE_TGID);
  #endif
    if (is_leader) {
	detach_pid(task, PIDTYPE_PGID);
	detach_pid(task, PIDTYPE_SID);
    }
#endif
}

/* 
 * cr_do_restore_pids
 *
 * NOTE:  Must only be called by a single thread.  This function acts on
 * every process contained in the restart request.  All other threads
 * should be blocked and prevented from returning to user space before entry
 * into this function.
 *
 * Acquires and releases tasklist_lock held for writing.
 */
static int
cr_do_restore_pids(cr_rstrt_req_t *req,
		   struct cr_context_tasklinkage *linkage,
		   int linkage_size)
{
    int retval = 0;
    int i;
    struct cr_context_tasklinkage *link;

    /*
     * Make a pass through for a sanity check
     */
    for (i=0; i<linkage_size; ++i) {
        link = &linkage[i];
	if (link->tl_task == NULL) {
            /* Means that we have an entry for which there is no
	     * corresponding task! */

	    /* XXX:  This can actually happen in normal cases if a thread
	     * exits...  Ignore for now...  Can probably pull this code
	     * out later. */
	    CR_ERR("%s [%d]:  ERROR:  No task for this PID. %d",
		     __FUNCTION__, current->pid, link->pid);
	    retval = -EINVAL;
	    goto out_nolock;
	}
    }


    /* Now grab tasklist_lock while we futz with everyones pointers */
    write_lock_irq(&tasklist_lock);

    /* Unlike the last pass, this pass is decisive.  If it doesn't work here,
     * then we _have_ to abort!  (or skip) */
    if (!cr_is_linkage_restorable(req, linkage, linkage_size)) {
        retval = -EBUSY;
        goto out;	
    }

    /* Pass 1: unhash all the pids */
    for (i=0; i<linkage_size; ++i) {
	cr_do_unhash(linkage[i].tl_task);
    }

    /* Pass 2: change all the simple links and ids */
    for (i=0; i<linkage_size; ++i) {
	struct task_struct *task;

        link = &linkage[i];
	task = link->tl_task;

	task->tgid = link->tgid;
    	task->pid = link->pid;
	/* XXX: NOT YET
	task->pgrp = link->pgrp;
	task->session = link->session;
	*/

	task->exit_signal = link->exit_signal;
	/* XXX:  We don't handle these correctly yet */
#if 0
	task->tty_old_pgrp = link->tty_old_pgrp;
	task->leader = leader;
#endif
    }


#if HAVE_2_4_TASK_LINKAGE
    /* Pass 3: rehash the pids */
    for (i=0; i<linkage_size; ++i) {
	cr_do_hash(linkage[i].tl_task);
    }
#elif HAVE_2_6_TASK_LINKAGE
    /* Keeping the pid hashes/lists correct requires multiple passes for some
     * RH9-type kernels, but we do it this way for them all 2.6 linkage cases
     */

    /* Pass 3a: relink/rehash only thread group leaders
     * Sometimes these must exist before their followers can be linked in.
     */
    for (i=0; i<linkage_size; ++i) {
	struct task_struct *task = linkage[i].tl_task;

        if (thread_group_leader(task)) {
		cr_do_hash(task);
		/* To look like we called exec(): */
		task->parent_exec_id = req->cr_restart_task->self_exec_id;
		task->self_exec_id = req->cr_restart_task->self_exec_id + 1;
	}
    }

    /* Pass 3b: relink/rehash non-leaders */
    for (i=0; i<linkage_size; ++i) {
	struct task_struct *task = linkage[i].tl_task;

        if (!thread_group_leader(task)) {
		cr_do_hash(task);
		/* To look like we were fork()ed by leader, who called exec(): */
		task->parent_exec_id = req->cr_restart_task->self_exec_id + 1;
		task->self_exec_id = req->cr_restart_task->self_exec_id + 1;
	}
    }
#endif

    /* on success retval is zero when we get here */

out:
    /* Let the rest of the system come back */
    write_unlock_irq(&tasklist_lock);
out_nolock:
    return retval;
}

/*
 * cr_restore_pids
 *
 * This function implements a barrier, and hence must be called by all
 * restarting tasks before their return to user space.  Once the barrier
 * condition is reached (tasks entered == tasks created) the last thread
 * to enter restores the linkage.  At that point, the other tasks are
 * woken, and all tasks all allowed to continue
 *
 * This grabs tasklist_lock.
 *
 * Use cr_load_linkage to place the linkage structure into memory before
 * this function is called.
 */
static int
cr_restore_pids(cr_rstrt_req_t *req, cr_rstrt_proc_req_t *proc_req)
{
    int retval;

    /* First, we need to stop, so that none of the children start to
     * execute until their PIDs have been restored. */

    if (atomic_dec_and_test(&proc_req->linkage_counter)) {
        /* True only for the last task.  All other tasks have also reached
	 * this point and will block at the barrier below while we do
	 * the remaining work. */

        CR_KTRACE_HIGH_LVL("Now restoring the pids... ");
        retval = cr_do_restore_pids(req, proc_req->linkage, proc_req->linkage_size);

	if (retval < 0) {
            CR_KTRACE_HIGH_LVL("Linkage restore failed.  Releasing reserved ids.");
#if HAVE_2_6_TASK_LINKAGE
            cr_release_ids(proc_req->linkage, proc_req->linkage_size);
#endif
	} else {
            CR_KTRACE_HIGH_LVL("Linkage restore finished...");
	}

	retval = 0;
    } else {
	/* Nothing to do, so just return successful */
	retval = 0;
    }

    /* Now wait until everything quiets down */
    cr_barrier_enter(&proc_req->linkage_barrier);

    /* XXX: check for broken barrier */

    return retval;
}

/*
 * cr_fgets
 *
 * Sort of like fgets.
 *   Returns >= 0 on success.
 *   0 return means string is NULL.
 *   Returns number of bytes read in bytesread.
 */
int
cr_fgets(char *buf, int size, struct file *filp, int *bytesread)
{
    int ret;
    int read;
    int len = 0;

    ret = -EINVAL;
    if (!(bytesread)) {
        CR_ERR("cr_fgets: NULL bytesread passed as argument!");
        goto out;        
    }
    read = 0;

    ret = -EINVAL;
    if (!(buf)) {
        CR_ERR("cr_fgets: NULL buffer pass as argument!");
        goto out_bytesread;
    }

    ret = -EINVAL;
    if ((size < 0) || (size > PAGE_SIZE)) {
        CR_ERR("cr_fgets:  Bad buffer size.");
        goto out;
    }

    ret = cr_kread(filp, &len, sizeof(len));
    if (ret < 0) {
        CR_ERR("cr_fgets: Bad file read.");
        goto out_bytesread;
    } else if (ret != sizeof(len)) {
        read = ret;
        ret = -EIO;
        CR_ERR("cr_fgets: Short read.");
        goto out_bytesread;
    }
    read = ret;

    if (len == 0) {
        /* old string was NULL.  i.e. never initialized rather than a
	 * pointer to an empty string */
        ret = 0;
        goto out_bytesread;
    } else if ((len < 0) || (len > PAGE_SIZE) || (len > size)) {
        ret = -EINVAL;
        CR_ERR("cr_fgets:  Bad string length %d in file.", len);
        goto out_bytesread;
    } 

    ret = cr_kread(filp, buf, len);
    if (ret < 0) {
        CR_ERR("cr_fgets:  Bad file read.");
        goto out;
    } else if (ret != len) {
        read += ret;
        ret = -EIO;
        CR_ERR("cr_fgets:  Short read.");
        goto out_bytesread;
    }
    read += ret;

out_bytesread:
    *bytesread = read;
out:
    return ret;
}

/*
 * cr_restore_files_struct
 *
 * Restore the files_struct (file table).
 *
 * This function just expands our fd array to the correct size,
 * returning max_fds or error.
 *
 * In the future, if we need any information that applies to all opened files
 * that info can be placed here.
 */
static int
cr_restore_files_struct(cr_rstrt_req_t *req)
{
    struct cr_files_struct cr_fs;
    struct file *c_filp;
    int retval = 0;

    c_filp = cr_src_gett(req, cr_files_obj);
    if (!c_filp) {
	CR_ERR("Unable to find context file to load cr_files_struct.");
	goto out_no_cfilp;
    }

    retval = cr_kread(c_filp, &cr_fs, sizeof(cr_fs));
    if (retval != sizeof(cr_fs)) {
	CR_ERR("cr_fs: Bad file read!");
	retval = -EIO;
	goto out_putt;
    }

    retval = 0;
    if (cr_fs.cr_obj != cr_files_obj) {
	CR_ERR("cr_fs: Invalid cr_files_struct!");
	retval = -EINVAL;
	goto out_putt;
    }

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (((unsigned long) cr_fs.cr_max_fds) > CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur) {
        CR_KTRACE_HIGH_LVL("Too many files %d > %lu", cr_fs.cr_max_fds, CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur);
	goto out_putt;
    }

    cr_file_lock_write(&current->files->file_lock);

    retval = expand_files(current->files, cr_fs.cr_max_fds-1);
    if (retval < 0) {
	CR_ERR("Failed to resize the fd array. (err=%d)", retval);
	goto out_unlock;
    }

#if HAVE_FILES_STRUCT_NEXT_FD
    current->files->next_fd = cr_fs.cr_next_fd;
#else
    {
      cr_fdtable_t *fdt = cr_fdtable(current->files);
      fdt->next_fd = cr_fs.cr_next_fd;
    }
#endif

    retval = cr_fs.cr_max_fds;

out_unlock:
    cr_file_lock_unwrite(&current->files->file_lock);
out_putt:
    cr_src_putt(c_filp);
out_no_cfilp:
    return retval;
}

/*
 * cr_load_file_info
 * 
 * Reads in a file_info from disk, returns it.
 */
static int
cr_load_file_info(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    struct file *filp;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    filp = cr_src_gett(req, cr_file_obj);
    if (!filp) {
        retval = -EBADF;
        goto out;
    }

    retval = cr_kread(filp, file_info, sizeof(*file_info));
    if (retval != sizeof(*file_info)) {
        CR_ERR("cr_load_file_info: Bad read! retval=%d", retval);
        retval = -EIO;
	goto out_putt;
    }

    retval = -EINVAL;
    if (file_info->cr_type != cr_file_info_obj) {
        CR_ERR("cr_load_file_info: Garbage in context file! (type=%d)", file_info->cr_type);
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(filp);
out:
    return retval;
}

/*
 * cr_fd_claim
 *
 * Atomically checks and claims an fd
 *
 * returns < 0 on conflict w/ an existing fd
 */
static int
cr_fd_claim(int fd)
{
    cr_fdtable_t *fdt;
    int retval;

    /* Mark the fd in use */
    cr_file_lock_write(&current->files->file_lock);
    fdt = cr_fdtable(current->files);
    if (FD_ISSET(fd, fdt->open_fds)) {
	retval = -EBUSY;
    } else {
	retval = 0;
	FD_SET(fd, fdt->open_fds);
	FD_CLR(fd, fdt->close_on_exec);
    }
    cr_file_lock_unwrite(&current->files->file_lock);

    return retval;
}

/*
 * cr_open_fd: open a file - accepts file descriptor as argument
 * 
 * NOTE:  Does not attempt to expand the fdset array.
 * Use cr_restore_files_struct to expand fdset to the correct size.
 *
 * NOTE:  Does not check limits on number of open files.
 */ 
static struct file *
cr_open_fd(int fd, char *path, int flags, int mode)
{
    int rc;
    struct file *filp;

    /* check to see if we have too many files open */
    filp = ERR_PTR(-EMFILE);
    if (fd >= CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur)
	goto out;

#if 0 /* If we were paranoid */
    filp = ERR_PTR(-EMFILE);
    if (fd >= current->files->max_fds)
	goto out;
#endif

    /* Mark the fd in use */
    rc = cr_fd_claim(fd);
    /* make sure it's not in use already.  shouldn't ever happen */
    if (rc) {
        CR_WARN("File opened unexpectedly.");
	filp = ERR_PTR(rc);
        goto out;
    }

    /* open the file */
    filp = filp_open(path, flags, mode);
    if (IS_ERR(filp)) {
        goto out;
    }

    /* install the filp in the fd array -- note race see fd_install */
    fd_install(fd, filp);

out:
    return filp;
}

/*
 * cr_anonymous_rename
 *
 * Rewrite a filename to an anonymous value
 *
 * len is strlen and size is buffer size
 */
static char *
cr_anonymous_rename(char *in_buf, unsigned long id)
{
    size_t len;
    char *out_buf, *p;

    len = strlen(in_buf);
    if (len >= PAGE_SIZE) {
	/* XXX: probably not what we want to do */
	len = PAGE_SIZE - 1;
    }

    out_buf = (char *) __get_free_page(GFP_KERNEL);
    if (out_buf == NULL) {
	goto out;
    }
    memcpy(out_buf, in_buf, len+1);

    p = out_buf + len;
    while ((p != out_buf) && (*p != '/')) {
	--p;
    }
    ++p;
    len = p - out_buf;

#if BITS_PER_LONG == 32
    if (len > PAGE_SIZE - 20) {
	CR_ERR("cr_anonymous_rename - unlinked name too long for renaming");
	goto out_free;
    }
    sprintf(p, ".blcr_%04x.%08lx", (unsigned int)current->pid, id);
#elif BITS_PER_LONG == 64
    if (len > PAGE_SIZE - 24) {
	CR_ERR("cr_anonymous_rename - unlinked name too long for renaming");
	goto out_free;
    }
    sprintf(p, ".blcr_%04x.%016lx", (unsigned int)current->pid, id);
#else
    #error "No value for BITS_PER_LONG"
#endif

    return out_buf;

out_free:
    free_page((unsigned long)out_buf);
out:
    return NULL;
}

/*
 * cr_restore_open_file
 *
 * reopens files
 */
static int
cr_restore_open_file(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    char *buf;
    struct file *c_filp;
    struct file *filp;
    struct cr_open_file open_file;
    int len;

    CR_KTRACE_FUNC_ENTRY();

    c_filp = cr_src_gett(req, cr_file_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        CR_ERR("Couldn't allocate buffer for file name.");
        retval = -ENOMEM;
        goto out_putt;
    }
    
    /* read in the cr_open_file struct */
    retval = cr_kread(c_filp, &open_file, sizeof(open_file));
    if (retval != sizeof(open_file)) {
        CR_ERR("cr_restore_open_file - Bad read.");
        retval = -EIO;
        goto out_free;
    }

    if (open_file.cr_type != cr_open_file_obj) {
        CR_ERR("cr_restore_open_file.  Garbage in context file!");
    }

    /* now read out the name */
    retval = cr_fgets(buf, PAGE_SIZE, c_filp, &len);
    if (retval < 0) {
        CR_ERR("cr_restore_open_file - pwd Bad read!");
        retval = -EIO;
        goto out_free;
    } else if (retval == 0) {
        /* 0 corresponds to NULL, should never have happened. */
        CR_ERR("A file with no name?");
	retval = -EINVAL;
	goto out_free;
    }

    if (do_not_restore) {
        retval = 0;
	goto out_free;
    }

    /* reopen file */
    filp = cr_open_fd(file_info->fd, buf, open_file.f_flags, open_file.i_mode);
    if (IS_ERR(filp)) {
	CR_WARN("Failed to open file '%s'", buf);
	retval = PTR_ERR(filp);
        goto out_free;
    }
    if (S_ISDIR(filp->f_dentry->d_inode->i_mode)) {
	retval = -EISDIR;
        goto out_free;
    }

    /* restore position in file */
    retval = sys_lseek(file_info->fd, open_file.f_pos, 0);
    if (retval < 0) {
        CR_ERR("Couldn't restore file pointer.");
	goto out_free;
    }

    if (!S_ISREG(filp->f_dentry->d_inode->i_mode)) {
	/* Skip the truncation on non-regular files */
    } else if ((filp->f_dentry == c_filp->f_dentry) ||
        (filp->f_dentry->d_inode == c_filp->f_dentry->d_inode) ||
        ((filp->f_dentry->d_inode->i_sb == c_filp->f_dentry->d_inode->i_sb) &&
         (filp->f_dentry->d_inode->i_ino == c_filp->f_dentry->d_inode->i_ino))) {
	/* Don't trucate the current input!! */
    } else if (filp->f_mode & FMODE_WRITE) {
        retval = sys_ftruncate(file_info->fd, open_file.i_size);
        if (retval < 0) {
            CR_ERR("Couldn't restore file length.");
	    goto out_free;
        }
    }

    retval = 0;

out_free:
    free_page((unsigned long) buf);
out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

/*
 * cr_restore_open_dir
 *
 * reopens directories
 */
static int 
cr_restore_open_dir(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    char *buf;
    struct file *c_filp;
    struct file *filp;
    struct cr_open_dir open_dir;
    int len;

    CR_KTRACE_FUNC_ENTRY();

    c_filp = cr_src_gett(req, cr_file_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        CR_ERR("Couldn't allocate buffer for file name.");
        retval = -ENOMEM;
        goto out_putt;
    }
    
    /* read in the cr_open_dir struct */
    retval = cr_kread(c_filp, &open_dir, sizeof(open_dir));
    if (retval != sizeof(open_dir)) {
        CR_ERR("cr_restore_open_dir - Bad read.");
        retval = -EIO;
        goto out_free;
    }

    if (open_dir.cr_type != cr_open_dir_obj) {
        CR_ERR("cr_restore_open_dir.  Garbage in context file!");
    }

    /* now read out the name */
    retval = cr_fgets(buf, PAGE_SIZE, c_filp, &len);
    if (retval < 0) {
        CR_ERR("cr_restore_open_dir - Bad read!");
        retval = -EIO;
        goto out_free;
    } else if (retval == 0) {
        /* 0 corresponds to NULL, should never have happened. */
        CR_ERR("A directory with no name?");
	retval = -EINVAL;
	goto out_free;
    }

    if (do_not_restore) {
	/* XXX: Do we really want non-dir to ever replace dir? */
        retval = 0;
	goto out_free;
    }

    /* reopen directory */
    filp = cr_open_fd(file_info->fd, buf, open_dir.f_flags, open_dir.i_mode);
    if (IS_ERR(filp)) {
	retval = PTR_ERR(filp);
        goto out_free;
    }
    if (!S_ISDIR(filp->f_dentry->d_inode->i_mode)) {
	retval = -ENOTDIR;
        goto out_free;
    }

    /* restore position */
    retval = sys_lseek(file_info->fd, open_dir.f_pos, 0);
    if (retval < 0) {
        CR_ERR("Couldn't restore file pointer.");
	goto out_free;
    }

    retval = 0;

out_free:
    free_page((unsigned long) buf);
out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

static int 
cr_restore_open_link(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

/*
 * f_flags - same as in user space
 * f_mode  - funny kernel thing (can reconstruct from f_flags)
 *   00 - no permissions needed
 *   01 - read permission needed
 *   10 - write permission needed
 *   11 - read/write permissions needed
 */
static int
cr_posix_flags_to_f_mode(int posix_flags)
{
    return ((posix_flags+1) & O_ACCMODE);

}

/*
 * This is more or less the same thing as open(), with the major exception
 * that we DO NOT BLOCK.  (Normal open() does block in certain cases.)
 * POSIX tells us that if we open write-only when no readers are present
 * we must either block, or fail w/ errno=ENXIO if O_NONBLOCK is set.
 * Therefore, we can't JUST set O_NONBLOCK.  We must ALSO mess w/ the
 * flags to open read/write when the user asks for write-only.
 *
 * XXX: this could easily break for a client attaching to a FIFO for
 * which it has write permission but not read permission.  We should
 * really just see about reproducing more of the fifo_open instead.
 */
int
cr_open_named_fifo(int fd, struct dentry *dentry, struct vfsmount *mnt, int flags)
{
    int retval = 0;
    struct file *filp;
    int tmp_flags;

    tmp_flags = flags | O_NONBLOCK;
    if (flags & O_WRONLY) {
        /* We promote W-only to RW, dropping the R later */
	tmp_flags = (tmp_flags & ~O_ACCMODE) | O_RDWR;
    }

    filp = dentry_open(dget(dentry), mntget(mnt), tmp_flags);
    if (IS_ERR(filp)) {
	retval = PTR_ERR(filp);
        goto out;
        /* get out */
    }

    /* Check here that we actually got a FIFO */
   if (!S_ISFIFO(filp->f_dentry->d_inode->i_mode)) {
	retval = -EEXIST;
        CR_ERR("encountered non-FIFO with same name as desired FIFO.");
        goto out;
   }
	
    /* Fixup: */
    cr_pipe_lock(filp->f_dentry->d_inode);
    filp->f_flags = flags;
    if (flags & O_WRONLY) {
        /* Fix up the mode and the inode if we faked RDWR */
        struct pipe_inode_info *pipe = filp->f_dentry->d_inode->i_pipe;
        filp->f_mode = cr_posix_flags_to_f_mode(O_WRONLY);
	filp->f_op = &write_fifo_fops;
	pipe->readers--;
	wake_up_interruptible(&pipe->wait);
    }
    cr_pipe_unlock(filp->f_dentry->d_inode);

    /* Install it */
    retval = cr_fd_claim(fd);
    if (retval < 0) {
        CR_ERR("File descriptor %d in use.", fd);
	goto out;
    }
    fd_install(fd, filp);

out:
    return retval;
}

/*
 *
 */

/* half of do_pipe - set up new file pointer to an existing pipe */
static int
cr_open_unnamed_pipe(cr_rstrt_req_t *req, struct cr_file_info *file_info, 
	struct cr_fifo *cf_fifo)
{
    int retval = -EINVAL;
    struct file *filp;
    struct pipe_inode_info *pipe;

    CR_KTRACE_LOW_LVL("Opening FIFO ID %p", cf_fifo->fifo_id);

    /* unnamed pipe */
    filp = get_empty_filp();

    if (!filp) {
        retval = -ENOMEM;
	goto out_err;
    }

    /* first the boring stuff */
    filp->f_pos     = cf_fifo->fifo_pos;
    filp->f_flags   = cf_fifo->fifo_flags;
    filp->f_mode    = cr_posix_flags_to_f_mode(filp->f_flags);
    filp->f_version = cf_fifo->fifo_version;

    /* now the dentry and vfsmnt, to attach us to the filesystem */
    cr_find_object(req->map, cf_fifo->fifo_dentry, (void **) &filp->f_dentry);
    if (!(filp->f_dentry)) {
	CR_ERR("cr_open_unnamed_pipe: Couldn't look up dentry %p in object map", cf_fifo->fifo_dentry);
        goto out_free;
    }

    cr_find_object(req->map, cf_fifo->fifo_vfsmnt, (void **) &filp->f_vfsmnt);
    if (!(filp->f_vfsmnt)) {
	CR_ERR("cr_open_unnamed_pipe: Couldn't look up vfsmnt %p in object map", cf_fifo->fifo_vfsmnt);
        goto out_free;
    }

    filp->f_dentry = dget(filp->f_dentry);
    if (!filp->f_dentry) {
	CR_ERR("dget(%p) failed", filp->f_dentry);
        goto out_free;
    }
        
    filp->f_vfsmnt = mntget(filp->f_vfsmnt);
    if (!filp->f_vfsmnt) {
	CR_ERR("mntget(%p) failed", filp->f_vfsmnt);
        goto out_dput;
    }

    /* now the file_operations f_op */
    pipe = filp->f_dentry->d_inode->i_pipe;
    switch(filp->f_mode) {
        case 1:
	  /* rdonly */
	  filp->f_op = &read_pipe_fops;
          pipe->readers++;
	  break;
	case 2:
	  /* wronly */
	  filp->f_op = &write_pipe_fops;
          pipe->writers++;
	  break;
	case 3:
	  /* rdwr */
	  filp->f_op = &rdwr_pipe_fops;
	  if (filp->f_mode & FMODE_READ) 
	    pipe->readers++;
          if (filp->f_mode & FMODE_WRITE) 
	    pipe->writers++;
	  break;
	default:
	  CR_ERR("Bad access mode for pipe.");
	  retval = -EBADF;
	  goto out_mntput;
	  break;
    }

    /* allocate the fd */
    retval = cr_fd_claim(file_info->fd);
    if (retval) {
        /* this should never happen.  Something else is already using our fd.  
	 * Race? Or maybe forgot to set do_not_restore... */
        CR_ERR("File descriptor %d in use.", file_info->fd);
        goto out_mntput;
    }

    /* install the fd */
    fd_install(file_info->fd, filp);

    return 0;

    /* only get here if something went wrong. */
out_mntput:
    mntput(filp->f_vfsmnt);
out_dput:
    dput(filp->f_dentry);
out_free:
    put_filp(filp);
out_err:
    return retval;
}

/* Use path_lookup() if present as a function or a macro */
#if defined(path_lookup) || defined(CR_KCODE_path_lookup) || defined(CR_EXPORTED_KCODE_path_lookup)
  /* Macro or function (possibly unexported) */
  #define cr_path_lookup path_lookup
#else
  /* based on linux/fs/namei.c:path_lookup in later 2.4.x kernels */
  static int cr_path_lookup(const char *path, unsigned flags, struct nameidata *nd)
  {
    int retval = 0;
    if (path_init(path, flags, nd)) {
      retval = path_walk(path, nd);
    }
    return retval;
  }
#endif

/* from on linux/fs/namei.c:lookup_create */
/* Differs in that the balancing up() takes place on error */
static struct dentry *cr_lookup_create(struct nameidata *nd, int is_dir)
{
#if defined(CR_EXPORTED_KCODE_lookup_create)
    struct dentry *dentry;

    dentry = lookup_create(nd, is_dir);
    if (IS_ERR(dentry)) {
      cr_inode_unlock(nd->dentry->d_inode);
    }
    return dentry;
#elif HAVE_LOOKUP_HASH
    struct dentry *dentry;

    cr_inode_lock(nd->dentry->d_inode);
    dentry = ERR_PTR(-EEXIST);
    if (nd->last_type != LAST_NORM) {
	goto out_fail;
    }
    dentry = lookup_hash(&nd->last, nd->dentry);
    if (IS_ERR(dentry)) {
	goto out_fail;
    }
    if (!is_dir && nd->last.name[nd->last.len] && !dentry->d_inode) {
	goto out_noent;
    }
    return dentry;

out_noent:
    dput(dentry);
    dentry = ERR_PTR(-ENOENT);
out_fail:
    cr_inode_unlock(nd->dentry->d_inode);
    return dentry;
#else
    #error "No lookup_create() mechanism exported"
    return ERR_PTR(-ENOSYS);
#endif
}

/* cr_mknod - based on linux/fs/namei.c:sys_mknod
 *
 * Creates regular files, fifos or devices, making them anonymous (unlinked)
 * if desired.
 * Returns a dentry for the resulting filesystem objects, and the corresponding
 * vfsmnt can be obtained in nd->mnt.  Together these two can be passed
 * to dentry_open(), even for an unlinked inode.
 * In the event of an error, no dput() or path_release() is required,
 * otherwise they are.
 * In the event that an object exists with the given name, it will be
 * returned.  Therefore the caller must check the type of the resulting
 * dentry (addition of a 'uniq' flag might change this behavior later).
 * It is also possible to get -EEXIST from the lookup_create step.
 */
static struct dentry *
cr_mknod(struct nameidata *nd, char *name, int mode, int unlinked_id)
{
    struct dentry * dentry;
    char *_name;
    int err;

    if (unlinked_id) {
	/* Generate a replacement name which we will use instead of the original one. */
	_name = cr_anonymous_rename(name, unlinked_id);
	if (!_name) {
	    CR_ERR("cr_mknod - failed to rename unlinked object");
	    err = -ENOMEM;
	    goto out;
	}
    } else {
	_name = name;
    }

    err = cr_path_lookup(_name, LOOKUP_PARENT, nd);
    if (err) {
	CR_ERR("Couldn't path_lookup for mknod %s.  err=%d.", name, err);
	goto out_free;
    }

    dentry = cr_lookup_create(nd, 0);
    if (IS_ERR(dentry)) {
	err = PTR_ERR(dentry);
	CR_ERR("Couldn't lookup_create for mknod %s.  err=%d.", name, err);
	goto out_release;
    }

    err = vfs_mknod(nd->dentry->d_inode, dentry, mode, 0 /* ignored */);
    if (unlinked_id && !err) { /* Note that we don't unlink if we failed to create */
	dget(dentry);	/* ensure unlink doesn't destroy the dentry */
	/* Note possibility of silent failure here: */
	(void)vfs_unlink(nd->dentry->d_inode, dentry);
	dput(dentry);
    }
    cr_inode_unlock(nd->dentry->d_inode);
    if (err == -EEXIST) {
	/* We fall through and return the dentry */
    } else if (err) {
	CR_ERR("Couldn't vfs_mknod %s.  err=%d.", name, err);
	goto out_put;
    }

    return dentry;

out_put:
    dput(dentry);
out_release:
    path_release(nd);
out_free:
    if (unlinked_id) {
	free_page((unsigned long)_name);
    }
out:
    return (struct dentry *)ERR_PTR(err);
}

static int
cr_creat_named_fifo(int fd, char *name, int flags, int mode, int unlinked)
{
    int retval;
    struct nameidata nd;
    struct dentry * dentry;

    /* make a new FIFO, or find the existing one */
    dentry = cr_mknod(&nd, name, mode, unlinked ? fd+1 : 0);
    if (IS_ERR(dentry)) {
	retval = PTR_ERR(dentry);
	CR_ERR("Failed to create named fifo %s.  err=%d.", name, retval);
	goto out;
    }

    /* open it */
    /* This will check for correct object type in the event of an existing object */
    retval = cr_open_named_fifo(fd, dentry, nd.mnt, flags);
    if (retval < 0) {
	CR_ERR("Couldn't open %s.  err=%d.", name, retval);
	goto out_dput;
    }

    retval = 0;

out_dput:
    dput(dentry);
    path_release(&nd);
out:
    return retval;
}

static int
cr_restore_pipe_buf(struct file *cf_filp, struct inode *p_inode, int buf_len, int do_not_restore_flag)
{
    struct pipe_inode_info *pipe;
    int retval;
    void *buf;

    if (!buf_len) {
        CR_KTRACE_LOW_LVL("Skipping empty pipe buffer %p", p_inode);
	retval = 0;
	goto out;
    }

    buf = (void *)kmalloc(buf_len, GFP_KERNEL);
    if (!buf) {
	retval = -ENOMEM;
	goto out;
    }

    retval = cr_kread(cf_filp, buf, buf_len);
    if (retval != buf_len) {
	CR_ERR("cr_restore_pipe_buf - Bad read.");
	retval = -EIO;
	goto out_free;
    }

    if (do_not_restore_flag) {
        CR_KTRACE_LOW_LVL("Unrestored pipe buffer %p length = %d", p_inode, buf_len);
        retval = 0;
        goto out_free;
    }

    if (cr_pipe_lock_interruptible(p_inode)) {
        retval = -ERESTARTSYS;
        goto out_free;
    }

    pipe = p_inode->i_pipe;
#if HAVE_PIPE_INODE_INFO_BASE
    if (PIPE_LEN(*p_inode)) {
        /* someone put data in here already -- bail */
        retval = -EBUSY;
	goto out_up;
    }

    if (buf_len) {
	memcpy((void *)PIPE_BASE(*p_inode), buf, buf_len);
        PIPE_LEN(*p_inode) = buf_len;
        CR_KTRACE_LOW_LVL("Pipe buffer %p length = %d inode = %lu", p_inode, buf_len, p_inode->i_ino);
	retval = 0;
    }
#else
    {   struct pipe_inode_info *pipe = p_inode->i_pipe;
	char * p;
        int nrbufs;

	if (pipe->nrbufs != 0) {
            /* someone put data in here already -- bail */
            retval = -EBUSY;
	    goto out_up;
	}

	nrbufs = 0;
	p = buf;
	while (buf_len) {
	    size_t count = buf_len;
	    struct pipe_buffer *buf = pipe->bufs + nrbufs;
	    struct page *page = alloc_page(GFP_HIGHUSER);
	    if (count > PAGE_SIZE)
		count = PAGE_SIZE;
	    memcpy(kmap(page), p, count);
	    kunmap(page);
	    buf->page = page;
	    buf->offset = 0;
	    buf->len = count;
	    buf->ops = &anon_pipe_buf_ops;
	    ++nrbufs;
	    buf_len -= count;
	    p += count;
	}
	pipe->curbuf = 0;
	pipe->nrbufs = nrbufs;
    }
#endif

out_up:
    cr_pipe_unlock(p_inode);
out_free:
    kfree(buf);
out:
    return retval;
}

/* 
 * pipe was instantiated already, so we connect our file descriptor
*  to the pipe inode.
*/
static int
cr_couple_fifo(cr_rstrt_req_t *req, struct cr_file_info *file_info, 
	struct cr_fifo *cf_fifo, char *name, int do_not_restore_flag)
{
    int retval;

    retval = -EINVAL;
    if (!cr_find_object(req->map, cf_fifo->fifo_id, NULL)) {
	CR_ERR("cr_couple_fifo - unable to find previously restored fifo (id=%p)", cf_fifo->fifo_id);
	goto out;
    }

    CR_KTRACE_LOW_LVL("%s:  phase 2: connecting second end.", (char *) name);
    /* no obvious way to distinguish named pipes from unnamed besides 
     * filename.  anything with a '/' at the beginning is a named fifo, 
     * and anything with a 'p' is a pipe. */
    if (*((char *) name) == '/') {
        /* named fifo */
        if (do_not_restore_flag) {
            retval = 0;
        } else {
	    struct dentry *dentry;
	    struct vfsmount *mnt;
	    cr_find_object(req->map, cf_fifo->fifo_dentry, (void **) &dentry);
	    cr_find_object(req->map, cf_fifo->fifo_vfsmnt, (void **) &mnt);
	    retval = cr_open_named_fifo(file_info->fd, dentry, mnt, cf_fifo->fifo_flags);
        }
        if (retval < 0) {
            CR_ERR("cr_couple_fifo:  unable to open named fifo: %s",
                   (char *) name);
            goto out;
        }
    } else if (*((char *) name) == 'p') {
        /* unnamed pipe */
        if (do_not_restore_flag) {
            retval = 0;
        } else {
            retval = cr_open_unnamed_pipe(req, file_info, cf_fifo);
        }
        if (retval < 0) {
            CR_ERR("cr_couple_fifo:  unable to open unnamed pipe: %s",
                   (char *) name);
            goto out;
        }
    } else {
        /* garbage */
        retval = -EINVAL;
        CR_ERR("bogus pipe filename in context file: %s", (char *) name);
        goto out;
    }


  out:
    return retval;
}

static int
cr_make_new_pipe(cr_rstrt_req_t *req, struct cr_file_info *file_info, 
	         struct cr_fifo *cf_fifo, char *name, int do_not_restore_flag)
{
    struct file *p_filp;
    int retval;

    CR_KTRACE_LOW_LVL("%s:  Phase 1: Making new pipe.", (char *) name);

    if (*((char *) name) == '/') {
        /* sanity check */
        if (!(S_IFIFO & cf_fifo->fifo_perms)) {
            CR_WARN("Pipe object not a FIFO!");
        }

        if (do_not_restore_flag) {
            retval = 0;
        } else {
            retval = cr_creat_named_fifo(file_info->fd, (char *) name,
                                         cf_fifo->fifo_flags,
                                         S_IFIFO | cf_fifo->fifo_perms,
					 file_info->unlinked);
        }

        if (retval < 0) {
            goto out;
        }
    } else if (*((char *) name) == 'p') {
        int pipe_fds[2];
        int tmp_fd;

        if (do_not_restore_flag) {
            retval = 0;
            goto out;
        }

        retval = do_pipe(pipe_fds);
        if (retval < 0) {
            goto out;
        }

        /* now have two pipe objects, close one of them. */
        if (cr_posix_flags_to_f_mode(cf_fifo->fifo_flags) == 1) {
            sys_close(pipe_fds[1]);
            tmp_fd = pipe_fds[0];
        } else if (cr_posix_flags_to_f_mode(cf_fifo->fifo_flags) == 2) {
            sys_close(pipe_fds[0]);
            tmp_fd = pipe_fds[1];
        } else {
            CR_ERR("Bad access mode for pipe.");
            goto out;
        }

        /* put the fd into the right place */
        if (tmp_fd != file_info->fd) {
            retval = sys_dup2(tmp_fd, file_info->fd);
            if (retval < 0) {
                CR_ERR("sys_dup2(%d,%d) failed.", tmp_fd, file_info->fd);
                sys_close(tmp_fd);
                /* we ignore the error from close here. */
                goto out;
            }

            /* we ignore the error from close here. */
            sys_close(tmp_fd);
        }

    } else {
        /* garbage */
        retval = -EINVAL;
        CR_ERR("Bogus pipe filename in context file: %s", (char *) name);
        goto out;
    }

    p_filp = fget(file_info->fd);
    if (!p_filp) {
        goto out;
    }

    /* now do the easy stuff */
    p_filp->f_pos = cf_fifo->fifo_pos;
    p_filp->f_flags = cf_fifo->fifo_flags;
    p_filp->f_version = cf_fifo->fifo_version;
    /* p_filp->f_fop was initialized correctly by do_pipe/mknod */

    /* fifo_id(inode), dentry, vfsmnt pointers into object table for subsequent open(s) */
    cr_insert_object(req->map, cf_fifo->fifo_id, p_filp->f_dentry->d_inode);
    cr_insert_object(req->map, cf_fifo->fifo_dentry, p_filp->f_dentry);
    cr_insert_object(req->map, cf_fifo->fifo_vfsmnt, p_filp->f_vfsmnt);

    fput(p_filp);

out:
    return retval;
}

static int 
cr_restore_open_fifo(cr_rstrt_req_t *req, struct cr_file_info *file_info,
       	int do_not_restore_flag)
{
    int retval = -ENOSYS;
    struct cr_fifo cf_fifo;
    struct file *cf_filp;
    struct file *p_filp;
    void *buf;
    int len;

    CR_KTRACE_FUNC_ENTRY();

    cf_filp = cr_src_gett(req, cr_fifo_obj);
    if (!cf_filp) {
        retval = -EBADF;
        goto out;
    }

    buf = (void *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_putt;
    }

    /* read fifo struct */
    retval = cr_kread(cf_filp, &cf_fifo, sizeof(cf_fifo));
    if (retval != sizeof(cf_fifo)) {
        CR_ERR("cr_restore_open_fifo - Bad read.");
        retval = -EIO;
        goto out_free;
    }

    CR_KTRACE_LOW_LVL("   Open fifo: id == %p.", cf_fifo.fifo_id);

    if (cf_fifo.cr_type != cr_fifo_obj) {
        CR_ERR
            ("cr_restore_open_fifo.  Garbage found in context file.  Wrong object type.");
        retval = -EINVAL;
        goto out_free;
    }

    /* Force do_not_restore_flag OFF for "internal" pipes */
    if (cf_fifo.fifo_internal) {
	do_not_restore_flag = 0;
	if (fcheck(file_info->fd) != NULL) {
	    sys_close(file_info->fd);
	}
    }

    /* read file name */
    retval = cr_fgets(buf, PAGE_SIZE, cf_filp, &len);
    if (retval < 0) {
        CR_ERR("pipe_name: Bad file read!");
        goto out_free;
    }
    
    /* now, actually rebuild the thing. */
    if (cf_fifo.fifo_len == ((unsigned int)-1)) {
        retval = cr_couple_fifo(req, file_info, &cf_fifo, buf,
                                do_not_restore_flag);
        if (retval < 0) {
            goto out_free;
        }

        /* 
         * the pipe buffer should have been restored when we first built the pipe
         */
    } else {
        retval = cr_make_new_pipe(req, file_info, &cf_fifo, buf,
                                  do_not_restore_flag);
        if (retval < 0) {
            goto out_free;
        }

	/* now read the pipe buffer data, and optionally restore it */
	p_filp = fget(file_info->fd);
	if (!p_filp) {
	    goto out_free;
	}
	retval = cr_restore_pipe_buf(cf_filp, p_filp->f_dentry->d_inode,
                                     cf_fifo.fifo_len, do_not_restore_flag);
	fput(p_filp);
	if (retval < 0) {
	    CR_ERR("could not restore pipe buffer");
	    goto out_free;
	}
    }

out_free:
    free_page((unsigned long) buf);
out_putt:
    cr_src_putt(cf_filp);
out:
    return retval;
}

static int 
cr_restore_open_socket(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    /* XXX:  Kludge to recover w/o restoring a socket */
    retval = 0;

    return retval;
}

static int 
cr_restore_open_chr(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_chrdev cf_chrdev;
    struct file *c_filp;

    CR_KTRACE_FUNC_ENTRY();

    c_filp = cr_src_gett(req, cr_chrdev_obj);
    if (!c_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_chrdev struct */
    retval = cr_kread(c_filp, &cf_chrdev, sizeof(cf_chrdev));
    if (retval != sizeof(cf_chrdev)) {
        CR_ERR("cr_restore_open_chr: Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* XXX:  Validate type */
    retval = -EINVAL;
    if (cf_chrdev.cr_type != cr_open_chr) {
        CR_ERR("cr_restore_open_chr: Garbage in context file.");
        goto out_putt;
    }

    CR_KTRACE_LOW_LVL("   Character device.  major=%d, minor=%d", cf_chrdev.cr_major, cf_chrdev.cr_minor);

    if (do_not_restore) {
        retval = 0;
	goto out_putt;
    }

    retval = 0;

out_putt:
    cr_src_putt(c_filp);
out:
    return retval;
}

static int 
cr_restore_open_blk(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int
cr_dup2(cr_rstrt_req_t *req, struct file *old_filp, int fd)
{
    struct file *new_filp;
    int retval = 0;

    /* translate the filp to its new value */
    if (!cr_find_object(req->map, old_filp, (void **) &new_filp)) {
        CR_ERR("cr_dup2: Could not locate restored filp %p.", old_filp);
	retval = -EINVAL;
        goto out;
    }
    if (new_filp == NULL) {
	/* The file was silently ignored before - continue to do so. */
        CR_WARN("cr_dup2: Skipping dup of ignored filp %p.", old_filp);
	goto out;
    }
    get_file(new_filp);

    /* allocate the fd */
    retval = cr_fd_claim(fd);
    if (retval) {
        /* this should never happen.  */
        CR_ERR("cr_dup2: File descriptor %d in use.", fd);
	fput(new_filp);
        goto out;
    }

    /* install it */
    fd_install(fd, new_filp);

out:
    return retval;
}

static int 
cr_restore_open_dup(cr_rstrt_req_t *req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_dup cf_dup;
    struct file *cf_filp;

    CR_KTRACE_FUNC_ENTRY();

    cf_filp = cr_src_gett(req, cr_fifo_obj);
    if (!cf_filp) {
        retval = -EBADF;
        goto out;
    }

    /* read in the cr_dup struct */
    retval = cr_kread(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
        CR_ERR("cr_restore_open_dup: Bad read.");
        retval = -EIO;
        goto out_putt;
    }

    /* Validate type (the only field we've got) */
    retval = -EINVAL;
    if (cf_dup.cr_type != cr_open_dup) {
        CR_ERR("cr_restore_open_dup: Garbage in context file.");
        goto out_putt;
    }

    /* ALWAYS honor dups so that overrides will catch all instances together */
    if (fcheck(file_info->fd) != NULL) {
	sys_close(file_info->fd);
    }

    retval = cr_dup2(req, file_info->orig_filp, file_info->fd);
    if (retval<0) {
        CR_ERR("cr_restore_open_dup: Could not cr_dup2(%p -> %d).", file_info->orig_filp, file_info->fd);
        goto out_putt;
    }
    CR_KTRACE_LOW_LVL("cr_dup2(%p -> %d) successful", file_info->orig_filp, file_info->fd);

    retval = 0;

out_putt:
    cr_src_putt(cf_filp);
out:
    return retval;
}

static int 
cr_end_of_file_restore(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    int retval = 0;

    CR_KTRACE_FUNC_ENTRY();

    /* Nothing to do here yet */

    return retval;
}

static int
cr_restore_file_locks(cr_rstrt_req_t *req, struct cr_file_info *file_info)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    /* Nothing to do here yet */

    return retval;
}

static int
cr_finish_all_file_restores(cr_rstrt_req_t *req, int max_fds)
{
    int retval = -ENOSYS;
    int fd;

    CR_KTRACE_FUNC_ENTRY();

    cr_file_lock_read(&current->files->file_lock);
    CR_KTRACE_LOW_LVL("current->files=%p", current->files);
    for(fd=0; fd<max_fds; ++fd) {
      struct file *filp;
      filp = fcheck(fd);
      if (filp) {
        CR_KTRACE_LOW_LVL("fd=%d filp=%p filp->dentry=%p d_count=%d d_inode=%p", fd, filp, filp->f_dentry, atomic_read(&filp->f_dentry->d_count), filp->f_dentry->d_inode);
      }
    }
    cr_file_lock_unread(&current->files->file_lock);

    return retval;
}

/*
 * cr_restore_all_files
 *
 * Main entry point for file recovery
 *
 * In the future, may want to set this so that it manipulates file "objects"
 * with load / restore methods
 */
static int
cr_restore_all_files(cr_rstrt_req_t *req)
{
    int do_not_restore_flag;
    int i, retval, max_fds;
    struct cr_file_info file_info;
    struct files_struct *files = current->files;
    cr_fdtable_t *fdt;

    /* close-on-exec() of caller's fds before we start restoring */
    CR_KTRACE_HIGH_LVL("close-on-exec of callers files");
    cr_file_lock_write(&files->file_lock);
    fdt = cr_fdtable(files);
    for (i = 0; i < fdt->max_fds; ++i) {
	if (FD_ISSET(i, fdt->close_on_exec)) {
	    cr_file_lock_unwrite(&files->file_lock);
	    sys_close(i);
	    cr_file_lock_write(&files->file_lock);
	    fdt = cr_fdtable(files);
        }
    }
    cr_file_lock_unwrite(&files->file_lock);

    /* load the files struct */
    CR_KTRACE_HIGH_LVL("recovering fs_struct...");
    retval = cr_restore_files_struct(req);
    if (retval < 0) {
        CR_ERR("Unable to cr_restore_files_struct");
        goto out;
    }
    max_fds = retval;

    file_info.cr_file_type = cr_bad_file_type;
    /* now do per file stuff */
    while (1) {
        retval = cr_load_file_info(req, &file_info);
	if (retval < 0) {
            CR_ERR("Error loading file_info.");
            goto out;
	}


	if (file_info.cr_file_type == cr_end_of_files) {
            /* finished */
            retval = cr_end_of_file_restore(req, &file_info);
            break;
	}

	/* Take first-cut at dnr flag.  Note some cases will ignore this:
	 * 1) dups always follow the dupped fd
	 * 2) internal pipes (those w/ both a reader and writter in the checkpoint)
	 */
        do_not_restore_flag = (fcheck(file_info.fd) != NULL);

        CR_KTRACE_LOW_LVL("   fd=%d dnr=%d", file_info.fd, do_not_restore_flag);

        switch(file_info.cr_file_type) {
        case cr_open_file:
            retval = cr_restore_open_file(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_directory:
            retval = cr_restore_open_dir(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_link:
            retval = cr_restore_open_link(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_fifo:
            retval = cr_restore_open_fifo(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_socket:
            retval = cr_restore_open_socket(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_chr:
            retval = cr_restore_open_chr(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_blk:
            retval = cr_restore_open_blk(req, &file_info, do_not_restore_flag);
            break;
        case cr_open_dup:
            retval = cr_restore_open_dup(req, &file_info, do_not_restore_flag);
            break;
        case cr_bad_file_type:
            /* fall through */
        default:
            CR_ERR("Unknown file type fd=%d type=%d", file_info.fd, file_info.cr_file_type);
            retval = -EBADF;
            goto out;
            break;
        }

        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fd %d (type=%d,err=%d)",
                   __FUNCTION__, current->pid, file_info.fd, file_info.cr_file_type, retval);
            goto out;
        }

	/* Restore close-on-exec flag */
        cr_file_lock_write(&files->file_lock);
	fdt = cr_fdtable(files);
        if (file_info.cloexec) {
	    FD_SET(file_info.fd, fdt->close_on_exec);
        } else {
	    FD_CLR(file_info.fd, fdt->close_on_exec);
	}
	cr_file_lock_unwrite(&files->file_lock);

        /* now acquire locks (unimplemented) */
        cr_restore_file_locks(req, &file_info);

	cr_file_lock_read(&files->file_lock);
        cr_insert_object(req->map, file_info.orig_filp, fcheck(file_info.fd));
	cr_file_lock_unread(&files->file_lock);
    };

    /* now for the second pass.  The above restore routines should have taken
     * care of loading all the data.  the easy work, now we do everything 
     * that has to be deferred until we have all the data
     */
    cr_finish_all_file_restores(req, max_fds);

out:
    return retval;
} 

/*
 * cr_chdir
 *
 * Our own version of the chdir() system call.  Differs in that it accepts
 * pathnames in kernel space, and calls path_init, path_walk, and path_release
 * directly.
 * 
 */
static int
cr_chdir(const char *path)
{
    struct nameidata nd;
    int retval;

    /* lookup the path */
#ifndef LOOKUP_POSITIVE
  /* XXX: is this correct? */
  #define LOOKUP_POSITIVE 0
#endif
    retval = cr_path_lookup(path,LOOKUP_POSITIVE|LOOKUP_FOLLOW|LOOKUP_DIRECTORY,&nd);
    if (retval)
        goto out;

    retval = cr_permission(nd.dentry->d_inode,MAY_EXEC,&nd);
    if (retval)
        goto dput_and_out;

    set_fs_pwd(current->fs, nd.mnt, nd.dentry);

dput_and_out:
    path_release(&nd);
out:
    return retval;
}

/*
 * cr_restore_fs
 *
 * This function does the work of restoring the data in our fs_struct, i.e.
 * the current working directory, and the umask.
 *
 * In the future, this may be modified to also restore a process after a
 * chroot(), if desirable.
 * 
 */
static int
cr_restore_fs_struct(cr_rstrt_req_t *req)
{
    int len;
    int retval;
    int umask;
    char *buf;
    struct file *filp;

    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
        retval = -ENOMEM;
        goto out_nomem;
    }

    filp = cr_src_gett(req, cr_fs_obj);
    if (!filp) {
        retval = -EBADF;
        goto out_nofile;
    }

    /* restore the umask */
    retval = cr_kread(filp, &umask, sizeof(umask));
    if (retval < sizeof(umask)) {
        CR_ERR("umask: Bad file read! retval=%d", retval);
        retval = -EIO;
        goto out_readerr;
    }

    /* read in root */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("root: Bad file read!");
        goto out_readerr;
    }

    /* read in altroot */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("altroot: Bad file read!");
        goto out_readerr;
    } else if (retval > 0) {
        CR_ERR("Warning:  chroot recovery not implemented.");
    }

    /* read in the current working directory */
    retval = cr_fgets(buf, PAGE_SIZE, filp, &len);
    if (retval < 0) {
        CR_ERR("pwd: Bad file read!");
        goto out_readerr;
    }

    /* should change root directory here. */
    /* cr_chroot() */

    /* now change the current working directory */
    retval = cr_chdir(buf);
    if (retval < 0) {
        CR_ERR("cr_chdir:  Unable to change directory to %s", buf);
	goto out_readerr;
    }    

out_readerr:
    cr_src_putt(filp);
out_nofile:
    free_page((unsigned long) buf);
out_nomem:
    return retval;
}

/*
 * This does half of a close.  It removes the fd from a files_struct, but doesn't
 * actually close the fd.
 */
int
cr_hide_filp(struct file *filp, struct files_struct *files)
{
    cr_fdtable_t *fdt;
    int found;
    int fd;

    found = 0;

    cr_file_lock_write(&files->file_lock);
    fdt = cr_fdtable(files);
    for (fd = 0; fd < fdt->max_fds; ++fd) {
	if (filp == fdt->fd[fd]) {
	    ++found;
            rcu_assign_pointer(fdt->fd[fd], NULL);
	    FD_CLR(fd, fdt->close_on_exec);
	    break;
	}
    }
    cr_file_lock_unwrite(&files->file_lock);

    if (found) {
	put_unused_fd(fd);
    } else {
	CR_ERR("File descriptor closed unexpectedly!");
    }

    return found;
}

// "lookup" a proc (thread group) by current->mm
static cr_rstrt_proc_req_t *
lookup_proc_req(cr_rstrt_req_t *req)
{
	cr_rstrt_proc_req_t *result = NULL;
	struct list_head *elem;
	const struct mm_struct *mm = current->mm;

        read_lock(&req->procs_lock);
	list_for_each(elem, &req->procs) {
	    cr_rstrt_proc_req_t *tmp = list_entry(elem, cr_rstrt_proc_req_t, list);
	    if (tmp->mm == mm) {
		result = tmp;
		break;
	    } else if (tmp->mm == NULL) {
	        result = tmp; // "wildcard"
	    }
	}
        read_unlock(&req->procs_lock);
	result->mm = mm; // In case match was the wildcard

	return result;
}

// cr_rstrt_task_complete
//
// Cleanup
static void cr_rstrt_task_complete(cr_task_t *cr_task, int lock_held)
{
    release_rstrt_req(cr_task->rstrt_proc_req->req);
    cr_task->rstrt_proc_req = NULL;
    if (lock_held) {
        __cr_task_put(cr_task);
    } else {
        cr_task_put(cr_task);
    }
}

// cr_rstrt_child
//
// Invoked by a freshly created child process.
//
// The real "restart" part of reconstructing a process image happens here.
//
// Returns pid of cr_restart on success, or a negative on failure
int cr_rstrt_child(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    cr_rstrt_proc_req_t *proc_req; 
    cr_task_t *cr_task = NULL;
    int retval;
    int restart_pid, old_pid;
    struct cr_context_tasklinkage *my_linkage = NULL;
    long state = TASK_RUNNING;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
	CR_ERR("%s: No rstrt_req attached to filp!", __FUNCTION__);
        goto out;
    }

    proc_req = lookup_proc_req(req);
    if (proc_req == NULL) {
	CR_ERR("%s: No rstrt_proc_req for this thread!", __FUNCTION__);
        goto out;
    }

    write_lock(&cr_task_lock);
    {
	cr_task = __cr_task_get(current, 1);
        if (cr_task == NULL) {
		write_unlock(&cr_task_lock);
		retval = -ENOMEM;
                goto out;
        }
	cr_task->rstrt_proc_req = proc_req;
    }
    write_unlock(&cr_task_lock);

    atomic_inc(&req->ref_count);
    atomic_inc(&req->barrier.count);
    clear_bit(0, &req->need_children);

    down(&proc_req->serial_mutex);
    /* Initialize parts of req relevant to restoring task linkage */
    /* First load the linkage */
    if (!test_and_set_bit(0, &proc_req->done_linkage)) {
	retval = cr_load_linkage(req, proc_req);
        if (retval < 0) {
            CR_WARN("Error reading linkage structure! err=%d", retval);
	    retval = -ENOMEM;
	    goto out_release;
        }

#if HAVE_2_6_TASK_LINKAGE
        /* Now we reserve all the ids (pid, pgid, sid) we will need.
         * Note that this can get_free_page(), and thus no locks are held yet.
         */
        if (cr_reserve_ids(proc_req->linkage, proc_req->linkage_size)) {
            CR_WARN("%s [%d]:  PID conflict found by cr_reserve_ids()", __FUNCTION__, current->pid);
	    retval = -EBUSY;
	    goto out_release;	
        }
#endif

        /*
         * Check to see if the linkage is restorable
         */
        read_lock(&tasklist_lock);
        retval = cr_is_linkage_restorable(req, proc_req->linkage, proc_req->linkage_size);
        read_unlock(&tasklist_lock);
        if (!retval) {
            CR_WARN("%s [%d]:  PID conflict found by cr_is_linkage_restorable()", __FUNCTION__, current->pid);
            retval = -EBUSY;
#if HAVE_2_6_TASK_LINKAGE
            cr_release_ids(proc_req->linkage, proc_req->linkage_size);
#endif
	    goto out_release;
        }
    }

    /* Restore credentials (uids and gids) */
    if (!test_and_set_bit(0, &proc_req->done_creds)) {
        CR_KTRACE_HIGH_LVL("%d: Restoring credentials", current->pid);
	retval = cr_restore_creds(proc_req); 
    }
    up(&proc_req->serial_mutex);

    /* Thaw thyself out (called by all threads) */
    {
        struct vmadump_threadundump_args thaw_args;
	long err;

	thaw_args.num_threads = proc_req->total_threads;
	thaw_args.fd = -1;              /* Ignored, but we initialize */
	thaw_args.flags = 0;            /* no flags */
	thaw_args.old_pid = -1;         /* Output, but initialize anyway */

        err = vmadump_thaw_threads(proc_req->file, get_pt_regs(current), &thaw_args);
	if (err < 0) {
	    CR_ERR("thaw_threads returned error, aborting. %ld", err);
#if HAVE_2_6_TASK_LINKAGE
	    /* Do this from only one thread */
	    if (!test_and_set_bit(0, &proc_req->done_abort)) {
                cr_release_ids(proc_req->linkage, proc_req->linkage_size);
	    }
#endif
	    if (req->result >= 0) req->result = err;
	    req->die = 1;
            goto out_release;
	}
        old_pid = thaw_args.old_pid;

	retval = 0;
    }

    /* Associate your old PID with your task in the linkage */
    my_linkage = cr_linkage_find_by_old_pid(proc_req->linkage, proc_req->linkage_size, old_pid);

    /* Map old->new task pointers for linkage restore (all threads) */
    if (my_linkage == NULL) {
        CR_ERR("%s [%d]: ERROR:  No linkage entry for my PID %d",
	       __FUNCTION__, current->pid, old_pid);
	/*
	 * NOTE:  We DO NOT ABORT here.  Instead, we wait for the linkage to be restored,
	 * and rely on the cr_restore_parents code to notice the problem (a NULL tl_task).
	 */
    } else {
	get_task_struct(current);
	cr_insert_object(req->map, my_linkage->myptr, current);
        my_linkage->tl_task = current;
    }

    /* Restore linkage (PID, session ID, etc.) (called by all threads) */
    restart_pid = current->pid; /* save for debugging (or other use) */
    CR_KTRACE_HIGH_LVL("Formerly %s PID %d", current->comm, old_pid);
    retval = cr_restore_pids(req, proc_req);
    if (retval < 0) {
        CR_ERR("%s [%d]:  Unable to restore pids!  (err=%d)",
	       __FUNCTION__, current->pid, retval);
        // XXX:  Get the hell out of here...
        // XXX:  goto ok here?
	if (req->result >= 0) req->result = retval;
	req->die = 1;
        goto out_put_task;
    }

    /* --- File recovery starts here --- */
    /* Exactly one thread should close the rstrt_req file descriptor(s)
     * but they should all wait until this has been done. */
    /* NOTE:  Past this point, you must jump to out_close to abort.  We are removing
     * our reference to the fd, which means do_exit() can't clean up after us. */ 

    down(&proc_req->serial_mutex);
    if (!test_and_set_bit(0, &proc_req->done_hide_cr_fds)) {
        /* Remove the rstrt_req descriptor from our open files. */
        cr_hide_filp(filp, current->files);
        cr_hide_filp(proc_req->file, current->files);
    }

    /* restore fs_struct */
    /* XXX:  put the bit tests inside the restore routines when we do process
     * groups and sessions.  This way, they can just decide they want to skip it.
     */
    if (!test_and_set_bit(0, &proc_req->done_fs)) {
        retval = cr_restore_fs_struct(req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fs!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
            goto out_close;
            // XXX:  same concern
        }
    }

    /* restore the open files */
    if (!test_and_set_bit(0, &proc_req->done_files)) {
        retval = cr_restore_all_files(req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore files!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
            goto out_close;
            // XXX:  same concern
        }
    }

    /* close the request descriptor */
    /* XXX: Need to handle differently when we split to multiple files? */
out_close:
    if (!test_and_set_bit(0, &proc_req->done_close_cr_fds)) {
	int rc;	/* don't overwrite error code w/ 0 */
	CR_KTRACE_LOW_LVL("%d: closing request descriptor", current->pid);
	rc = filp_close(filp, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing checkpoint control descriptor!  "
		   "(err=%d)", rc);
	    retval = rc;
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	}

	CR_KTRACE_LOW_LVL("%d: closing context file descriptor", current->pid);
	rc = filp_close(proc_req->file, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing context file descriptor!"
		   "(err=%d)", rc);
	    retval = rc;
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	}
    }
    up(&proc_req->serial_mutex);

    /* restore handlers (if any)  (all threads) */
    if (!req->die) {
	/* XXX: move to cr_rstrt_task_complete()? */
	if ((my_linkage != NULL) && (my_linkage->cr_task.phase != CR_NO_PHASE)) {
    	    if (!cr_find_object(req->map, my_linkage->cr_task.orig_filp, (void **) &cr_task->filp)) {
		CR_ERR("unable to locate checkpoint control descriptor");
	        if (req->result >= 0) req->result = retval;
	        req->die = 1;
                goto out_put_task;
	    }
	    cr_task->phase = my_linkage->cr_task.phase;
	    cr_task->handler = my_linkage->cr_task.handler;
	    atomic_inc(&cr_task->ref_count);
	}
    }
	    
out_put_task:
    /* Balance the get_task_struct() call made when task was added to object map */
    write_lock_irq(&tasklist_lock);
    put_task_struct(current);
    write_unlock_irq(&tasklist_lock);
out_release:
    /* On failure we need to ensure everyone fails, and inform the requester */
    if ((req->result >= 0) && (retval < 0)) {
	req->result = retval;
    }
    {
        /* XXX: move cr_rstrt_task_complete now that handler re-registration is done */
        if (proc_req->linkage) {
	    if (my_linkage != NULL && my_linkage->cr_task.stopped) {
	        state = TASK_STOPPED;
	    }
        }
        if (retval == 0) {
            /* Return > zero for success restart.  We return the pid of cr_restart, which we want anyway. */
	    set_current_state(state);	/* XXX: move to cr_rstrt_task_complete? */
            retval = req->cr_restart_task->tgid;
        }
    }

    /* done reading the context file - release it back to the requester */
    if (atomic_dec_and_test(&proc_req->final_counter)) {
        req->state = CR_RSTRT_STATE_REQUESTER;
        wake_up(&req->wait);
    }

    /* Wait for the requester and all tasks to reach end-of-file */
    cr_barrier_notify(&req->barrier);
    if (!req->die) {
        cr_barrier_wait(&req->barrier);
    }

    // We can freeze or abort here as a debugging aid.
    if (req->signal) {
	send_sig_info(req->signal, NULL, current);
    }

    if (req->die) { /* set by us or another thread */
	send_sig_info(SIGKILL, NULL, current);
	cr_rstrt_task_complete(cr_task, 0);
    } else if (cr_task->phase == CR_NO_PHASE) {
        // task has no handlers, so finish up now.
	cr_rstrt_task_complete(cr_task, 0);
    } else {
        // cr_rstrt_task_complete() will run from HAND_RSTRT ioctl
    }
out:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

/**
 * cr_rstrt_complete
 * @filp: The file on which the ioctl was received.
 *
 * DESCRIPTION:
 * Implements the HAND_RSTRT ioctl() which indicates to the kernel that
 * the calling task has completed running its RESTART handlers.  Thus
 * we can prevent checkpoint requests from overlapping these handlers.
 * (XXX: need to move stuff out of rstrt_child to get this right.)
 *
 * This is the function that the handlers will call into when
 * the RESTART portion of callbacks are finished.
 *
 */
int cr_rstrt_complete(struct file *filp)
{
    cr_task_t *cr_task;
    int retval = 0;

    CR_KTRACE_FUNC_ENTRY();

    // Lookup this task
    retval = -ESRCH;
    cr_task = cr_task_get(current);
    if (!cr_task) {
	// No matching task found.
        CR_ERR("%s: No matching task found!", __FUNCTION__);
	goto out_noput;
    }

    // Find the matching request
    if (!cr_task->rstrt_proc_req) {
	// No matching request found.
        CR_ERR("%s: No matching request found!", __FUNCTION__);
	goto out;
    }

    // Complete it
    cr_rstrt_task_complete(cr_task, 0);
    retval = 0;

out:
    cr_task_put(cr_task);

out_noput:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

int cr_rstrt_req_release(struct file *filp, cr_pdata_t *priv)
{
    if (priv == NULL) {
        CR_WARN("Warning:  Called with NULL priv!");
        return -EINVAL;
    }

    if (priv->rstrt_req) {
	// Just a simple requester
	release_rstrt_req(priv->rstrt_req);
    } else {
	// If a "restart child" dies unexpectedly, then we need to clean up.
	// However, the release is run in the context of the *last* thread to
	// exit/close/etc, rather then in each - so go looking for them.
	struct list_head *elem, *next;
	write_lock(&cr_task_lock);
	list_for_each_safe(elem, next, &cr_task_list) {
	    cr_task_t *cr_task = list_entry(elem, cr_task_t, task_list);
	    if ((cr_task->filp == filp) && cr_task->rstrt_proc_req) {
		cr_rstrt_task_complete(cr_task, 1);
	    }
	}
	write_unlock(&cr_task_lock);
    }

    return 0;
}

int cr_rstrt_src(struct file *filp, char *ubuf)
{
    cr_task_t *cr_task;
    cr_rstrt_proc_req_t *proc_req;
    cr_rstrt_req_t *req;
    char *buf;
    char *path = NULL;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -ENOMEM;
    buf = (char *) __get_free_page(GFP_KERNEL);
    if (!buf) {
    	goto out_nobuf;
    }

    retval = -ESRCH;
    cr_task = cr_task_get(current);
    if (!cr_task) {
	goto out;
    }
    proc_req = cr_task->rstrt_proc_req;
    cr_task_put(cr_task);
    if (!proc_req) {
	goto out;
    }
    req = proc_req->req;
    if (!req) {
	goto out;
    }

    path = cr_location2path(&(req->src), buf, PAGE_SIZE);
    if (IS_ERR(path)) {
        retval = PTR_ERR(path);
    } else if (path) {
        retval = copy_to_user(ubuf, path, strlen(path));
    } else {
        retval = -EBADF;
    }

out:
    free_page((unsigned long) buf);
out_nobuf:

    return retval;
}
