/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_trigger.c,v 1.23 2005/07/07 21:30:31 phargrov Exp $
 */

#include "cr_module.h"

//
// Private functions
//

// XXX:
// Phase1 and Phase2 could/should deal w/ thread groups (aka pthreads
// processes) more intelligently.
//
// When we support process groups and/or sessions, we _could_ let a
// Phase2 task run as soon as all its own Phase1 siblings are done.
// In fact, if we ever want a Phase1 callback in one process to talk
// to a Phase2 callback in another, then we MUST do it this way.

// do_trigger(task, req)
//
// Sends a trigger (currently just signal 63) to the given task.
// If the task is STOPPED, we also wake it so it can receive the
// signal.
//
// XXX: Should (probably) never fail, but could return -EAGAIN if the
// kernel can't queue anymore RT signals at this time.
// XXX: When we finally move away from the signal, the trigger will
// need to invoke the cr_task->handler, or the default trigger if NULL.
static int
do_trigger(cr_task_t *cr_task, cr_chkpt_req_t *req)
{
	struct task_struct *task = cr_task->task;
	struct siginfo info;
	int retval;

	memset(&info, 0, sizeof(info));	/* security */

	info.si_signo = CR_SIGNUM;
	info.si_errno = 0;
	info.si_code = SI_KERNEL;
	info.si_pid  = req->signal;	/* our 32bits of payload */

	/* Note that we don't recalc_sigpending.  It is not needed, because
	   bad things would have happened if CR_SIGNUM was already pending. */
#if HAVE_2_4_SIGNAL_STRUCT
	spin_lock_irq(&task->sigmask_lock);
	sigdelset(&task->blocked, CR_SIGNUM);
	spin_unlock_irq(&task->sigmask_lock);
#elif HAVE_2_6_SIGNAL_STRUCT
	spin_lock_irq(&task->sighand->siglock);
	sigdelset(&task->blocked, CR_SIGNUM);
	spin_unlock_irq(&task->sighand->siglock);
#endif

	retval = send_sig_info(CR_SIGNUM, &info, task);
	CR_KTRACE_LOW_LVL("triggered pid %d w/ retval=%d", task->pid, retval);

	/* Save the "stoppedness" of the task for later restore */
	cr_task->stopped = (task->state == TASK_STOPPED);
	if (!retval && (cr_task->stopped)) {
		wake_up_process(task);
	}
	return retval;
}

//
// cr_trigger_phase1(req)
//
// Sends a trigger for phase1:
//    All Phase1 tasks (if any) are sent the trigger.
//    If no Phase1 tasks exist then skip to Phase2.
//
// XXX: Should (probably) never fail, thus currently always returns 0.
//
int cr_trigger_phase1(cr_chkpt_req_t *req)
{
	struct list_head *elem;
	cr_task_t *cr_task;
	int found = 0;

	// Send trigger to all of the Phase1 tasks on the list
	read_lock(&req->lock);
	list_for_each(elem, &req->tasks) {
		cr_task = list_entry(elem, cr_task_t, req_list);

		if (cr_task->phase == CR_PHASE1) {
			found = 1;
			do_trigger(cr_task, req);
			// XXX: check error conditions
		}
	}
	read_unlock(&req->lock);

	// If no Phase1 tasks were found then skip to Phase2
	if (!found) {
		cr_trigger_phase2(req);
	}

	return 0;
}

//
// cr_trigger_phase2(req)
//
// Sends the trigger the Phase2 (ie normal/application) tasks.
//
// XXX: Should (probably) never fail, thus currently always returns 0.
//
int cr_trigger_phase2(cr_chkpt_req_t *req)
{
	struct list_head *elem;
	cr_task_t *cr_task;

	// Send trigger to all of the Phase2 (and "undeclared") tasks.
	read_lock(&req->lock);
	list_for_each(elem, &req->tasks) {
		cr_task = list_entry(elem, cr_task_t, req_list);

		if (cr_task->phase != CR_PHASE1) {
			do_trigger(cr_task, req);
			// XXX: check error conditions
		}
	}
	read_unlock(&req->lock);

	return 0;
}
