/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_core.c,v 1.147 2007/02/09 18:06:42 phargrov Exp $
 */

#define _LARGEFILE64_SOURCE 1   /* For O_LARGEFILE */

#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <dlfcn.h>
#include <features.h>

// XXX: needed as long as we do some save/restore work in user space:
#include <termios.h>	// XXX: for termios
#include <sys/param.h>	// XXX: for getcwd() call
#include <sys/time.h>	// XXX: for set/getitimer

#ifndef O_LARGEFILE
  #define O_LARGEFILE 0
#endif

#include "cr_private.h"

//
// Private variables
//

//
// Public variables
//

// How many threads are actively trying to checkpoint (PENDING + ACTIVE)
cri_atomic_t cri_live_count = {0,};

// Red-black lock for critical sections
cri_rb_lock_t cri_cs_lock = CRI_RB_LOCK_INITIALIZER;

//
// Private functions
//

// Tell the kernel we are ready to be saved
// This version just invokes CR_OP_HAND_CHKPT
static int do_checkpoint(int flags)
{
    int rc;
#if CRI_DO_TERMIOS
    struct termios ti;
#endif
    struct itimerval    real_timer;
    struct itimerval    virt_timer;
    struct itimerval    prof_timer;

    // "GLOBAL" CHECKPOINT STUFF HERE:

    // Save the itimers
    getitimer(ITIMER_REAL,    &real_timer);
    getitimer(ITIMER_VIRTUAL, &virt_timer);
    getitimer(ITIMER_PROF,    &prof_timer);

#if CRI_DO_TERMIOS
    // Save the termios structure
    do {
	int tty = __cri_open("/dev/tty", O_WRONLY | O_NOCTTY, 0, NULL);
	
	if (tty >= 0) {
	    rc = tcgetattr(tty, &ti);
	    if (rc < 0) {
		CRI_ABORT("Trouble saving termios");
	    }
	    (void)__cri_close(tty, NULL);
	}
    } while (0);
#endif

    // This does the actual kernel call to take the checkpoint:
    rc = cri_syscall(CR_OP_HAND_CHKPT, flags);
    if (rc < 0) {
	return -errno;
    }

    if (rc) {
	// "GLOBAL" RESTART STUFF HERE

#if CRI_DO_TERMIOS
	// Restore the termios structure
	do {
	    int tty = __cri_open("/dev/tty", O_WRONLY | O_NOCTTY, 0, NULL);
	
	    if (tty >= 0) {
		int err = tcsetattr(tty, TCSANOW, &ti);
		if (err < 0) {
		    CRI_ABORT("Trouble restoring termios");
		}
		(void)__cri_close(tty, NULL);
	    }
	} while (0);
#endif

	// Restore the itimers
	setitimer(ITIMER_REAL,    NULL, &real_timer);
	setitimer(ITIMER_VIRTUAL, NULL, &virt_timer);
	setitimer(ITIMER_PROF,    NULL, &prof_timer);
    } else {
	// "GLOBAL" CONTINUE STUFF HERE
    }

    return rc;
}

static int abort_checkpoint(int flags)
{
    int rc;

    rc = cri_syscall(CR_OP_HAND_ABORT, flags);

    return rc;
}

// enter_idle_state(info)
//
// Transition to CR_STATE_IDLE, regardless of starting state
static inline void enter_idle_state(cri_info_t *info)
{
    cri_atomic_write(&info->cr_state, CR_STATE_IDLE);
    cri_atomic_write(&info->cr_cs_count, 1);
}

// Signal handler to dispatch the request
//
// XXX: This will eventually all be done by the kernel which will
// invoke the proper handler directly w/o using a signal number.
// Must NOT be static because we need to be able to dlsym() it
// when trying to mix static and shared libs.
/* static */ void cri_sig_handler(int signr, siginfo_t *siginfo, void *context)
{
    int saved_errno = errno;
#if LIBCR_TRACING
    int pid = (int)getpid();
#endif
    cri_info_t *info = cri_info_location();	// thread-specific

    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] signal %d received", pid, signr);
    LIBCR_TRACE(LIBCR_TRACE_INFO,
		"checkpoint will deliver signal %d", siginfo->si_pid);

    if (!info || !info->handler) {
	int rc;

	// No state and/or handler, so do this the "easy way"
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] START", pid);
	cri_atomic_inc(&cri_live_count);
	rc = do_checkpoint(0);
	(void)cri_atomic_dec_and_test(&cri_live_count);
	if (rc < 0) {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] do_checkpoint failed", pid);
	}
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] DONE", pid);
    } else {
	// Run the handler appropriate to this thread
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] running handler %p", pid, info->handler);
	info->handler(signr, siginfo, context);
    }

    errno = saved_errno;
}

static void cri_request(int fd, const char *filename, int scope)
{
    /* Static data: */
    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
    static char prev_name[PATH_MAX];

    struct cr_chkpt_req req;
    int rc;

    req.cr_target	= 0;		// 0 = self
    req.cr_scope	= scope;
    req.cr_secs		= 0;		// 0 = unbounded
    req.cr_opts		= NULL;
    req.cr_fd		= fd;
    req.dump_format	= cr_format_vmadump;
    req.signal		= 0;		// 0 = no signal
    req.flags		= 0;

    /* Serialize */
    pthread_mutex_lock(&lock);

    // If there is a previous request un-reaped then reap it now.
    do {
	rc = cri_syscall(CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG);
	if (rc < 0) {
	    if (errno == EINVAL) {
		// No request was pending - not an error.
		break;
	    } else if (errno == EAGAIN) {
		// Wait for the request to complete
		// If interrupted by a signal the loop will try again
		cri_syscall(CR_OP_CHKPT_DONE, (uintptr_t)NULL);
	    } else if (prev_name[0]) {
		// Failed - cleanup file from previous failure
	        LIBCR_TRACE(LIBCR_TRACE_INFO, "Removing supposedly bogus context file "
		            "(%s), because rc=%d and errno=%d (%s)",
		            prev_name, rc, errno, cr_strerror(errno));
	        if (unlink(prev_name) < 0) {
	            CRI_ABORT("unable to remove aborted/failed checkpoint file %s: %s\n", 
		              prev_name, strerror(errno));
	        }
	    }
	}
    } while (rc < 0);

    // Issue the new request
    rc = cri_syscall(CR_OP_CHKPT_REQ, (uintptr_t)&req);
    if (rc) {
	CRI_ABORT("CR_OP_CHKPT_REQ returned %d w/ errno=%d (%s)\n", rc, errno, cr_strerror(errno));
    }

    // Don't return until the request is either pending or completed.
    // This is important; otherwise we would have races with critical
    // sections which follow the cr_request() call.
    //
    // We can detect pending state because the live count is non-zero.
    // We can detect completed because a reap will succeed (or fail
    // with an errno other than EAGAIN).
    // When restarting we expect the reap to fail w/ errno = EINVAL,
    // (if the fd is restored, since there won't be any checkpoint
    // request anymore) or with EBADF if the fd is not restored.
    rc = 0;
    CRI_WHILE_COND_YIELD(
	!cri_atomic_read(&cri_live_count) &&
	((rc = cri_syscall(CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG)) < 0) &&
	errno == EAGAIN
    );
    LIBCR_TRACE(LIBCR_TRACE_INFO, "Reap loop finished with rc=%d and errno=%d (%s)",
		rc, errno, cr_strerror(errno));

    if (filename) {
        /* Remove checkpoint file if an unexpected error occurred (see above) */
        if (rc && (errno != EAGAIN) && (errno != EINVAL) && (errno != EBADF)) {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "Removing supposedly bogus context file "
		        "(%s), because rc=%d and errno=%d (%s)",
		        filename, rc, errno, cr_strerror(errno));
	    if (unlink(filename) < 0) {
	        CRI_ABORT("unable to remove aborted/failed checkpoint file %s: %s\n", 
		          filename, strerror(errno));
	    }
	    prev_name[0] = '\0';
	} else {
	    /* Save name for possible later clean up */
	    strncpy(prev_name, filename, PATH_MAX);
	}
    } else {
	prev_name[0] = '\0';
    }

    pthread_mutex_unlock(&lock);
}

//
// Public functions
//

// Request an application-initiated checkpoint
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
void cr_request_fd(int fd)
{
    if (!cri_info_location()) {
        CRI_ABORT("called w/o first calling cr_init()");
    }

    cri_request(fd, NULL, CR_SCOPE_PROC);
}

// XXX: Needs to go away w/ resilution of bug #77
//
// Request an application-initiated checkpoint of own process
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
// XXX: error handling needs work (ABORT is not the best policy)
void cri_request_file_scoped(const char *filename, int scope)
{
    int fd;

    if (!cri_info_location()) {
        CRI_ABORT("called w/o first calling cr_init()");
    }

    fd = __cri_open(filename, O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE, 0600, NULL);
    if (fd < 0) {
	CRI_ABORT("open(%s)", filename);
    }
    cri_request(fd, filename, scope);
    (void)__cri_close(fd, NULL);	/* silent failure possible */
}

// Request an application-initiated checkpoint of own process
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
// XXX: error handling needs work (ABORT is not the best policy)
void cr_request_file(const char *filename)
{
    cri_request_file_scoped(filename, CR_SCOPE_PROC);
}

// Request an application-initiated checkpoint
// The actual processing might be delayed by critical sections.
//
// Thread-safe, but not reentrant
void cr_request(void)
{
#if 1 // Until VMAdump code deals w/ directory targets properly
#define CONTEXT_NAMELEN 16 /* context.XXXXX + \0 + extra */
    char filename[CONTEXT_NAMELEN];
    snprintf(filename, CONTEXT_NAMELEN, "context.%d", (int)getpid());
    cr_request_file(filename);
#else
    if (!cri_info_location()) {
        CRI_ABORT("called w/o first calling cr_init()");
    }

    cri_request(CR_DEST_CWD, NULL, CR_SCOPE_PROC);
#endif
}

int cr_checkpoint(int flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int my_id;
    int outer_most;
    int index;
    int retval;
    int rc;
#if LIBCR_TRACING
	int pid = (int)getpid();
#endif

    // Check for improper calls
    if (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE) {
	CRI_ABORT("STATE ERROR.  Called while not in the ACTIVE state.");
    }

    // Save callers id
    my_id = info->run.index;
    outer_most = (my_id == info->cr_cb_count);

    // Loop calling callbacks until they have all been called and then
    // invoke do_checkpoint() to do the real work.
    //
    // This strange loop ensures that after the callback with (index == 0)
    // we next call do_checkpoint() even if callback0 is NULL or returns
    // without calling back into cr_checkpoint().
    do {
	if (flags & (CR_CHECKPOINT_TEMP_FAILURE|CR_CHECKPOINT_PERM_FAILURE)) {
	    // We've detected failure of the invoking callback
	    // We we must now call the kernel and stop stacking of callbacks.
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CANCELLING CHECKPOINT (flags=%d)",
			pid, flags);
	    if (abort_checkpoint(flags) == -1)
		info->run.rc = -errno;
	    index = info->run.index = -1;
	    break;
	}
    
	index = --info->run.index;
	if (index >= 0) {
	    // Call the next callback, skipping NULLs

	    if (info->cr_cb[index].func != NULL) {
		info->run.id = index;
		rc = (*info->cr_cb[index].func)(info->cr_cb[index].arg);
		if (rc) {
		    CRI_ABORT("Callback %d returned %d - ABORTING\n",
			      index, rc);
		}
	    }
	} else {
	    LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CHECKPOINTING (flags=%d)",
			pid, flags);
	    info->run.rc = do_checkpoint(flags);
	    if (info->run.rc < 0) {
		LIBCR_TRACE(LIBCR_TRACE_INFO,
			    "[%d] do_checkpoint() FAILED", pid);
	    } else if (info->run.rc) {
		LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] RESTORING", pid);
	    } else {
		LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] CONTINUING", pid);
	    }
	}
    } while (info->run.index >= 0);

    // Once the outer-most invocation calls CR_OP_HAND_{CONT,RSTRT}
    // the kernel might checkpoint us again.  Save the return value now
    // because a second checkpoint could modify info->run.rc.
    // We take this measure even though the final return is currently
    // ignored by cri_start_checkpoint().
    retval = info->run.rc;

    /* Restore saved id for possible use by the caller (in replace_self). */
    info->run.id = my_id;

    // Perform cleanup only on the final/outer-most return
    if (outer_most) {
	enter_idle_state(info);
	cri_atomic_dec_and_test(&cri_live_count);
	if (retval > 0) {
	    rc = cri_syscall(CR_OP_HAND_RSTRT, CRI_SYSCALL_NOARG);
	    if (rc < 0) {
		CRI_ABORT("cri_syscall(CR_OP_HAND_RSTRT) failed w/ rc=%d errno=%d", rc, errno);
	    }
	} else {
	    rc = cri_syscall(CR_OP_HAND_CONT, CRI_SYSCALL_NOARG);
	    if (rc < 0) {
		CRI_ABORT("cri_syscall(CR_OP_HAND_CONT) failed w/ rc=%d errno=%d", rc, errno);
	    }
	}
    }

    return retval;
}

//
// Internal functions
//

// cri_checkpoint_info_init()
//
// Initialize a thread-local struct cr_checkpoint_info
void cri_checkpoint_info_init(cri_info_t *info, siginfo_t *siginfo)
{
    info->cr_checkpoint_info.signal = siginfo->si_pid;
    info->cr_checkpoint_info.dest = NULL;	// filled-in on demand
    info->cr_restart_info.src = NULL;		// filled-in on demand
}

// cri_info_free()
//
// dtor for thread-specific data
void
cri_info_free(void *info)
{
    if (info) {
	free(info);
	pthread_setspecific(cri_info_key, NULL);
    }
}


// cri_info_init()
//
// Routine to initialize the list of registered callbacks and other data
// On first call per thread, allocates the thread-specific cri_info structure.
cri_info_t* cri_info_init(void)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int rc;

    if (!info) {
	info = malloc(sizeof(cri_info_t));
	if (!info) {
	    CRI_ABORT("Failed to allocate cri_info structure");
	}

	memset(info, 0, sizeof(cri_info_t));
	enter_idle_state(info);
	info->handler = NULL;
	info->next_id = 0;

	rc = pthread_setspecific(cri_info_key, info);
	if (rc != 0) {
	    CRI_ABORT("pthread_setspecific() returned %d", rc);
	}
    }

    return info;
}

// cri_start_checkpoint()
//
// Routine to process the list of registered callbacks
void cri_start_checkpoint(cri_info_t *info, int is_thread)
{
    if (cri_cmp_swap(&info->cr_state, CR_STATE_PENDING, CR_STATE_ACTIVE)) {
	// State is valid, so start running callbacks
#if LIBCR_TRACING
	int pid = (int)getpid();
#endif
	info->run.index = info->cr_cb_count;
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] START", pid);
	(void)cr_checkpoint(0);
	LIBCR_TRACE(LIBCR_TRACE_INFO, "[%d] DONE", pid);
    } else if (!is_thread) {
	CRI_ABORT("STATE ERROR.  Probably an excess call to cr_leave_cs().");
    } else {
	// False wake up of a checkpoint thread - not an error.
    }
}

/* Initialization logic.
 *
 * XXX: need a balancing _fini()?
 */
static int raninit = 0;

#ifdef CRI_SA_RESTORER
  CRI_SA_RESTORER
#endif

static void _do_init(void *replace)
{
    int rc;
    struct k_sigaction	ksa;
    int signum;

    //
    // Initialize tracing
    //
    libcr_trace_init();

    //
    // Initialize pthread-dependent parts if appropriate
    //
    cri_pthread_init();

    //
    // Setup signal handler
    //
    if (replace) {
	/* Check that we're registered already and the reregister */
	rc = __cri_ksigaction(CR_SIGNUM, NULL, &ksa, (_NSIG/8), &errno);
	if (ksa.ksa_sigaction != replace) {
	    CRI_ABORT("Failed to reregister signal %d in process %d from static constructor",
		      CR_SIGNUM, (int)getpid());
	}
    } else if (CR_SIGNUM != (signum = __libc_allocate_rtsig(0))) {
	CRI_ABORT("Failed to allocate signal %d in process %d: got signal %d instead",
		  CR_SIGNUM, (int)getpid(), signum);
    }
    ksa.ksa_sigaction = cri_sig_handler;
    ksa.ksa_flags = SA_RESTART | SA_SIGINFO;
    #ifdef CRI_SA_RESTORER
      ksa.ksa_flags |= SA_RESTORER;
      ksa.ksa_restorer = &cri_sa_restorer;
    #endif
    rc = sigfillset(&ksa.ksa_mask);
    if (rc != 0) {
	CRI_ABORT("sigfillset() failed: %s", strerror(errno));
    }
    rc = __cri_ksigaction(CR_SIGNUM, &ksa, NULL, (_NSIG/8), &errno);
    if (rc != 0) {
	CRI_ABORT("sigaction() failed: %s", strerror(errno));
    }
}

/* Initialization entry point.
 *
 * If shared lib, compiled with -DPIC, and use magic '_init' function, which
 * is always loaded just once.
 *
 * If compiled statically, use the 'constructor' attribute to run at
 * startup.
 *
 * TODO:  This doesn't completely solve the problem:  
 *  1)  When statically linked, if no other functions from this library used,
 *  linker won't link the library into the app, and so constructor not run.  
 *	If linking libcr and not calling any functions, then you get what you
 *	deserve.  However, since every client of libcr must call cr_init(),
 *	we need only ensure cr_init() will reference something in cr_core.o.
 *	In fact, cr_init() calls cri_info_init() which is in cr_core.o.
 *	Thus we are certain that any client that actually uses the library
 *	will include the constructor.  This issue is solved.
 *  2) If statically linked AND started with cr_run, init function will be run
 *  twice (and with dynamic lib having its own copy of all data structures).
 *	This is now (almost) solved by ensuring that only the statically linked
 *	version is ever run.  This works because the shared version will
 *	run its initializer first and the static one will overwrite the
 *	signal handler with the one in the static library.  Thus nobody
 *	will ever invoke code in the shared library once the static lib's
 *	constructor has run.  Since this is certain to happen before main(),
 *	we are certain that in the unlikely case that a checkpoint is
 *	requested before the static lib's constructor runs, no callbacks
 *	will be missed. XXX: untested
 *  3) We are unable to interpose on pthread_create, fork or vfork.
 *
 *  For now we still build only shared libs.
 */
#ifdef PIC
void _init(void)
{
    if (!raninit++)
	_do_init(NULL);
}
#else
static void __attribute__((constructor)) _cr_static_init(void)
{
    if (!raninit++) {
	// dlsym(NULL, ) segfaults for a static executable.
	// Use dlopen first to test that first.
	void *handle = dlopen(NULL, RTLD_LAZY);
	if (handle) {
	    _do_init(dlsym(handle, "cri_sig_handler"));
	    dlclose(handle);
	} else {
	    _do_init(NULL);
	}
    }
}
#endif
