/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: lampd.c,v 1.44 2005/01/15 04:51:56 phargrov Exp $
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <time.h>
#include <libgen.h>
#include <libcr.h>
#include <sys/socket.h>
#include <sys/un.h>
#ifdef HAVE_CONFIG_H
  #include "blcr_config.h"
#endif

#define LAMBOOT_HOSTFILE_STEM   "/tmp/lampd_hosts."
#define CR_DIR_PATTERN		"/tmp/lampd_%s_%d"

#define PMI_HOSTKEY	"hostname:%d"
#define PMI_CMD		"command"
#define PMI_ARG0	"arg:0"
#define PMI_ARG1	"arg:1"

#ifndef LAM_BIN
  #ifdef LAM_HOME
    #define LAM_BIN LAM_HOME "/bin"
  #else
    #error Must #define LAM_HOME or LAM_BIN to locate LAM/MPI executables
  #endif
#endif
#ifndef CR_BIN
  #ifdef CR_HOME
    #define CR_BIN CR_HOME "/bin"
  #else
    #error Must #define CR_HOME or CR_BIN to locate BLCR executables
  #endif
#endif

enum {
  cmd_hostname,
  cmd_crdir,
  cmd_exit,
};

/* Protypes for the PMI functions: */
#if HAVE_PMI_H
  #include <pmi.h>
#else
  /* prototypes for PMI (Process Management Interface) */

  extern int PMI_Init(int *spawned);
  extern int PMI_Initialized(int *initialized);
  extern int PMI_Finalize(void);
  extern int PMI_Barrier(void);

  extern int PMI_Get_size(int *size);
  extern int PMI_Get_rank(int *rank);
  extern int PMI_KVS_Get_my_name(char kvsname[], int length);
  extern int PMI_KVS_Get_name_length_max(int *length);
  extern int PMI_KVS_Get_key_length_max(int *length);
  extern int PMI_KVS_Get_value_length_max(int *length);

  extern int PMI_KVS_Get(const char kvsname[], const char key[], char value[], int length);
  extern int PMI_KVS_Put(const char kvsname[], const char key[], const char value[]);
  extern int PMI_KVS_Commit(const char kvsname[]);
#endif

/* globals: */
static int debug = 1;
static int is_restart = 0;
static char *debug_filename = NULL;
static FILE *debug_file = NULL;
static int debug_fd = -1;
static const char *argv0;
static volatile pid_t child = 0;
static pid_t my_pid = 0;
static int rank = -1;
static int size = -1;
static char *cr_base_dir = NULL;
static char *unique = NULL;
static char *hostnames = NULL;
static char *context = NULL;
static volatile int watcher_pid = -1;
static int fd_max;
static sigset_t sigmask;
static char * kvsname;
static int valuelen;
static int namelen; 
static int keylen;
static int exit_signal;
static cr_client_id_t my_crid;
static char *myname;

#define debug_printf(ARGS...) \
  if (debug) {\
    fprintf(debug_file, ARGS);\
    fflush(debug_file);\
  }

#define debug_system(ARGS...) \
  if (debug) {\
    char *_cmd = sappendf(sappendf(NULL, ARGS), " 1>&%d 2>&1", debug_fd); \
    fprintf(debug_file, "system('%s'):\n", _cmd);\
    fcntl(debug_fd, F_SETFD, 0); \
    system(_cmd); \
    fcntl(debug_fd, F_SETFD, FD_CLOEXEC); \
    fflush(debug_file);\
    free(_cmd); \
  }

char *sappendf(char *s, const char *fmt, ...)
{
        va_list args;
	int old_len, add_len;
                                                                                                             
	/* compute length of thing to append */
        va_start(args, fmt);
	add_len = vsnprintf(NULL, 0, fmt, args);
        va_end(args);

	/* grow the string, including space for '\0': */
	if (s) {
		old_len = strlen(s);
	    	s = realloc(s, old_len + add_len + 1);
	} else {
		old_len = 0;
		s = malloc(add_len + 1);
	}

	/* append */
        va_start(args, fmt);
	vsprintf((s+old_len), fmt, args);
        va_end(args);

	return s;
}

void fatal(const char *fmt, ...)
{
        va_list args;
	char *progname;
	int is_initd;
	PMI_Initialized(&is_initd);

	progname = strdup(argv0);
	fprintf(stderr, "%s: ", basename(progname));
	free(progname);

	if (fmt != NULL) {
        	va_start(args, fmt);
		vfprintf(stderr, fmt, args);
        	va_end(args);
	}
	
	if (is_initd) {
    		(void)PMI_Finalize();
	}

	exit(-1);
}

void usage(void) {
    fprintf(stderr, "usage: %s LAM_MPI_PROGRAM [args...]\n", argv0);
}

void snd_cmd(int cmd, const char *arg0, const char *arg1)
{
    char *cmd_string = alloca(valuelen+1);
    int rc;


    snprintf(cmd_string, valuelen, "%d", cmd);
    rc = PMI_KVS_Put(kvsname, PMI_CMD, cmd_string);
    if (rc != 0) {
	goto error_out;
    }
    if (arg0) {
        rc = PMI_KVS_Put(kvsname, PMI_ARG0, arg0);
	if (rc != 0) {
	    goto error_out;
	}
    }
    if (arg1) {
        rc = PMI_KVS_Put(kvsname, PMI_ARG1, arg1);
	if (rc != 0) {
	    goto error_out;
	}
    }
    rc = PMI_KVS_Commit(kvsname);
    if (rc != 0) {
	goto error_out;
    }

    debug_printf("Sending cmd %d\n", cmd);
    rc = PMI_Barrier();
    if (rc != 0) {
	goto error_out;
    }
    debug_printf("Cmd sent\n");
    rc = PMI_Barrier();
    if (rc != 0) {
	goto error_out;
    }
    debug_printf("Cmd acknowledged\n");
    return;

error_out:
    fatal("lost connection to PMI\n");
    return;
}

int rcv_cmd(char *arg0, char *arg1)
{
    char cmd_string[8];	/* XXX size ? */
    int rc;

    debug_printf("Entering cmd barrier\n");
    rc = PMI_Barrier();
    if (rc != 0) {
	goto error_out;
    }
    debug_printf("Left cmd barrier\n");
    	
    rc = PMI_KVS_Get(kvsname, PMI_CMD, cmd_string, valuelen+1);
    rc = PMI_KVS_Get(kvsname, PMI_ARG0, arg0, valuelen+1);
    rc = PMI_KVS_Get(kvsname, PMI_ARG1, arg1, valuelen+1);

    debug_printf("Recvd cmd = %d\n", atoi(cmd_string));
    return atoi(cmd_string);

error_out:
    fatal("lost connection to PMI\n");
    return -1;
}

void ack_cmd(void)
{
    int rc;

    debug_printf("Entering ACK barrier\n");
    rc = PMI_Barrier();
    if (rc != 0) {
	fatal("lost connection to PMI\n");
    }
    debug_printf("Left ACK barrier\n");
}

void put_hostname(void)
{
    char *key = alloca(keylen+1);
    int rc;

    myname = malloc(valuelen+1);
    gethostname(myname, valuelen);
    snprintf(key, keylen, PMI_HOSTKEY, rank);
    rc = PMI_KVS_Put(kvsname, key, myname);
}

static void my_exit(int status)
{
    (void)PMI_Finalize();

    /* try to exit such that our exit code is same as our child */
    if (WIFEXITED(status)) {
        /* easy to mimic normal return */
        exit(WEXITSTATUS(status));
    } else if (WIFSIGNALED(status)) {
        /* disable generation of a 'core' */
        int sig = WTERMSIG(status);
        struct rlimit r;

        r.rlim_cur = r.rlim_max = 0;
        (void)setrlimit(RLIMIT_CORE, &r);
                                                                                                             
        /* now raise the signal */
        signal(sig, SIG_DFL);
	raise(sig);
	fatal("unexpectedly survived signal %d\n", sig);
    } else {
        fatal("Unexpected status %d from child\n", status);
    }
}


static void sigforward(int sig)
{
	signal(sig, SIG_IGN);
	if (child > 0) {
		debug_printf("forwarding signal %d to child %d\n", sig, child);
		kill(child, sig);
	}
}

static void lam_watcher_obituary(void) {
    debug_printf("lam_watcher: exiting unexpectedly\n");
}

static void lam_watcher(int ctrl_fd)
{
    int i, rc, fd;
    char *path;
    char *cmd;
    char *nodefile;

    /* Log our death if other than by exec() */
    atexit(&lam_watcher_obituary);

    /* redirect output, close almost everything else */
    dup2(debug_fd, STDERR_FILENO);
    dup2(debug_fd, STDOUT_FILENO);
    close(STDIN_FILENO);
    for (i = 3; i < fd_max; ++i) {
	if (i != ctrl_fd && i != debug_fd) {
	    close(i);
	}
    }
    fcntl(ctrl_fd, F_SETFD, FD_CLOEXEC);	/* redundant? */
    fcntl(debug_fd, F_SETFD, FD_CLOEXEC);	/* redundant? */

    /* private process group */
    setpgid(0,0);

    /* Create hosts file for LAM */
    nodefile = sappendf(NULL, LAMBOOT_HOSTFILE_STEM "%s", unique);
    fd = open(nodefile, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    if (fd < 0) {
	fatal("lamwatcher: unable to create '%s': %s\n", nodefile, strerror(errno));
    }
    write(fd, hostnames, strlen(hostnames));
    close(fd);

    /* setup LAM path */
    path = sappendf(NULL, LAM_BIN ":/bin:/usr/bin:%s", getenv("PATH"));
    setenv("PATH", path, 1);
    free(path);

    /* start up LAM daemons */
    if (debug) {
        cmd = sappendf(NULL, LAM_BIN "/lamboot -v %s", nodefile);
    } else {
        cmd = sappendf(NULL, LAM_BIN "/lamboot %s", nodefile);
    }
    debug_printf("lamwatcher: Going to run lamboot now...\n");
    rc = system(cmd);
    free(cmd);

    /* remove hosts file */
    (void)unlink(nodefile);	/* may fail silently */
    free(nodefile);

    /* "return" our result.
     * Can ignore write() failures since subsequent read() will also fail. */
    (void)write(ctrl_fd, &rc, sizeof(int));
    if (rc != 0) {
	debug_printf("lamwatcher: lamboot returned %d\n", rc);
	exit(rc);
    }

    /* wait for parent to close or die */
    debug_printf("lamwatcher: waiting for parent lampd\n");
    do {
	char dummy;
	rc = read(ctrl_fd, &dummy, 1);
    } while ((rc < 0) && (errno == EINTR));
    debug_printf("lamwatcher: parent lampd closed/exited\n");

    /* Use lamds to help cleanup the symlink */
    cmd = sappendf(NULL, LAM_BIN "/lamexec N rm -f %s", cr_base_dir);
    debug_printf("lamwatcher: Going to run system(%s)...\n", cmd);
    rc = system(cmd);
    free(cmd);
    
    /* kill lamds */
    debug_printf("lamwatcher: Going to exec lamhalt now...\n");
    (void)execl(LAM_BIN "/lamhalt", "lamhalt", NULL);
    exit(-1);
}


static int run_lamboot(void)
{
    int fds[2];
    int result;
    int rc;

    /* Fork a child do the work */
    rc = socketpair(PF_LOCAL, SOCK_STREAM, 0, fds);
    if (rc < 0) {
	perror("socketpair()");
	return(-1);
    }
    (void)fcntl(fds[0], F_SETFD, FD_CLOEXEC);
    (void)fcntl(fds[1], F_SETFD, FD_CLOEXEC);
    watcher_pid = fork();
    if (watcher_pid == 0) {
	/* I am the child, launch stuff */
	rc = close(fds[0]);
	if (rc < 0) {
	    perror("close(socketpair/child)");
	    exit(1);
	}
	lam_watcher(fds[1]);
    } else if (watcher_pid > 0) {
    	rc = close(fds[1]);
	if (rc < 0) {
	    perror("close(socketpair/parent)");
	    return(-1);
	}
    } else {
	perror("fork()");
	exit(1);
    }

    /* wait for a result from our child */
    do {
      rc = read(fds[0], &result, sizeof(int));
    } while ((rc < 0) && (errno == EINTR));
    if (rc < 0) {
	perror("read(pipe)");
	return(-1);
    }

    return result;
}

void run_checkpoint(const char *dir) {
    char *cmd;
    const struct cr_checkpoint_info *info;
    int rc;

    info = cr_get_checkpoint_info();
    context = sappendf(NULL, "%s/context.%d", dir, child);
    cmd = sappendf(NULL, CR_BIN "/cr_checkpoint --file=%s", context);
    if (info->signal) {
        cmd = sappendf(cmd, " --signal=%d", info->signal);
    }
    cmd = sappendf(cmd, " --pid %d", child);

    debug_printf("about to run '%s'\n", cmd);
    rc = system(cmd);
    free(cmd);

    debug_printf("got rc/errno %d/%d from checkpoint\n", rc, errno);
}

void run_mpirun(int argc, char **argv)
{
    char **args;
    int i;

    args = (char **)calloc(argc + 3, sizeof(char *));
    args[0] = "mpirun";
    args[1] = "-np";
    args[2] = sappendf(NULL, "%d", size);
    debug_printf("Going to run '%s %s %s", args[0], args[1], args[2]);
    for (i = 1; i < argc; i++) {
	args[i+2] = argv[i];
	debug_printf(" %s", argv[i]);
    }
    args[argc + 2] = NULL;
    debug_printf("'\n");

    child = fork();
    if (child == 0) {
	for (i = 3; i < fd_max; ++i) (void)close(i);
	(void)execvp(LAM_BIN "/mpirun", args);
	fatal("execvp(mpirun) failed: %s\n", strerror(errno));
    } else if (child < 0) {
	fatal("fork() failed: %s\n", strerror(errno));
    } else {
	debug_printf("created child (mpirun) w/ pid=%d\n", child);
    }
    free(args[2]);
    free(args);
}

void run_restart(void)
{
    debug_printf("gonna run '" CR_BIN "/cr_restart %s'\n", context);

    child = fork();
    if (child == 0) {
	int i;
	for (i = 3; i < fd_max; ++i) (void)close(i);
	(void)execl(CR_BIN "/cr_restart", "cr_restart", context, NULL);
	fatal("execl(cr_restart) failed: %s\n", strerror(errno));
    } else if (child < 0) {
	fatal("fork() failed: %s\n", strerror(errno));
    }

    debug_printf("spawned replacement child (cr_restart) w/ pid=%d\n", child);
}

void prep_crdir(const char *base_dir, const char *dest_dir)
{
    int rc;

    debug_printf("symlink(%s, %s)\n", dest_dir, base_dir);
    rc = symlink(dest_dir, base_dir);
}

int my_signal_callback(void *arg)
{
    int rc;
    const struct cr_checkpoint_info *info = cr_get_checkpoint_info();
    void (*saved_handler)(int);

    exit_signal = info->signal;
    if (exit_signal) {
	saved_handler = signal(exit_signal, SIG_IGN);
    }

    rc = cr_checkpoint(CR_CHECKPOINT_READY);

    if (rc < 0) {
	fatal("checkpoint failed: %s\n", strerror(errno));
    } else if (rc) {
        /* restore original signal mask */
        sigprocmask(SIG_SETMASK, &sigmask, NULL);

	/* exec to RESTART from "context" */
	(void)execl("/proc/self/exe", argv0, "--restart", context, NULL);
	fatal("execl(self) failed: %s\n", strerror(errno));
    } else {
	/* CONTINUE */
	debug_printf("CONTINUE\n");
    }

    return 0;
}

int my_thread_callback(void *arg)
{
    int rc;
    const struct cr_checkpoint_info *info = cr_get_checkpoint_info();
    char *tmp = strdup(info->dest);
    char *dir = dirname(tmp);

    /* Ensure that LAM/MPI can see checkpoint dir at a "well known" location */
    prep_crdir(cr_base_dir, dir);
    snd_cmd(cmd_crdir, cr_base_dir, dir);

    /* CHECKPOINT our child if running */
    if (child > 0) {
	run_checkpoint(dir);
    }

    free(tmp);

    rc = cr_checkpoint(CR_CHECKPOINT_READY);
    if (rc < 0) {
	fatal("checkpoint failed: %s\n", strerror(errno));
    } else if (rc) {
	/* RESTART */
    } else {
	/* CONTINUE */
        if (context) {
            free(context);
            context = NULL;
        }
    }

    return 0;
}

void master(int argc, char **argv)
{
    char *value = alloca(valuelen+1);
    char *key = alloca(keylen+1);
    int status = 0;
    int i, rc;

    rc = cr_register_callback(&my_thread_callback, NULL, CR_THREAD_CONTEXT);
    if (rc < 0) {
	fatal("cr_register_callback() failed\n");
    }

    /* get all hostnames */
    put_hostname();
    snd_cmd(cmd_hostname, NULL, NULL);
    hostnames = NULL;
    for (i=0; i<size; ++i) {
	snprintf(key, keylen, PMI_HOSTKEY, i);
	rc = PMI_KVS_Get(kvsname, key, value, valuelen+1);
	hostnames=sappendf(hostnames, "%s\n", value);
    }

    /* get something unique for this session */
    if (unique == NULL) {
	unique = malloc(namelen);
	PMI_KVS_Get_my_name(unique, namelen);
    }

    cr_base_dir = sappendf(NULL, CR_DIR_PATTERN, myname, (int)my_pid);

    if (!is_restart) {
	    /* NORMAL CASE, run mpirun */
	    putenv(sappendf(NULL, "LAM_MPI_SESSION_SUFFIX=%s", unique));
	    putenv(sappendf(NULL, "LAM_MPI_SSI_cr_base_dir=%s", cr_base_dir));
	    run_lamboot();
	    run_mpirun(argc, argv);
    } else {
	    /* RESTART case */
	    char *tmp, *dir;

	    context = argv[2];

	    tmp = strdup(context);
	    dir = dirname(tmp);
	    prep_crdir(cr_base_dir, dir);
	    snd_cmd(cmd_crdir, cr_base_dir, dir);
	    free(tmp);

	    run_lamboot();
	    run_restart();
    }

    signal(SIGTERM, &sigforward);
    signal(SIGINT, &sigforward);
    do {
	    rc = waitpid(-1, &status, 0);
    } while((rc == watcher_pid) ||
		((rc < 0) && ((errno == EINTR) || (errno == ECHILD))));
    debug_printf("child %d exited with status/signal %d/%d\n", rc, WEXITSTATUS(status),WTERMSIG(status));
    signal(SIGTERM, SIG_DFL);
    signal(SIGINT, SIG_DFL);

    if (WIFSIGNALED(status) && (WTERMSIG(status) == exit_signal)) {
	status = 0;
    }

    snprintf(value, valuelen, "%d", status);
    snd_cmd(cmd_exit, value, NULL);

    my_exit(status);
}

void slave(int argc, char **argv)
{
    char *arg0 = alloca(valuelen+1);
    char *arg1 = alloca(valuelen+1);
    int cmd;

    do {
	cmd = rcv_cmd(arg0, arg1);

	switch (cmd) {
	  case cmd_exit:
	    ack_cmd();
	    my_exit(atoi(arg0));
	    break;

	  case cmd_hostname:
	    put_hostname();
	    ack_cmd();
	    break;

	  case cmd_crdir:
	    if (cr_base_dir) {
		free(cr_base_dir);
	    }
	    cr_base_dir = strdup(arg0);
	    prep_crdir(arg0, arg1);
	    ack_cmd();
	    break;
	}
    } while (1);	
}

int main(int argc, char **argv)
{
    int rc;
    int inited;
    sigset_t nullset;
		
    my_pid = getpid();

    sigemptyset(&nullset);
    sigprocmask(SIG_BLOCK, &nullset, &sigmask); /* saves original mask */
    argv0 = argv[0];
#if 0
    fd_max = sysconf(_SC_OPEN_MAX);
#else
    fd_max = 32;
#endif

    /* Check args */
    if (argc < 2) {
      usage();
      fatal("missing program argument\n");
    } else if (strcmp("--help", argv[1]) == 0) { 
      usage();
      exit(0);
    } else if (strcmp("--version", argv[1]) == 0) { 
      printf("lampd $Revision: 1.44 $\n");
      exit(0);
    }
    is_restart = (strcmp("--restart", argv[1]) == 0);

    /* Get basic data from MPD */
    rc = PMI_Init (&inited);
    if (rc != 0) {
      fatal("unable to connect to a PMI implementation such as mpd\n");
    }
    rc = PMI_Get_rank (&rank);
    if (rc != 0) {
      fatal("unable to get my rank from the PMI implementation\n");
    }
    rc = PMI_Get_size(&size);
    if (rc != 0) {
      fatal("unable to get job size from the PMI implementation\n");
    }
    PMI_KVS_Get_name_length_max(&namelen);
    PMI_KVS_Get_key_length_max(&keylen);
    PMI_KVS_Get_value_length_max(&valuelen);
    kvsname = malloc(1+namelen);
    rc = PMI_KVS_Get_my_name(kvsname, 1+namelen);
    if (!strcmp("mykvs", kvsname)) {
	fatal("You do not appear to be running under MPICH's mpdrun/mpirun (KVS name='mykvs')\n");
    }
    if (rc != 0) {
      fatal("unable to get namespace from the PMI implementation\n");
    }

    /* Create a debugging output file */
    if (debug) {
	debug_filename = sappendf(NULL, "/tmp/lampd_debug.%s-%d", kvsname, rank);
	debug_file = fopen(debug_filename, "a");
	if (debug_file == NULL) {
	  fatal("unable to open log file '%s' for writing : %s\n", 
		debug_filename, strerror(errno));
	}
	setlinebuf(debug_file);
	debug_fd = fileno(debug_file);
	fcntl(debug_fd, F_SETFD, FD_CLOEXEC);
	/* XXX need to rm these debug files eventually */
    }

    debug_printf("rank=%d of %d\n", rank, size);
    /* Initialize cr */
    my_crid = cr_init();
    if (my_crid < 0) {
	fatal("cr_init() failed\n");
    }
    rc = cr_register_callback(&my_signal_callback, NULL, CR_SIGNAL_CONTEXT);
    if (rc < 0) {
	fatal("cr_register_callback() failed\n");
    }

    if (rank == 0) {
       master(argc, argv);
    } else {
	slave(argc, argv);
    }

    /* NOTREACHED */

    return -1;
}
