/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2007, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: child.c,v 1.2 2007/02/21 20:00:30 phargrov Exp $
 *
 * Test of pipes w/ child and grandchild caputured via CR_SCOPE_TREE
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>

#include "crut.h"
#include "crut_util.h"

enum {
	MSG_CHILD_GOOD,
	MSG_CHILD_BAD,
	MSG_PARENT_READY,
	MSG_PARENT_DONE = 42
};

static int fds1[2], fds2[2];
static int inpipe, outpipe;
static int parent, child;

static void do_read(unsigned char expect) {
    char c;
    int rc;

    do { rc = read(inpipe, &c, 1); } while ((rc < 0) && (errno == EINTR));
    if ((rc != 1) || (c !=expect)) {
	CRUT_DEBUG("rc=%d errno=%d expect=0x%x got=0x%x\n",
		   rc, errno, (unsigned int)expect, (unsigned int)c);
	exit((rc == 1) ? c : -1);
    }
}

static void do_write(unsigned char c) {
    int rc;
    do { rc = write(outpipe, &c, 1); } while ((rc < 0) && (errno == EINTR));
    if (rc != 1) {
	CRUT_DEBUG("rc=%d errno=%d val=0x%x\n", rc, errno, (unsigned int)c);
	exit(-1);
    }
}

static void sigpipe(int signo) {
    fprintf(stderr, "%d exiting on SIGPIPE\n", getpid());
    exit(-1);
}

static void *thread_main(void *arg) 
{
    while(1) { pthread_testcancel(); sleep(1); }
    return NULL;
}

static int
child_setup(void **p)
{
    int retval;

    signal(SIGPIPE, &sigpipe);	

    CRUT_DEBUG("Creating pipes.");
    retval = pipe(fds1);
    if (retval < 0) {
	CRUT_FAIL("Error initializing pipes.");
	goto out;
    }
    retval = pipe(fds2);
    if (retval < 0) {
	CRUT_FAIL("Error initializing pipes.");
	goto out;
    }

    CRUT_DEBUG("Forking child.");
    child = fork();
    if (child < 0) {
	CRUT_FAIL("Error forking child.");
	retval = child;
	goto out;
    } else if (!child) {
	/* In the child */
	parent = getpid();
	child = fork();
        if (child < 0) {
	    CRUT_FAIL("Error forking grandchild.");
	} else if (!child) {
	    /* In the grandchild */
	    pthread_t th;
	    inpipe  = fds1[0]; (void)close(fds1[1]);
            outpipe = fds2[1]; (void)close(fds2[0]);
	    if (pthread_create(&th, NULL, &thread_main, NULL) != 0) {
		exit(1);
	    }
	    while (1) {
	        CRUT_DEBUG("Grandchild entering read()");
	        do_read(MSG_PARENT_READY);
	        CRUT_DEBUG("Grandchild sees ppid %d", getppid());
	        do_write((getppid() == parent) ? MSG_CHILD_GOOD : MSG_CHILD_BAD);
	    }
	    exit(-1); // unreached
	} else {
	    int status;
            do {
	        retval = waitpid(child, &status, 0);
            } while ((retval < 0) && (errno == EINTR)); /* retry on non-fatal signals */
            if ((retval != child) || !WIFEXITED(status) || (WEXITSTATUS(status) != MSG_PARENT_DONE)) {
	        CRUT_FAIL("Error running grandchild.");
            } else {
	        CRUT_DEBUG("Grandchild done.");
	    }
            exit(WIFEXITED(status) ? WEXITSTATUS(status) : -1);
	}
	exit(-1); // unreached
    } else {
	/* In the parent */
	CRUT_DEBUG("Forked child %d", child);
	inpipe  = fds2[0]; (void)close(fds2[1]);
	outpipe = fds1[1]; (void)close(fds1[0]);
    }

    do_write(MSG_PARENT_READY);
    do_read(MSG_CHILD_GOOD);

    *p = NULL;

out:
    return retval;
}

static int
child_precheckpoint(void *p)
{
    int retval = 0;

    CRUT_DEBUG("Getting ready to checkpoint");

    do_write(MSG_PARENT_READY);
    do_read(MSG_CHILD_GOOD);

    return retval;
}

static int teardown(void)
{
    int status, retval;

    do_write(MSG_PARENT_DONE);
    do {
	retval = waitpid(child, &status, 0);
    } while ((retval < 0) && (errno == EINTR)); /* retry on non-fatal signals */
    if ((retval != child) || !WIFEXITED(status) || (WEXITSTATUS(status) != MSG_PARENT_DONE)) {
	CRUT_FAIL("Error running child.");
    } else {
	retval = 0;
    }
    CRUT_DEBUG("Child %d exited", child);

    return retval;
}

static int
child_continue(void *p)
{
    int retval = 0;

    CRUT_DEBUG("Continuing after checkpoint.");

    do_write(MSG_PARENT_READY);
    do_read(MSG_CHILD_GOOD);

    retval = teardown();

    return retval;
}

static int
child_restart(void *p)
{
    int retval = 0;

    CRUT_DEBUG("Restarting from checkpoint.");

    do_write(MSG_PARENT_READY);
    do_read(MSG_CHILD_GOOD);

    return retval;
}

static int
child_teardown(void *p)
{
    int retval = 0;

    CRUT_DEBUG("Entering child_teardown.");

    retval = teardown();

    return retval;
}

int
main(int argc, char *argv[])
{
    int ret;
    struct crut_operations child_test_ops = {
	test_name:"child",
        test_description:"Child tester.  Tests checkpoint of children via CR_SCOPE_TREE.",
	test_setup:child_setup,
	test_precheckpoint:child_precheckpoint,
	test_continue:child_continue,
	test_restart:child_restart,
	test_teardown:child_teardown,
    };

    /* add the test */
    crut_add_test(&child_test_ops);

    ret = crut_main(CR_SCOPE_TREE, argc, argv);

    return ret;
}
