/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_checkpoint.c,v 1.55 2007/01/31 06:05:45 phargrov Exp $
 */

#define _LARGEFILE64_SOURCE 1   /* For O_LARGEFILE */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <libgen.h>
#include <dirent.h>
#include <getopt.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <assert.h>

#ifndef O_LARGEFILE
  #define O_LARGEFILE 0
#endif

#include "libcr.h"

/* can't use argv[0] to get name, since libtool screws it up */
#define MY_NAME	"cr_checkpoint"

static int verbose = 0;

static char *to_remove;

static void die(int code, const char *format, ...)
		__attribute__ ((noreturn, format (printf, 2, 3)));
static void print_err(const char *format, ...)
		__attribute__ ((format (printf, 1, 2)));

static void usage(FILE *stream)
{
    fprintf(stream, 
"Usage: " MY_NAME " [options] ID\n"
"\n"
"General options:\n"
"  -v, --verbose          print progress messages to stderr.\n"
"  -?, --help             print this message and exit.\n"
"      --version          print version information and exit.\n"
"\n"
"Options for scope of the checkpoint:\n"
"  -p, --pid, --process   ID identifies a process id (default).\n"
"  -T, --tree             ID identifies a process id.  It and all\n"
"                         of its descendants are to be checkpointed.\n"
"  -g, --pgid, --group    ID identifies a process group id.\n"
"  -s, --sid, --session   ID identifies a session id.\n"
"\n"
"Options for destination location of the checkpoint:\n"
"  -c, --cwd              checkpoint saved as a single 'context.ID' file in\n"
"                         cr_checkpoint's working directory (default).\n"
"  -d, --dir DIR          checkpoint saved in new directory DIR, with one\n"
"                         'context.ID' file per process (unimplemented).\n"
"  -f, --file FILE        checkpoint saved as FILE.\n"
"  -F, --fd FD            checkpoint written to an open file descriptor.\n"
"\n"
"Options for creation/replacement policy for checkpoint files:\n"
"      --atomic           checkpoint created/replaced atomically (default).\n"
"      --backup[=NAME]    checkpoint created atomically, and any existing \n"
"                         checkpoint backed up to NAME or *.~1~, *.~2~, etc.\n"
"      --clobber          checkpoint written incrementally to target, \n"
"                         overwriting any pre-existing checkpoint.\n"
"  These options are ignored if the destination is a file descriptor.\n"
"\n"
"Options for signal sent to process(es) after checkpoint:\n"
"      --run              no signal sent: continue execution (default).\n"
"  -S, --signal NUM       signal NUM sent to all processes/threads.\n"
"      --stop             SIGSTOP sent to all processes/threads.\n"
"      --term             SIGTERM sent to all processes/threads.\n"
/* "      --kill             SIGKILL sent to all processes/threads.\n" */
"      --abort            SIGABRT sent to all processes/threads.\n"
"\n"
"Options for file system synchronization (default is --sync):\n"
"      --sync             fsync checkpoint file(s) to disk (default).\n"
"      --nosync           do not fsync checkpoint file(s) to disk.\n"
"\n"
"Misc Options:\n"
"  -t, --time SEC         allow only SEC seconds for target to complete\n"
"                         checkpoint (default: wait indefinitely).\n"
    );
}

static void print_version()
{
    printf(MY_NAME " version %s\n", CR_RELEASE_VERSION);
}

/*
 * die(code, format, args...)
 *
 * Sends an error message to stderr and exits program with indicated exit code.
 */
static void die(int code, const char *format, ...)
{
    va_list args;
    struct stat s;

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    if (to_remove != NULL) 
	if (!stat(to_remove, &s)) 
	    unlink(to_remove);

    exit(code);
}

/*
 * print_err(format, args...)
 *
 * Works like fprintf(stderr, format, args...), but preserves errno.
 * No return value, since used only on error paths.
 */
static void print_err(const char *format, ...)
{
    int saved_errno = errno;
    va_list args;

    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);

    errno = saved_errno;
}



static int
readint(const char *arg, const char *argv0)
{
    int val;
    char *endptr;

    val = strtol(arg, &endptr, 10);
    if ((*arg == '\0') || (*endptr != '\0')) {
	usage(stderr);
	exit(-1);
    }

    return val;
}

static inline int ftype_ok(mode_t mode) 
{
    return S_ISREG(mode) || S_ISFIFO(mode) || S_ISSOCK(mode) || S_ISDIR(mode);
}

static int 
openfile(const char *filename)
{
    int fd;
    struct stat st;
    
    if ((fd = open(filename, O_WRONLY|O_CREAT|O_NOCTTY|O_NONBLOCK|O_LARGEFILE, 0400)) < 0) {
	print_err("Unable to open file '%s': %s\n", filename, strerror(errno));
	return -1;
    }
    if (fstat(fd, &st) < 0) {
	print_err("Unable to fstat opened file: %s\n", strerror(errno));
	return -1;
    }
    if (!ftype_ok(st.st_mode)) {
	print_err("File '%s' is not an acceptable file type (i.e. not a "
		  "regular file, socket, or FIFO)\n", filename);
	errno = EINVAL;
	return -1;
    } else if (S_ISDIR(st.st_mode)) {
	print_err("File '%s' is a directory: use the '--dir' flag for "
		  "directories\n", filename);
	errno = EISDIR;
	return -1;
    }

    return fd;
}

#if 0
static int 
opendirectory(const char *dirname)
{
    int fd;
    struct stat st;
    
    if ( (fd = open(dirname, O_NOCTTY|O_NONBLOCK)) == -1) {
	print_err("Unable to open directory '%s': %s\n", dirname,
		  strerror(errno));
	return -1;
    }
    if (fstat(fd, &st)) {
	print_err("Unable to fstat opened file: %s\n", strerror(errno));
	return -1;
    }
    if (!S_ISDIR(st.st_mode)) {
	print_err("'%s' is not a directory\n", dirname);
	errno = ENOTDIR;
	return -1;
    }

    return fd;
}
#endif

/* file synchronization behaviors */
enum {
    ATOMIC,
    BACKUP,
    CLOBBER
};

/* Values for long options w/o single-char equivalents. */
enum {
   opt_run = 1000,
   opt_stop,
   opt_term,
   opt_kill,
   opt_abort,
   opt_sync,
   opt_nosync,
   opt_atomic,
   opt_backup,
   opt_clobber,
   opt_version
};

/* Type of destination */
enum {
   dest_default,
   dest_file,
   dest_dir,
   dest_fd,
   dest_cwd
};

int main(int argc, char **argv)
{
    struct cr_chkpt_req req;
    int dest_type = dest_default;
    int chkpt_fd = -1;
    int do_atomic = 1;
    int do_backup = 0;
    int do_sync = 1;
    char * chkpt_file = NULL;	/* file to checkpoint to */
    char * chkpt_dir = NULL;	/* directory to checkpoint to */
    char * chkpt_to = NULL;	/* name of checkpoint file or dir */
    char * rename_to = NULL;	/* final checkpoint file/dir, if different */
    char * backup_to = NULL;	/* backup file/dir, if needed */
    char * parent_dir = NULL;   /* parent directory of checkpoint */

    int secs = 0;
    int err;
    pid_t target = 0;
    cr_scope_t target_type = CR_SCOPE_PROC;
    int signal = 0;

    /* Parse cmdline options */
    char * shortflags = "f:d:F:S:pgsct:vh?";  /* 1 colon == requires argument */
    struct option longflags[] = {
	/* target_type: */
	{ "pid",     no_argument      , 0, 'p' },
	{ "process", no_argument      , 0, 'p' },
	{ "pgid",    no_argument      , 0, 'g' },
	{ "group",   no_argument      , 0, 'g' },
	{ "sid",     no_argument      , 0, 's' },
	{ "session", no_argument      , 0, 's' },
	{ "tree",    no_argument      , 0, 'T' },
	/* destination: */
	{ "file",    required_argument, 0, 'f' },
	{ "dir",     required_argument, 0, 'd' },
	{ "fd",      required_argument, 0, 'F' },
	{ "cwd",     no_argument,       0, 'c' },
	/* creation/replacement policy */
	{ "atomic",  no_argument,	0, opt_atomic },
	{ "backup",  optional_argument,	0, opt_backup },
	{ "clobber", no_argument,	0, opt_clobber },
	/* signal options: */
	{ "signal",  required_argument, 0, 'S' },
	{ "run",     no_argument,       0, opt_run },
	{ "stop",    no_argument,       0, opt_stop },
	{ "term",    no_argument,       0, opt_term },
	{ "kill",    no_argument,       0, opt_kill },
	{ "abort",   no_argument,       0, opt_abort },
	/* fsync options: */
	{ "sync",    no_argument,       0, opt_sync},
	{ "nosync",  no_argument,       0, opt_nosync},
	/* misc options: */
	{ "time",    required_argument, 0, 't' },
	{ "verbose", no_argument,       0, 'v' },
	{ "help",    no_argument,       0, '?' },
	{ "version", no_argument,       0, opt_version },
	{ 0,	     0,			0,  0  }
    };

    while (1) {
	int longindex = -1;
	int opt = getopt_long(argc, argv, shortflags, longflags, &longindex);

	if (opt == -1)
	    break;		    /* reached last option */

	switch (opt) {
	    case 'p':
		target_type = CR_SCOPE_PROC;
		break;
	    case 'g':
		target_type = CR_SCOPE_PGRP;
		break;
	    case 's':
		target_type = CR_SCOPE_SESS;
		break;
	    case 'T':
		target_type = CR_SCOPE_TREE;
		break;
	    case 'f':
		if (dest_type != dest_default)
		    die (EINVAL, "conflicting destinations provided\n");
	 	dest_type = dest_file;
		chkpt_file = strdup(optarg);
		if (!chkpt_file) 
		    die(ENOMEM, "strdup failed on string '%s'\n", optarg);
		break;
	    case 'd':
#if 1
		die(EINVAL, "-d flag not yet implemented\n");
#else
		if (dest_type != dest_default)
		    die (EINVAL, "conflicting destinations specified\n");
	 	dest_type = dest_dir;
		chkpt_dir = strdup(optarg);
		if (!chkpt_dir) 
		    die(ENOMEM, "strdup failed on string '%s'\n", optarg);
		break;
#endif
	    case 'F':
		if (dest_type != dest_default)
		    die (EINVAL, "conflicting destinations specified\n");
	 	dest_type = dest_fd;
		chkpt_fd = readint(optarg, argv[0]);
		break;
	    case 'c':
		if (dest_type != dest_default)
		    die (EINVAL, "conflicting destinations specified\n");
	 	dest_type = dest_cwd;
		/* nothing to do */
		break;
	    case opt_atomic:
		do_atomic = 1;
		do_backup = 0;
		break;
	    case opt_backup:
		do_atomic = 1;
		do_backup = 1;
		if (optarg) {
		    backup_to = strdup(optarg);
		    if (!backup_to)
			die(errno, "Error during strdup: %s\n", strerror(errno));
		}
		break;
	    case opt_clobber:
		do_atomic = 0;
		do_backup = 0;
		break;
	    case 'S':
		signal = readint(optarg, argv[0]);
		if ((signal <= 0) || (signal >= _NSIG)) {
		    die(EINVAL,
			"Valid signal numbers are from 1 to %d, inclusive.\n",
			_NSIG - 1);
		}
		break;
	    case opt_run: /* --run */
		signal = 0;
		break;
	    case opt_stop: /* --stop */
		signal = SIGSTOP;
		break;
	    case opt_term: /* --term */
		signal = SIGTERM;
		break;
	    case opt_kill: /* --kill */
		signal = SIGKILL;
		break;
	    case opt_abort: /* --abort */
		signal = SIGABRT;
		break;
	    case opt_nosync: /* --nosync */
		do_sync = 0;
		break;
	    case opt_sync: /* --sync */
		do_sync = 1;
		break;
	    case 't':
		secs = readint(optarg, argv[0]);
		if (secs < 0) {
		    die(EINVAL, "Time limit must be non-negative.\n");
		}
		break;
	    case 'v':
		++verbose;
		break;
	    case opt_version:
		print_version();
		exit(0);
		break;
	    case '?':
	    case 'h':
		usage(stdout);
		exit(0);
	    default:
		usage(stderr);
		exit(-1);
	}
    }

    /* Grab id if exactly one arg left.
       optind is a magic global set by getopt_long, and is index of first 
       non-flag parameter, if any */
    if (argc - optind == 1) {
	target = readint(argv[optind], argv[0]);
    } else {
	usage(stderr);
    }

    if (dest_type == dest_fd) {
	#define NAMELEN 64 
	char buf[NAMELEN];
	snprintf(buf, NAMELEN, "<fd%d>", chkpt_fd);
	chkpt_to = strdup(buf);
	if (!chkpt_to)
	    die(errno, "Could not duplicate string '%s'\n", buf);
    } else {
	char buf[NAMELEN];

	switch (dest_type) {
	    case dest_dir:
		chkpt_to = chkpt_dir;
		break;

	    case dest_default: /* default -> cwd */
	    case dest_cwd:
		snprintf(buf, NAMELEN, "context.%d", target);
		chkpt_file = strdup(buf);
		if (!chkpt_file)
		    die(errno, "Could not duplicate string '%s'\n", buf);
		/* fall through */
	    case dest_file:
		chkpt_to = chkpt_file;
		break;

	    default:
		die(1, "Invalid dest_type\n");
	}

	/* get parent directory name */
	parent_dir = strdup(chkpt_to);
	if (!parent_dir)
	    die(errno, "Error in strdup: %s", strerror(errno));
	parent_dir = dirname(parent_dir);

	if (do_atomic) {
	    char *base = strdup(chkpt_to);
	    if (!base)
		die(errno, "Error in strdup: %s\n", strerror(errno));
	    /* save final target name */
	    rename_to = chkpt_to;  
	    /* use '.context.pid.tmp'-style name for checkpoint file */
	    chkpt_to = (char *)malloc(strlen(chkpt_to) + 10);
	    if (!chkpt_to)
		die(errno, "Malloc failed!\n");
	    strcpy(chkpt_to, parent_dir);
	    strcat(chkpt_to, "/.");
	    strcat(chkpt_to, basename(base));
	    strcat(chkpt_to, ".tmp");
	    free(base);
	}
    }

    if (verbose) 
	printf("targetfile='%s', parent dir='%s', rename=%s\n", 
		chkpt_to, parent_dir, rename_to);

    /* TODO:  make sure no other checkpoint is occurring to the same file? */
    if (chkpt_fd >= 0) {
	/* silently ignore the atomic/backup flags */
	do_atomic = do_backup = 0;
    } else if (chkpt_file) {
	if ((chkpt_fd = openfile(chkpt_to)) == -1)
	    die(errno, "Unable to open checkpoint file '%s'\n", chkpt_file);
	/* after this point, remove checkpoint file if there's an error */
	to_remove = chkpt_to;
    } else {
	die(EINVAL, "directories not yet supported\n");
    }


    /* BUILD THE REQUEST */
    req.cr_target = target;
    req.cr_scope = target_type;
    req.cr_secs = secs;	/* 0 == unbounded */
    req.cr_opts = NULL;
    req.cr_fd = chkpt_fd;
    req.dump_format = cr_format_vmadump;
    req.signal = signal;
    req.flags = 0;

    /* CONNECT TO THE KERNEL */
    err = cri_connect();
    if (err < 0) {
      if (errno == ENOSYS) {
	  die(errno, "Checkpoint failed: support missing from kernel\n");
      } else {
          die(errno, "Failed cri_connect(): %s\n", cr_strerror(errno));
      }
    }

    /* ISSUE THE REQUEST */
    err = cri_syscall(CR_OP_CHKPT_REQ, (uintptr_t)&req);
    if (err < 0) {
	if (errno == CR_ENOSUPPORT) {
	    die(errno, "Checkpoint failed: support missing from application\n");
	} else {
	    die(errno, "ioctl(" CR_CTRL_FILE ",  CR_OP_CHKPT_REQ): %s\n",
	        cr_strerror(errno));
	}
    } else if (verbose) {
	fprintf(stderr, "checkpoint request issued\n");
    }

    /* WAIT FOR THE REQUEST TO COMPLETE */
    err = cri_syscall(CR_OP_CHKPT_DONE, (uintptr_t)NULL);
    if (err < 0) {
	die(errno, "ioctl(" CR_CTRL_FILE ", CR_OP_CHKPT_DONE): %s\n",
	    cr_strerror(errno));
    } else if (verbose) {
	fprintf(stderr, "checkpoint request completed\n");
    }

    /* REAP THE COMPLETED REQUEST */
    err = cri_syscall(CR_OP_CHKPT_REAP, CRI_SYSCALL_NOARG);
    if (err < 0) {
	if (errno == CR_ETEMPFAIL) {
	    die(errno, "Checkpoint cancelled by application: try again later\n");
	} else if (errno == CR_EPERMFAIL) {
	    die(errno, "Checkpoint cancelled by application: unable to checkpoint\n");
	} else if (errno == CR_ENOSUPPORT) {
	    die(errno, "Checkpoint failed: support missing from application\n");
	} else {
	    die(errno, "ioctl(" CR_CTRL_FILE ", CR_OP_CHKPT_REAP): %s\n",
		cr_strerror(errno));
	}
    }

    if (do_backup) {
	struct stat s;

	assert(do_atomic);
	/* only backup if there's a pre-existing checkpoint */
	if (!stat(rename_to, &s)) {
	    if (!backup_to) {
		int bnum = 1;
		backup_to = (char *)malloc(strlen(rename_to) + 10);
		while (1) {
		    strcpy(backup_to, rename_to);
		    sprintf(backup_to, "%s.~%d~", rename_to, bnum++);
		    if (stat(backup_to, &s)) 
			break;
		    if (bnum > (1<<16))
			die(-1, "This is absurd\n");
		}
	    }
	    if (rename(rename_to, backup_to))
		die(errno, "Unable to rename '%s' to '%s': %s\n",
		    rename_to, backup_to, strerror(errno));
	}
    }
    if (do_atomic) {
	assert(rename_to && strlen(rename_to));
	if (rename(chkpt_to, rename_to))
	    die(errno, "Unable to rename '%s' to '%s': %s\n",
		chkpt_to, rename_to, strerror(errno));
    }

    if (do_sync) {
	DIR *dir;

	/* COMMIT TO DISK */
	err = fsync(chkpt_fd);
	if (err < 0)
	    die(errno, "Error syncing checkpoint to disk: %s\n", 
		strerror(errno));
	if (parent_dir != NULL) {
	    /* sync parent directory, too, to ensure checkpoint shows up */
	    if (!(dir = opendir(parent_dir)))
		die(errno, "unable to opendir(%s): %s\n",
		    parent_dir, strerror(errno));
	    /* ignore fsync errors that might be from using NFS */
	    if (fsync(dirfd(dir)) && errno != EROFS && errno != EINVAL)
		fprintf(stderr, "Warning: unable to sync directory '%s': errno=%d\n",
			parent_dir, errno);
	    /* TODO: if checkpoint to directory, fsync all context.pid files */
	}
    }

    return 0;
}
