/*-------------------------------------------------------------------------
 *  vmadtheard.c: tests multithreaded dumps
 *
 *  Copyright (C) 1999-2001 by Erik Hendriks <erik@hendriks.cx>
 *    Originally derived from vmadtest.c
 *    by Eric Roman <ERoman@lbl.gov>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: vmadthread.c,v 1.5 2002/08/29 22:38:19 eroman Exp $
 *-----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/mman.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "vmadump.h"

#ifndef __NR_vmadump
#include <sys/bproc.h>
#endif

#define GROWCHUNK 10240

#define STACK_SIZE 8192

#define DEFAULT_FILE "dump"
#if HAVE_BINFMT_VMADUMP
#define DUMPMODE 0777
#else
#define DUMPMODE 0666
#endif

extern char **environ;

void sighandler(void) { fprintf(stderr, "In sighandler!\n"); }

#ifdef __NR_vmadump
int vmadump(int fd, int flags) {
    return syscall(__NR_vmadump, VMAD_DO_DUMP, fd, flags);
}

int vmaundump(int fd) {
    return syscall(__NR_vmadump, VMAD_DO_UNDUMP, fd);
}

int vmathreaddump(struct vmadump_threaddump_args *args) {
    return syscall(__NR_vmadump, VMAD_DO_THREADDUMP, args);
}

int vmathreadundump(struct vmadump_threadundump_args *args) {
    return syscall(__NR_vmadump, VMAD_DO_THREADUNDUMP, args);
}
#else
int vmadump(int fd, int flags) {
    return bproc_dump(fd, flags);
}

int vmaundump(int fd) {
    return bproc_undump(fd);
}

#endif

void
grow_mm(void)
{
    char *foo;

    fprintf(stderr, "PID %d: Testing memory ops...\n", getpid());

    foo = alloca(GROWCHUNK);
    memset(foo, 0, GROWCHUNK);
    fprintf(stderr, ".");
    fflush(stderr);

    foo = malloc(GROWCHUNK);
    memset(foo, 0, GROWCHUNK);
    fprintf(stderr, "+");
    fflush(stderr);
}

int
dumping_thread_brains(void *arg)
{
    int old_pid;
    int retval;

    old_pid = getpid();

    retval = vmathreaddump(arg);

    if (!retval) {
      fprintf(stderr, "PID %d.  AKA The PID formerly known as: %d\n", 
	  getpid(), old_pid);
      // grow_mm();                        /* just once, to avoid death */
    }

    return retval;
}

/* new threads call undump to revive the original process */
int
undumping_thread_brains(void *arg)
{
    int retval;

    retval = vmathreadundump(arg);

    /* If we get here, something has gone wrong */
    return retval;
}

void create_threads(int (*thread_brains)(void *), int num_threads, void *args)
{
    int i;

    fprintf(stderr, "%d creating %d threads: ", getpid(), 
	((struct vmadump_threaddump_args *)args)->num_threads-1);

    /* spawn off additional threads as necessary */
    for (i = 0; i < num_threads; ++i) {
        void *addr;
	int pid;

	/* get stack space for the thread */
	addr = mmap(NULL, sizeof(int) * STACK_SIZE,
		 PROT_READ | PROT_WRITE | PROT_EXEC,
		 MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);

	pid = clone(thread_brains, addr + STACK_SIZE,
			 CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND,
			 args);
	fprintf(stderr, "%d ", pid);
    }
    fprintf(stderr, "\n");

}

int 
restore_threads(int fd, int num_threads)
{
    struct vmadump_threadundump_args args;
    int r;

    /* create an args structure describing what's going on */
    args.num_threads = num_threads;
    args.flags = 0;
    args.fd = fd;

    /* create some new threads */
    create_threads(undumping_thread_brains, num_threads - 1, &args);

    /* now restore ourselves */
    r = vmathreadundump(&args);    
    
    return r;
}

void
dump_threads(char *filename, int flags, int threads)
{
    int fd, r;
    struct vmadump_threaddump_args args;

    fprintf(stderr, "Dumping to: %s\n", filename);
    fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, DUMPMODE);
    if (fd == -1) {
	perror("open");
	exit(1);
    }

    args.num_threads = threads;
    args.flags = flags;
    args.fd = fd;

    /* NOTE:  Main thread is dumped, too.  So spawn one less than threads */
    create_threads(dumping_thread_brains, threads - 1, &args);

    /* now dump */
    r = vmathreaddump(&args);
    fprintf(stderr, "vmathreaddump(%d,%p)=%d\n", fd, &args, r);
    if (r == -1) {
	perror("VMAD_DO_THREADDUMP");
	exit(1);
    }

    close(fd);
    if (r) {
	fprintf(stderr, "dump complete - exiting.\n");
	exit(0);
    }

    /* r == 0... Which means that we were dumped and we're getting
     * here because we've been undumped. */
    fprintf(stderr, "Whoo!  I've been undumped!  Kick ass!\n");
    fprintf(stderr, "New pid: %d\n", (int) getpid());

    // grow_mm();
}

void usage(char *arg0) {
    printf(
"Usage: %s -d [-f file] [-e] [-l] [-o]\n"
"       %s -u [-f file]\n"
"\n"
"       This program is a simple test program for the Beowulf VMAdump\n"
"       system.  It will try to dump itself (with -d) or restore a dump\n"
"       with -u.\n"
"\n"
"       -1         Dump one thread.\n"
"       -2         Dump two threads.\n"
"       -4         Dump four threads.\n"
"\n"
"       -d         Dump to a file.\n"
"       -u         Undump from a file.\n"
"       -f file    Dump to/from file instead of the default (%s)\n"
"       -F         Perform a floating point operation before (un)dumping.\n"
"\n"
"       The following options only affect dumping.  (-d)\n"
"       -e         Use VMAD_DUMP_EXEC flag.\n"
"       -l         Use VMAD_DUMP_LIBS flag.\n"
"       -o         Use VMAD_DUMP_OTHER flag.\n"
, arg0, arg0, DEFAULT_FILE);
}

double c;

int main(int argc, char *argv[]) {
    int i;
    int w_pid;
    int status;

    int threads = 2;
    int fd, r,c ;
    int dump=-1;
    int dump_flags = 0;
    int fpop = 0;
    char *filename = DEFAULT_FILE;
    
    while ((c=getopt(argc, argv, "m124hvf:Fduelox")) != EOF) {
	switch (c) {
	case 'h':
	case 'v': usage(argv[0]); exit(0);
        case '1': threads = 1; break;
        case '2': threads = 2; break;
        case '4': threads = 4; break;
	case 'f': filename = optarg; break;
	case 'd': dump = 0; break;
	case 'u': dump = 1; break;
	case 'e': dump_flags |= VMAD_DUMP_EXEC; break;
	case 'l': dump_flags |= VMAD_DUMP_LIBS; break;
	case 'o': dump_flags |= VMAD_DUMP_OTHER; break;
	case 'F': fpop = 1; break;
	default: exit(1);
	}
    }

    if (fpop) {
	double a, b;
	a = 5.3; 
	b = getpid();
	c = a * b;
    }

    switch (dump) {
    case 0:
	/* Set a sighandler to test sighandler save/restore */
	signal(SIGUSR1, (void (*)(int)) sighandler);
	dump_threads(filename, dump_flags, threads);
        for (i=1; i<threads; i++) {
            w_pid = waitpid(-1, &status, __WCLONE);
	    if (w_pid < 0) {
	        perror("waitpid");
	    }

            printf("w_pid = %d:\t", w_pid);
            printf("status = %d\t", status);
            printf("WIFEXITED = %d\t", WIFEXITED(status));
            printf("WEXITSTATUS = %d\n", WEXITSTATUS(status));
        }
	_exit(0);

	break;
    case 1:
	fprintf(stderr, "Undumping from: %s\n", filename);
	fd = open(filename, O_RDONLY);
	if (fd == -1) { perror("open"); exit(1); }
	r = restore_threads(fd, threads);
	fprintf(stderr, "VMAD_DO_UNDUMP failed: err=%d (%s)\n",
		errno, strerror(errno));
	break;
    default:
	usage(argv[0]);
	exit(1);
  }
  exit(0);
}

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
