/*-------------------------------------------------------------------------
 *  vmadump_threads.c:  Virtual Memory Area dump/restore routines
 *
 *  2004, The Regents of the University of California, through Lawrence
 *  Berkeley National Laboratory (subject to receipt of any required
 *  approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 *  Portions may be copyrighted by others, as may be noted in specific
 *  copyright notices within specific files.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: vmadump_threads.c,v 1.9 2007/02/14 21:30:18 phargrov Exp $
 *
 * THIS FILE SPECIFIC TO BLCR <http://ftg.lbl.gov/checkpoint>
 * It handles thread-management for dumps of multithreaded processes.
 * This file is shared by both the Linux 2.4 and 2.6 versions of vmadump.
 *-----------------------------------------------------------------------*/

#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  #ifndef EXPORT_SYMTAB
    #define EXPORT_SYMTAB
  #endif
  #include <linux/autoconf.h>
  #if defined(CONFIG_SMP) && ! defined(__SMP__)
    #define __SMP__
  #endif
  #if defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
    #define MODVERSIONS
  #endif
  #if defined(MODVERSIONS)
    #include <linux/modversions.h>
  #endif
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>

#define __VMADUMP_INTERNAL__
#include "vmadump.h"
#include "blcr_config.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  #undef MOD_INC_USE_COUNT
  #define MOD_INC_USE_COUNT
  #undef MOD_DEC_USE_COUNT
  #define MOD_DEC_USE_COUNT
#endif 

#ifdef DEFINE_SPINLOCK
  DEFINE_SPINLOCK(thread_req_list_lock);
#else
  spinlock_t thread_req_list_lock = SPIN_LOCK_UNLOCKED;
#endif
LIST_HEAD(thread_req_list);

#ifndef thread_group_leader
  #define thread_group_leader(p) ((p)->pid == (p)->tgid)
#endif

/* -------------------------------------------------------------------
 * Routines for tracking outstanding thread requests
 */
static struct vmadump_thread_req *
alloc_thread_req(struct task_struct *task, int threadop, struct file *file, int num_threads, int flags)
{
    struct vmadump_thread_req *req;
    
    req = kmalloc(sizeof(struct vmadump_thread_req), GFP_ATOMIC);
    if (req) {
	req->mm = task->mm;                    
        req->num_threads = num_threads;        
        atomic_set(&req->num_awake, num_threads);
        init_waitqueue_head(&req->wait);
        atomic_set(&req->num_frozen, num_threads);
        req->file = file;
	sema_init(&req->sema, 1);
        req->done_leader = 0;
        req->flags = flags;
        atomic_set(&req->ref_count, 1);
	INIT_LIST_HEAD(&req->req_list);
	req->type = threadop;
	req->error = 0;
    }

    return req;
}

/*
 * Note:  Must be called with thread_req_list_lock held
 */
static struct vmadump_thread_req *
find_thread_req(struct task_struct *task)
{
    struct list_head *tmp;
    struct vmadump_thread_req *req;

    list_for_each(tmp, &thread_req_list) {
        req = list_entry(tmp, struct vmadump_thread_req, req_list);
        if (req->mm == task->mm) {
	    atomic_inc(&req->ref_count);
	    return req;
	}
    }

    return NULL;
}

static struct vmadump_thread_req *
get_thread_req(int threadop, struct task_struct *task, struct file *file, 
               int num_threads, int flags)
{
    struct vmadump_thread_req *req;

    spin_lock(&thread_req_list_lock);
    req = find_thread_req(task);
    if (!req) {
	/* we're the first one's in.  Put a req in for everyone else */
	req = alloc_thread_req(task, threadop, file, num_threads, flags);
	if (!req) {
	    goto err;
	}

	/* ... put us on the list */
	list_add(&req->req_list, &thread_req_list);
    }

err:
    spin_unlock(&thread_req_list_lock);
    return req;
}


static void
put_thread_req(struct vmadump_thread_req *req)
{
    spin_lock(&thread_req_list_lock);
    if(atomic_dec_and_test(&req->ref_count)) {
        list_del(&req->req_list);
        kfree(req);
    }
    spin_unlock(&thread_req_list_lock);
}

long
vmadump_thaw_threads(struct file *file, struct pt_regs *regs, 
                     struct vmadump_threadundump_args *args)
{
    struct vmadump_thread_req *req;
    int pid;
    int retval = 0;
    int i_am_done = 0;

    MOD_INC_USE_COUNT;

    req = get_thread_req(VMAD_THREAD_UNDUMP, current, file,
			 args->num_threads, args->flags);

    if (!req) {
	retval = -ENOMEM;
	goto err;
    }

    if (req->type != VMAD_THREAD_UNDUMP) {
	retval = -EAGAIN;
	goto err1;
    }

    /* XXX: This is paranoia, can it go away? */
    if (!atomic_read(&req->num_awake)) {
	/* if num_threads is zero it means that either this req has
	 * happened already, or that it's in progress now.  In any case,
	 * this process no longer has any reason to be a part of this req.
	 */
	printk("vmadump[%d]: Warning: Threaddump invoked after barrier\n",
	       current->pid);
	retval = -EAGAIN;
	goto err1;
    }

    /* First a barrier to ensure all threads are accounted for.
     * XXX: this ensures they are all "here", while is is probably
     * sufficient just to ensure they are in cr_rstrt_child().
     */
    if (atomic_dec_and_test(&req->num_awake)) {
        wake_up(&req->wait);
    } else {
        wait_event_interruptible(req->wait, !atomic_read(&req->num_awake));
	if (atomic_read(&req->num_awake)) {
	    retval = -EINTR;
	    goto err1;
        }
    }

    /* Now let exactly one thread-group leader do a full undump, and
     * ensure everyone else does REGSONLY.
     * This and the matching logic at dump time together ensure that we
     * preserve the thread-group structure that was spawned in user-space.
     */
    if (thread_group_leader(current)) {
        down(&req->sema);
	if (!req->done_leader) {
	    retval = vmadump_thaw_proc(0, req->file, regs, 0, &pid);
	    req->error = retval;
	    memcpy(req->comm, current->comm, sizeof(req->comm));
#ifdef ARCH_HAS_SETUP_ADDITIONAL_PAGES
	    vmad_vdso_save(req);
#endif
	    i_am_done = 1;
	    req->done_leader = 1;
	    wake_up(&req->wait);
	}
	up(&req->sema);
    }
    if (!i_am_done) {
        wait_event_interruptible(req->wait, req->done_leader);
	if (!req->done_leader) {
	    retval = -EINTR;
	    goto err1;
        }
        down(&req->sema);
        if (!req->error) {
	    retval = vmadump_thaw_proc(0, req->file, regs,
				       VMAD_DUMP_REGSONLY, &pid);
	    req->error = retval;
	    memcpy(current->comm, req->comm, sizeof(current->comm));
#ifdef ARCH_HAS_SETUP_ADDITIONAL_PAGES
	    vmad_vdso_load(req);
#endif
        } else {
	    retval = req->error;
        }
        up(&req->sema);
    }

    if (!retval) {
        args->old_pid = pid;
    }

err1:
    put_thread_req(req);
err:
    MOD_DEC_USE_COUNT;
    return retval;
}

long
vmadump_freeze_threads(struct file *file, struct pt_regs *regs,
                       struct vmadump_threaddump_args *args)
{
    struct vmadump_thread_req *req;
    int retval = 0;
    int i_am_done = 0;

    MOD_INC_USE_COUNT;

    req = get_thread_req(VMAD_THREAD_DUMP, current, file,
			 args->num_threads, args->flags);

    if (!req) {
	retval = -ENOMEM;
	goto err;
    }

    if (req->type != VMAD_THREAD_DUMP) {
	retval = -EAGAIN;
	goto err1;
    }

    /* XXX: This is paranoia, can it go away? */
    if (!atomic_read(&req->num_awake)) {
	/* if num_threads is zero it means that either this req has
	 * happened already, or that it's in progress now.  In any case,
	 * this process no longer has any reason to be a part of this req.
	 */
	printk("vmadump[%d]: Warning: Threaddump invoked after barrier\n",
	       current->pid);
	retval = -EAGAIN;
	goto err1;
    }

    /* First a barrier to ensure all threads are accounted for.
     * XXX: this ensures they are all "here", while is is probably
     * sufficient just to ensure they are in cr_sump_self().
     */
    if (atomic_dec_and_test(&req->num_awake)) {
        wake_up(&req->wait);
    } else {
        wait_event_interruptible(req->wait, !atomic_read(&req->num_awake));
	if (atomic_read(&req->num_awake)) {
	    retval = -EINTR;
	    goto err1;
        }
    }

    /* Now let exactly one thread-group leader do a full dump, and
     * ensure everyone else does REGSONLY.
     * This and the matching logic at undump time together ensure that we
     * preserve the thread-group structure that was spawned in user-space.
     */
    if (thread_group_leader(current)) {
	down(&req->sema);
	if (!req->done_leader) {
	    retval = vmadump_freeze_proc(0, file, regs, req->flags,
					 req->num_threads);
	    i_am_done = 1;
	    req->done_leader = 1;
	    wake_up(&req->wait);
	}
	up(&req->sema);
    }
    if (!i_am_done) {
        wait_event_interruptible(req->wait, req->done_leader);
	if (!req->done_leader) {
	    retval = -EINTR;
	    goto err1;
        }
        down(&req->sema);
	retval = vmadump_freeze_proc(0, req->file, regs,
				     args->flags | VMAD_DUMP_REGSONLY,
				     req->num_threads);
        up(&req->sema);
    }

    /* now a second barrier, to make sure that we've all written */
    if (atomic_dec_and_test(&req->num_frozen)) {
        wake_up(&req->wait);
    } else {
        wait_event_interruptible(req->wait, !atomic_read(&req->num_frozen));
	if (atomic_read(&req->num_frozen)) {
	    retval = -EINTR;
	    goto err1;
        }
    }

err1:
    put_thread_req(req);
err:
    MOD_DEC_USE_COUNT;
    return retval;
}

EXPORT_SYMBOL(vmadump_freeze_threads);
EXPORT_SYMBOL(vmadump_thaw_threads);
